#undef USE_2P
#undef USE_ISEND
#define USE_FLOWCTL

#define SEND_TIME
#define PERFORMANCE
#undef PERF_SHORT
#undef DELAY
#undef FREQUENCIES

#undef USE_GAMMA_TEST

#undef STREAM
#ifdef STREAM
#define SEND_TIME
#define max(a,b) ((a) > (b) ? (a) : (b))
#define NSTREAM max(10,(len ? (10000000/len) : 10000))
#else
#define NSTREAM 1
#endif

#undef CORRECTNESS

#ifdef USE_FLOWCTL
#define gamma_SEND(a,b,c)  gamma_send_flowctl(a,b,c)
#define gammaSEND2P(a,b,c,d,e)  gamma_send_2p_flowctl(a,b,c,d,e)
#define gamma_ISEND(a,b,c)  gamma_isend_flowctl(a,b,c)
#define gammaISEND2P(a,b,c,d,e)  gamma_isend_2p_flowctl(a,b,c,d,e)
#else
#define gamma_SEND(a,b,c)  gamma_send(a,b,c)
#define gammaSEND2P(a,b,c,d,e)  gamma_send_2p(a,b,c,d,e)
#define gamma_ISEND(a,b,c)  gamma_isend(a,b,c)
#define gammaISEND2P(a,b,c,d,e)  gamma_isend_2p(a,b,c,d,e)
#endif

#include <linux/gamma/libgamma.h>
#include <stdio.h>
#include <stdlib.h>
#include "measure.h"

#undef TRACE
#define MASTER_NODE 0
#define SLAVE_NODE 1
#define PORT 1


unsigned char *data,*data1;
int len;

#ifdef USE_2P
void handler(void) {
	gamma_switch_to_buffer(PORT,data+12,len);
	return;
}
#endif


int main(int argc, char **argv) {

#ifdef PERFORMANCE
#ifdef SEND_TIME
	double record_mc[NPROVE];
#endif
	double record_mm[NPROVE];
#endif
	int i,me;
#ifdef CORRECTNESS
	int j;
#endif
#ifdef USE_ISEND
	int hndl;
#endif
#ifdef USE_GAMMA_TEST
	int prev_val;
#endif
	time_586
		t_send_s,
		t_rec;
#ifdef SEND_TIME
	time_586
		t_send_e;
#endif
#ifdef PERFORMANCE
	double time_overhead;
#endif

	gamma_init(2,argc,argv);
   	len = atoi(argv[1]);
	me = gamma_my_node();
#ifdef USE_GAMMA_TEST
	prev_val = 0;
#endif
	/*  Allocate data space (+4096 to allow page alignment)  */
#ifdef USE_2P
	data1 = (unsigned char *)malloc(4096+len*sizeof(unsigned char)+12);
#else
	data1 = (unsigned char *)malloc(4096+len*sizeof(unsigned char));
#endif
	/*  Align to page boundary  */
	data = (unsigned char *)(((unsigned long)data1+4096) & (~4095));

	/*  Pin communication data structures in RAM.  */
#if 0
#ifdef USE_2P
	gamma_mlock(data,len+12);
#else
	gamma_mlock(data,len);
#endif
#endif

#ifdef CORRECTNESS
	/*  Init the message to be exchanged.  */
#ifdef USE_2P
	for (j=0;j<len+12;j++) {
		if (me == MASTER_NODE)
			data[j] = (unsigned char)j;
		else
			data[j] = (unsigned char)0;
	}
#else
	for (j=0;j<len;j++) {
		if (me == MASTER_NODE)
			data[j] = (unsigned char)j;
		else
			data[j] = (unsigned char)0;
	}
#endif
#endif


	if (me == MASTER_NODE) {
#ifdef USE_2P
		gamma_set_active_port_2p(
			PORT,
			SLAVE_NODE,gamma_my_par_pid(),PORT,
			handler,handler,
			PORT,GO_BACK,data,len+12,
			12
		);
#else
		gamma_set_passive_port(
			PORT,
			SLAVE_NODE,gamma_my_par_pid(),PORT,
			PORT,GO_BACK,data,len
		);
#endif
	}
	else {
#ifdef USE_2P
		gamma_set_active_port_2p(
			PORT,
			MASTER_NODE,gamma_my_par_pid(),PORT,
			handler,handler,
			PORT,GO_BACK,data,len+12,
			12
		);
#else
		gamma_set_passive_port(
			PORT,
			MASTER_NODE,gamma_my_par_pid(),PORT,
			PORT,GO_BACK,data,len
		);
#endif
	}

#ifdef PERFORMANCE
	/*  Computing the gamma_time() overhead...  */
	for (i=0;i<NPROVE;i++) {
		gamma_time(t_send_s);
 		gamma_time(t_rec);
		record_mm[i] = gamma_time_diff(t_rec, t_send_s);
	}
	{
	double Mmm,medmm,varmm;
	mean_variance(record_mm,&time_overhead,&Mmm,&medmm,&varmm);
	}
#endif

	gamma_sync();

	for (i=0;i<NPROVE;i++) {
#ifdef STREAM
		int stream;
#endif

		if (me == MASTER_NODE) {
#ifdef CORRECTNESS
#ifdef USE_2P
			for (j=0;j<len+12;j++) data[j] = (unsigned char)j;
#else
			for (j=0;j<len;j++) data[j] = (unsigned char)j;
#endif
			gamma_sync();
#endif

#ifdef PERFORMANCE
			gamma_time(t_send_s);
#endif

#ifdef STREAM
			for (stream = NSTREAM; stream; stream--) {
#endif

#ifdef USE_ISEND
#ifdef USE_2P
			hndl = gammaISEND2P(PORT,data,12,data+12,len);
#else
			hndl = gamma_ISEND(PORT,data,len);
#endif
#else
#ifdef USE_2P
			gammaSEND2P(PORT,data,12,data+12,len);
#else
			gamma_SEND(PORT,data,len);
#endif
#endif

#ifdef STREAM
			}
#endif
#ifdef PERFORMANCE
#ifdef SEND_TIME
			gamma_time(t_send_e);
#endif
#endif
#ifdef TRACE
printf("[%d]: master sent\n",i);fflush(stdout);
#endif
#ifdef USE_GAMMA_TEST
			while ((gamma_test(PORT) - prev_val) < 1);
			prev_val = gamma_test(PORT);
#else
			gamma_wait(PORT,1);
#endif
#ifdef TRACE
printf("[%d]: master recvd\n",i);fflush(stdout);
#endif
#ifdef PERFORMANCE
 			gamma_time(t_rec);
#endif
#ifdef USE_ISEND
			gamma_wsend(hndl);
#endif

#ifdef CORRECTNESS
			/*  Check data correctness.  */
#ifdef USE_2P
                        for (j=0;j<len+12;j++)
                                if (data[j] != (unsigned char)j) {
                                        int k;
                                        printf("PING:  data error\n");
                                        for (k=j-5;k<j+10;k++)
						printf("%d:%d ",k,data[k]);
                                        printf("\n");
					break;
                                }
#else
                        for (j=0;j<len;j++)
                                if (data[j] != (unsigned char)j) {
                                        int k;
                                        printf("PING:  data error\n");
                                        for (k=j-5;k<j+10;k++)
						printf("%d:%d ",k,data[k]);
                                        printf("\n");
					break;
                                }
#endif
#endif

#ifdef PERFORMANCE
#ifdef SEND_TIME
			record_mc[i] =
			gamma_time_diff(t_send_e, t_send_s)-time_overhead;
#endif
#ifdef STREAM
			record_mm[i] =
			gamma_time_diff(t_rec, t_send_s)-time_overhead;
#else
			record_mm[i] =
			gamma_time_diff(t_rec, t_send_s)/2-time_overhead;
#endif
#endif

		}
		else {
#ifdef CORRECTNESS
#ifdef USE_2P
			for (j=0;j<len+12;j++) data[j] = 0;
#else
			for (j=0;j<len;j++) data[j] = 0;
#endif
			gamma_sync();
#endif

#ifdef USE_GAMMA_TEST
			while ((gamma_test(PORT) - prev_val) < NSTREAM);
			prev_val = gamma_test(PORT);
#else
			gamma_wait(PORT,NSTREAM);
#endif
#ifdef TRACE
printf("[%d]: slave recvd\n",i);fflush(stdout);
#endif
#ifdef STREAM
			gamma_send_flowctl(PORT,NULL,0);
#else
#ifdef USE_ISEND
#ifdef USE_2P
			hndl = gammaISEND2P(PORT,data,12,data+12,len);
			gamma_wsend(hndl);
#else
			hndl = gamma_ISEND(PORT,data,len);
			gamma_wsend(hndl);
#endif
#else
#ifdef USE_2P
			gammaSEND2P(PORT,data,12,data+12,len);
#else
			gamma_SEND(PORT,data,len);
#endif
#endif
#endif
#ifdef TRACE
printf("[%d]: slave sent\n",i);fflush(stdout);
#endif
		}

	}

	/*
	 *  Force printout of kernel-level time measurements,
	 *  if enabled.  Otherwise nothing happens.
	 */
	gamma_timers(t_send_s,t_rec);

#if 0
	gamma_munlockall();
#endif
	free(data1);
	gamma_exit();

#ifdef PERFORMANCE
	if (me == MASTER_NODE) {
#ifdef PERF_SHORT
#ifdef SEND_TIME
		double mmc,Mmc,medmc,varmc;
		mean_variance(record_mc,&mmc,&Mmc,&medmc,&varmc);
#ifdef DELAY
	        printf("%d %g\n",len,medmc/NSTREAM);
#else
		if (!len)
		        printf("%d %g\n",len,medmc/NSTREAM);
		else
		        printf("%d %g\n",len,NSTREAM*len/medmc);
#endif
#else
		double mmm,Mmm,medmm,varmm;
		mean_variance(record_mm,&mmm,&Mmm,&medmm,&varmm);
#ifdef DELAY
	        printf("%d %g\n",len,medmm/NSTREAM);
#else
		if (!len)
		        printf("%d %g\n",len,medmm/NSTREAM);
		else
		        printf("%d %g\n",len,NSTREAM*len/medmm);
#endif
#endif
#else
#ifdef SEND_TIME
		double mmc,Mmc,medmc,varmc;
#endif
#ifndef STREAM
		double mmm,Mmm,medmm,varmm;
#endif
		printf("\n\n");
		printf("PERFORMANCE (microsec, MB/sec) (lowest < average < highest):\n\n");
#ifdef SEND_TIME
		mean_variance(record_mc,&mmc,&Mmc,&medmc,&varmc);
	        printf("total send time        %g < %g < %g\n",
			mmc,medmc,Mmc);
#ifdef STREAM
		if (!len)
	        printf("avg. send time         %g < %g < %g\n",
			mmc/NSTREAM,medmc/NSTREAM,
			Mmc/NSTREAM);
		else
#else
		if (len)
#endif
	        printf("send throughput        %g < %g < %g\n",
			NSTREAM*len/Mmc,NSTREAM*len/medmc,
			NSTREAM*len/mmc);
#endif
#ifndef STREAM
		mean_variance(record_mm,&mmm,&Mmm,&medmm,&varmm);
	        printf("total delay            %g < %g < %g\n",
			mmm,medmm,Mmm);
#ifdef STREAM
		if (!len)
	        printf("avg. delay             %g < %g < %g\n",
			mmm/NSTREAM,medmm/NSTREAM,
			Mmm/NSTREAM);
		else
#else
		if (len)
#endif
	        printf("end-to-end throughput  %g < %g < %g\n",
			NSTREAM*len/Mmm,NSTREAM*len/medmm,
			NSTREAM*len/mmm);
#endif
		printf("\n\n");

	        printf("VARIANCE:  \n");
#ifdef SEND_TIME
	        printf("send time        %g\n", varmc);
#endif
#ifndef STREAM
	        printf("end-to-end delay %g\n", varmm);
#endif
		printf("\n\n");

#ifdef FREQUENCIES
	        printf("FREQUENCIES:\n");
#ifdef SEND_TIME
		printf("send time:\n");
		frequencies(record_mc,mmc,Mmc);
#endif
#ifndef STREAM
		printf("end-to-end delay:\n");
#endif
		frequencies(record_mm,mmm,Mmm);
#endif
		fflush(stdout);

#endif
	}
#endif
	return 0;
}
