/***************************************************************************
    COPYRIGHT NOTE

    GAMMA:		Genoa Active Message MAchine
			BIND2 layer:
			Basic Interface for Network Device Drivers.
			Derived from the Linux network driver 3c59x.c

    Copyright		(C) 2000-2001 Alessio Menale, Giuseppe Ciaccio,
			Salvatore Filippone
    e-mail:		a.menale@tiscalinet.it, ciaccio@disi.unige.it,
			sfilippone@uniroma2.it

    This program  or include  file  is free software,  either  original  or
    derived from other  software made available  under the terms of the GNU
    General Public License. You can redistribute and/or modify this program
    or include  file under the  terms of the GNU  General Public License as
    published  by the  Free Software Foundation;  either version 2  of  the
    License, or (at your option) any later version.

    THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR   IMPLIED
    WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
    MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
    NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT,  INDIRECT,
    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
    NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
    USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
    ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
    THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

    You should have received a copy of the  GNU General Public License along
    with this program; if not, write  to the Free Software Foundation, Inc.,
    675 Mass Ave, Cambridge, MA 02139, USA.

    END OF COPYRIGHT NOTE
****************************************************************************/

#define MAX_FLUSH_WAIT  100000	/*  Max waiting (usec) for Tx ring to flush  */
#define GAMMA_TX_RING_SIZE 128	/*
				 *  Size of Tx descriptor ring.
				 */
#define GAMMA_RX_RING_SIZE 1024	/*
				 *  Size of Rx descriptor ring.
				 *  We need a really large ring size due to
				 *  the static credit-based flow control.
				 */
#ifdef GAMMA_FULL_DUPLEX
#define MAX_CREDIT_SIZE 28
#define LOW_WATERMARK 4		/*  Must be > 0  */
#else
#define MAX_CREDIT_SIZE 30
#define LOW_WATERMARK 2		/*  Must be > 0  */
#endif
				/*
				 *  The sum MAX_CREDIT_SIZE + LOW_WATERMARK
				 *  must NOT exceed the following quantity:
				 *  (GAMMA_RX_RING_SIZE-20)/(total_num_nodes-1)
				 */
#define FRAME_SIZ 1500		/*  Max payload chunk = 1476 bytes  */
#define STEPS_THEN_STOP 1

/*======================================================================*/
/*  TX MACROS                                                           */
/*======================================================================*/

#define tx_optimal_frag(l) FRAME_SIZ

#define tx_common_decl_and_init \
struct vortex_private *vp = (struct vortex_private *)gamma_device->priv;\
long ioaddr = gamma_device->base_addr;\
int i

#define tx_ring_head (vp->dirty_tx % GAMMA_TX_RING_SIZE)
#define tx_ring_tail (vp->cur_tx % GAMMA_TX_RING_SIZE)
#define tx_ring_occupancy ((vp->cur_tx - vp->dirty_tx) % GAMMA_TX_RING_SIZE)

#define tx_status() \
(unsigned long)(vp->tx_ring[vp->dirty_tx % GAMMA_TX_RING_SIZE].status)

#define tx_reset_and_restart() vp->dirty_tx = vp->cur_tx

#define tx_enq_pkt(header_ptr,header_size,payload_ptr,payload_size) \
for (;;) { /*  Allows easy escapes in fall-through programming style  */\
if (header_size+payload_size <= HEADER_SIZE+MAX_1ST_PIECE_SIZE) {\
	struct 	boom_tx_desc *prev_desc, *descriptor;\
	unsigned long entry;\
\
	prev_desc = &vp->tx_ring[(vp->cur_tx-1) % GAMMA_TX_RING_SIZE];\
	entry = vp->cur_tx % GAMMA_TX_RING_SIZE;\
\
	memcpy((unsigned char *)header_ptr + header_size,\
		bus_to_virt((unsigned long)payload_ptr),payload_size);\
\
	descriptor = &vp->tx_ring[entry];\
	descriptor->next = 0;\
	descriptor->status = cpu_to_le32(header_size+payload_size);\
	descriptor->frag[0].addr = \
		cpu_to_le32(\
pci_map_single(vp->pdev,header_ptr,header_size+payload_size,PCI_DMA_TODEVICE)\
		);\
	descriptor->frag[0].length = \
		cpu_to_le32((header_size+payload_size) | LAST_FRAG);\
\
	prev_desc->next = \
		cpu_to_le32(\
			vp->tx_ring_dma + entry * sizeof(struct boom_tx_desc)\
		);\
	outl(vp->tx_ring_dma + entry * sizeof(struct boom_tx_desc),\
		ioaddr + DownListPtr);\
	outw(DownUnstall, ioaddr + EL3_CMD);\
	break;\
}\
{\
	struct 	boom_tx_desc *prev_desc, *descriptor;\
	unsigned long entry;\
\
	prev_desc = &vp->tx_ring[(vp->cur_tx-1) % GAMMA_TX_RING_SIZE];\
	entry = vp->cur_tx % GAMMA_TX_RING_SIZE;\
\
	descriptor = &vp->tx_ring[entry];\
	descriptor->next = 0;\
	descriptor->status = cpu_to_le32(header_size+payload_size);\
	descriptor->frag[0].addr = \
		cpu_to_le32(\
pci_map_single(vp->pdev,header_ptr,header_size,PCI_DMA_TODEVICE)\
		);\
	descriptor->frag[0].length = cpu_to_le32(header_size);\
	descriptor->frag[1].addr = cpu_to_le32(payload_ptr);\
	descriptor->frag[1].length = cpu_to_le32((payload_size) | LAST_FRAG);\
\
	prev_desc->next = \
		cpu_to_le32(\
			vp->tx_ring_dma + entry * sizeof(struct boom_tx_desc)\
		);\
	outl(vp->tx_ring_dma + entry * sizeof(struct boom_tx_desc),\
		ioaddr + DownListPtr);\
	outw(DownUnstall, ioaddr + EL3_CMD);\
	break;\
}\
}

#define tx_enq_pkt1(pkt_ptr,pkt_size) \
{\
struct 	boom_tx_desc *prev_desc, *descriptor;\
unsigned long entry;\
\
prev_desc = &vp->tx_ring[(vp->cur_tx-1) % GAMMA_TX_RING_SIZE];\
entry = vp->cur_tx % GAMMA_TX_RING_SIZE;\
\
descriptor = &vp->tx_ring[entry];\
descriptor->next = 0;\
descriptor->status = cpu_to_le32(pkt_size);\
descriptor->frag[0].addr = \
	cpu_to_le32(\
		pci_map_single(vp->pdev,pkt_ptr,pkt_size,PCI_DMA_TODEVICE)\
	);\
descriptor->frag[0].length = cpu_to_le32((pkt_size) | LAST_FRAG);\
\
prev_desc->next = \
	cpu_to_le32(vp->tx_ring_dma + entry * sizeof(struct boom_tx_desc));\
outl(vp->tx_ring_dma + entry * sizeof(struct boom_tx_desc),\
	ioaddr + DownListPtr);\
outw(DownUnstall, ioaddr + EL3_CMD);\
}

#define tx_start()

#define tx_forw_tail() vp->cur_tx++

#define tx_flush_ring(tx_desc_indx,n,i) \
{\
unsigned long entry;\
i = 1;\
for (;;) {\
	time_586 a,b;\
	entry = vp->dirty_tx % GAMMA_TX_RING_SIZE;\
	gamma_time(a);\
	for (;;) {\
		if (\
			inl(ioaddr + DownListPtr) !=\
			vp->tx_ring_dma + entry * sizeof(struct boom_tx_desc)\
		) {\
			vp->dirty_tx++;\
			break;\
		}\
		barrier();\
		gamma_time(b);\
		if ((unsigned long)gamma_time_diff_ticks(b,a) > (n)*CLOCK) {\
			i = 0;\
			break;\
		}\
	}\
	if (!i) {\
printk("tx_flush_ring(): failure at pos %d, tx_status %8.8lx (%8.8lx, %8.8lx) (tail is at %d).\n",\
vp->dirty_tx % GAMMA_TX_RING_SIZE,\
tx_status(),(unsigned long)inw(ioaddr + EL3_STATUS),\
(unsigned long)inb(ioaddr + TxStatus),\
vp->cur_tx % GAMMA_TX_RING_SIZE);\
/* resetting the NIC... */\
		outb(0, ioaddr + TxStatus);\
		outw(TxReset, ioaddr + EL3_CMD);\
		for (i=100000;i;i--)\
			if (!(inw(ioaddr + EL3_STATUS) & CmdInProgress))\
				break;\
		outw(TxEnable, ioaddr + EL3_CMD);  /*  Re-enable Tx  */\
		i = 1;  /*  Ignoring failure  */\
		vp->dirty_tx++;\
	}\
	if (tx_ring_head == (tx_desc_indx)) break;\
}\
}

#define tx_flush_ring_1(tx_desc_indx,n,i) \
{\
unsigned long entry;\
i = 1;\
for (;;) {\
	time_586 a,b;\
	entry = vp->dirty_tx % GAMMA_TX_RING_SIZE;\
	gamma_time(a);\
	for (;;) {\
		if (\
			inl(ioaddr + DownListPtr) !=\
			vp->tx_ring_dma + entry * sizeof(struct boom_tx_desc)\
		) {\
			vp->dirty_tx++;\
			break;\
		}\
		barrier();\
		gamma_time(b);\
		if ((unsigned long)gamma_time_diff_ticks(b,a) > (n)*CLOCK) {\
			i = 0;\
			break;\
		}\
	}\
	if (!i) {\
printk("tx_flush_ring_1(): failure at pos %d, tx_status %8.8lx (%8.8lx, %8.8lx) (tail is at %d, frame_no is %d)\n",\
vp->dirty_tx % GAMMA_TX_RING_SIZE,\
tx_status(),(unsigned long)inw(ioaddr + EL3_STATUS),\
(unsigned long)inb(ioaddr + TxStatus),\
vp->cur_tx % GAMMA_TX_RING_SIZE,\
read_send_header_frameno((l2s *)bus_to_virt(vp->tx_ring[entry].frag[0].addr))\
);\
/* resetting the NIC... */\
		outb(0, ioaddr + TxStatus);\
		outw(TxReset, ioaddr + EL3_CMD);\
		for (i=100000;i;i--)\
			if (!(inw(ioaddr + EL3_STATUS) & CmdInProgress))\
				break;\
		outw(TxEnable, ioaddr + EL3_CMD);  /*  Re-enable Tx  */\
		i = 0;\
		vp->dirty_tx++;\
	}\
	if (tx_ring_head == (tx_desc_indx)) break;\
}\
}

#define tx_wait_ring_room(n,i) \
{\
unsigned long entry;\
int guard = GAMMA_TX_RING_SIZE; \
i = 1;\
for (;guard;guard--) {\
	time_586 a,b;\
	entry = vp->dirty_tx % GAMMA_TX_RING_SIZE;\
	gamma_time(a);\
	for (;;) {\
		if (\
			inl(ioaddr + DownListPtr) !=\
			vp->tx_ring_dma + entry * sizeof(struct boom_tx_desc)\
		) {\
			vp->dirty_tx++;\
			break;\
		}\
		barrier();\
		gamma_time(b);\
		if ((unsigned long)gamma_time_diff_ticks(b,a) > (n)*CLOCK) {\
			i = 0;\
			break;\
		}\
	}\
	if (!i) {\
printk("tx_wait_ring_room(): failure at pos %d, tx_status %8.8lx (%8.8lx, %8.8lx) (tail is at %d).\n",\
vp->dirty_tx % GAMMA_TX_RING_SIZE,\
tx_status(),(unsigned long)inw(ioaddr + EL3_STATUS),\
(unsigned long)inb(ioaddr + TxStatus),\
vp->cur_tx % GAMMA_TX_RING_SIZE);\
/* resetting the NIC... */\
		outb(0, ioaddr + TxStatus);\
		outw(TxReset, ioaddr + EL3_CMD);\
		for (i=100000;i;i--)\
			if (!(inw(ioaddr + EL3_STATUS) & CmdInProgress))\
				break;\
		outw(TxEnable, ioaddr + EL3_CMD);  /*  Re-enable Tx  */\
		i = 1;  /*  Ignoring failure  */\
		if (guard) \
			vp->dirty_tx++;\
	}\
	if (tx_ring_has_room()) break;\
}\
if (!guard) {\
printk("tx_wait_ring_room(): Tx ring blocked at pos %d (tail is at %d).  Bug in NIC fw?  Resetting...\n",vp->dirty_tx, vp->cur_tx);\
	vp->dirty_tx = vp->cur_tx;\
}\
}

#define tx_close()

/*======================================================================*/
/*  RX MACROS                                             	        */
/*======================================================================*/

#define rx_common_decl_and_init \
struct vortex_private *vp = (struct vortex_private *)gamma_device->priv;\
long ioaddr = gamma_device->base_addr;\
int entry;\
int rx_status

#define poll_common_decl_and_init \
struct vortex_private *vp = (struct vortex_private *)gamma_device->priv;\
long ioaddr = gamma_device->base_addr;\
int entry;\
int rx_status

#define rx_status() rx_status

#define rx_ring_tail (vp->cur_rx % GAMMA_RX_RING_SIZE)
#define rx_ring_head (vp->dirty_rx % GAMMA_RX_RING_SIZE)
#define read_rx_ring_tail(tail) tail = vp->cur_rx
#define write_rx_ring_tail(tail) vp->cur_rx = tail
#define forward_rx_ring_tail() vp->cur_rx++

#ifdef NO_IRQ_RECV
#define clear_rx_irqs()
#else
#define clear_rx_irqs() \
outw(AckIntr | UpComplete | IntReq | IntLatch, ioaddr + EL3_CMD);
#endif

#define rx_read_pkt_info() \
{\
entry = vp->cur_rx % GAMMA_RX_RING_SIZE;\
rx_status = le32_to_cpu(vp->rx_ring[entry].status);\
}

#define rx_err_generic() (rx_status & RxDError)
#define rx_err_pktnotavail() (!(rx_status & RxDComplete))

#define discard_pkt() \
{\
/*  Pop top Rx packet.  */\
entry = vp->dirty_rx % GAMMA_RX_RING_SIZE;\
vp->rx_ring[entry].status = 0;	/* Clear complete bit. */\
outw(UpUnstall, ioaddr + EL3_CMD);\
vp->dirty_rx++;\
vp->cur_rx++;\
}

#define rx_header(h) (h) = (l2s *)(bus_to_virt(vp->rx_ring[entry].addr))

#define invoke_standard_rx(dev) boomerang_rx(dev)

#define rx_suspend_nic() \
{\
unsigned long ioaddr = gamma_device->base_addr;\
/*Disable interrupts by clearing the interrupt mask.*/\
outw(SetIntrEnb | 0x0000, ioaddr + EL3_CMD);\
/*Ack pending IRQs*/\
outw(AckIntr | IntReq | IntLatch, ioaddr + EL3_CMD);\
gamma_irq_disabled = 1;\
}

#define rx_resume_nic() \
{\
struct vortex_private *vp = (struct vortex_private *)gamma_device->priv;\
unsigned long ioaddr = gamma_device->base_addr;\
/* Enable interrupts by setting the interrupt mask.*/\
outw(vp->intr_enable, ioaddr + EL3_CMD);\
gamma_irq_disabled = 0;\
}
