/***************************************************************************
    COPYRIGHT NOTE

    GAMMA:		Genoa Active Message MAchine
			BIND2 layer:
			Basic Interface for Network Device Drivers.
			Derived from the Linux network driver 3c59x.c

    Copyright		(C) 2003 Marco Maniezzo
    e-mail:		marco.maniezzo@polito.it

    This program  or include  file  is free software,  either  original  or
    derived from other  software made available  under the terms of the GNU
    General Public License. You can redistribute and/or modify this program
    or include  file under the  terms of the GNU  General Public License as
    published  by the  Free Software Foundation;  either version 2  of  the
    License, or (at your option) any later version.

    THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR   IMPLIED
    WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
    MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
    NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT,  INDIRECT,
    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
    NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
    USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
    ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
    THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

    You should have received a copy of the  GNU General Public License along
    with this program; if not, write  to the Free Software Foundation, Inc.,
    675 Mass Ave, Cambridge, MA 02139, USA.

    END OF COPYRIGHT NOTE
****************************************************************************/

#define MAX_FLUSH_WAIT  200000	/*  Max waiting (usec) for Tx ring to flush  */
#define GAMMA_TX_RING_SIZE 512	/*
				 *  Size of Tx descriptor ring.
				 */
#define GAMMA_RX_RING_SIZE 1024	/*
				 *  Size of Rx descriptor ring.
				 *  We need a really large ring size due to
				 *  the static credit-based flow control.
				 */
#ifdef GAMMA_FULL_DUPLEX
#define MAX_CREDIT_SIZE 28
#define LOW_WATERMARK 4		/*  Must be > 0  */
#else
#define MAX_CREDIT_SIZE 30
#define LOW_WATERMARK 2		/*  Must be > 0  */
#endif
				/*
				 *  The sum MAX_CREDIT_SIZE + LOW_WATERMARK
				 *  must NOT exceed the following quantity:
				 *  (GAMMA_RX_RING_SIZE-20)/(total_num_nodes-1)
				 */
#define FRAME_SIZ 1500		/*  Max payload chunk = 1476 bytes  */
#define STEPS_THEN_STOP 10

/*======================================================================*/
/*  TX MACROS                                                           */
/*======================================================================*/

#define tx_optimal_frag(l) FRAME_SIZ

#define tx_common_decl_and_init \
PUM_DEVICE_BLOCK pUmDevice = ( PUM_DEVICE_BLOCK )gamma_device->priv;\
PLM_DEVICE_BLOCK pDevice   = ( PLM_DEVICE_BLOCK )pUmDevice;\
int i\



#define tx_ring_head	  pDevice->SendConIdx

#define tx_ring_tail 	  pDevice->SendProdIdx	

#define tx_ring_occupancy ( pDevice->SendProdIdx - pDevice->SendConIdx ) % GAMMA_TX_RING_SIZE


#define tx_status() (unsigned long)0\



#define tx_reset_and_restart()\
{\
	LM_ResetAdapter(pDevice);\
	printk(KERN_WARNING "GAMMA: TX_RESET_AND_RESTART CALLED\n");\
}\







#define tx_enq_pkt(header_ptr,header_size,payload_ptr,payload_size)\
{\
long flags;\
PT3_SND_BD pSendBd;\
PT3_SND_BD pShadowSendBd;\
LM_UINT32 Value32;\
LM_UINT32 Idx;\
LM_PHYSICAL_ADDRESS tempAddress;\
\
flags = bcm5700_lock(pUmDevice);\
\
\
for (;;)\
{\
	if(pUmDevice->do_global_lock && pUmDevice->interrupt)\
	{\
		netif_stop_queue(gamma_device);\
		pUmDevice->tx_queued = 1;\
		if (!pUmDevice->interrupt)\
		{\
			netif_wake_queue(gamma_device);\
			pUmDevice->tx_queued = 0;\
		}\
		printk(KERN_WARNING "Enq_pkt error: NIC locked\n");\
		break;\
	}\
	\
\
	if(header_size+payload_size <= HEADER_SIZE+MAX_1ST_PIECE_SIZE)\
	{\
		if (atomic_read(&pDevice->SendBdLeft) < 1)\
		{\
			netif_stop_queue(gamma_device);\
			pUmDevice->tx_full = 1;\
    			if (atomic_read(&pDevice->SendBdLeft) >= 1) \
    			{\
    				netif_wake_queue(gamma_device);\
				pUmDevice->tx_full = 0;\
    			}\
    			printk(KERN_WARNING "enq_pkt() ERROR: no send bd left!!!\n");\
			break;\
		}\
\
		\
		\
		memcpy((unsigned char *)header_ptr + header_size,\
			       	bus_to_virt((unsigned long)payload_ptr),payload_size);\
		\
		\
		atomic_sub(1, &pDevice->SendBdLeft);\
		Idx = pDevice->SendProdIdx;\
		pSendBd = &pDevice->pSendBdVirt[Idx];\
	    	pShadowSendBd = &pDevice->ShadowSendBd[Idx];\
\
	    	pSendBd->HostAddr.High =  0;\
    		pShadowSendBd->HostAddr.High = 0;\
\
\
	    	bcm_set_addr(&tempAddress, pci_map_single(pUmDevice->pdev,\
				       	(unsigned char *)header_ptr, \
					header_size + payload_size, \
					PCI_DMA_TODEVICE));\
\
    		pSendBd->HostAddr.Low = tempAddress.Low;\
	    	Value32 = header_size + payload_size;\
\
    		Idx = (Idx + 1) & T3_SEND_RCB_ENTRY_COUNT_MASK;\
\
	    	pSendBd->u1.Len_Flags = (Value32 << 16) | SND_BD_FLAG_END;\
    		pShadowSendBd->u1.Len_Flags = (Value32 << 16) | SND_BD_FLAG_END;\
\
	    	REG_WR(pDevice, Mailbox.SendNicProdIdx[0].Low, Idx);\
\
    		if(T3_CHIP_REV(pDevice->ChipRevId) == T3_CHIP_REV_5700_BX)\
  			REG_WR(pDevice, Mailbox.SendNicProdIdx[0].Low, Idx);\
\
		pDevice->SendProdIdx = Idx;\
		break;\
	}\
	else\
	{\
		if (atomic_read(&pDevice->SendBdLeft) < 2)\
		{\
			netif_stop_queue(gamma_device);\
			pUmDevice->tx_full = 1;\
    			if (atomic_read(&pDevice->SendBdLeft) >= 2) \
    			{\
    				netif_wake_queue(gamma_device);\
				pUmDevice->tx_full = 0;\
    			}\
    			printk(KERN_WARNING "enq_pkt() ERROR: no send bd left!!!\n");\
			break;\
		}\
\
		atomic_sub(2, &pDevice->SendBdLeft);\
		Idx = pDevice->SendProdIdx;\
\
/*Enqueing header*/\
\
		pSendBd = &pDevice->pSendBdVirt[Idx];\
	    	pShadowSendBd = &pDevice->ShadowSendBd[Idx];\
\
	    	pSendBd->HostAddr.High =  0;\
    		pShadowSendBd->HostAddr.High = 0;\
\
\
	    	bcm_set_addr(&tempAddress, pci_map_single(pUmDevice->pdev,\
				       	(unsigned char *)header_ptr, \
					header_size, \
					PCI_DMA_TODEVICE));\
\
    		pSendBd->HostAddr.Low = tempAddress.Low;\
	    	Value32 = header_size;\
\
    		Idx = (Idx + 1) & T3_SEND_RCB_ENTRY_COUNT_MASK;\
\
	    	pSendBd->u1.Len_Flags = (Value32 << 16);\
    		pShadowSendBd->u1.Len_Flags = (Value32 << 16);\
\
\
/*Enqueing payload*/\
\
		if(Idx == 0)\
		{\
			pSendBd = &pDevice->pSendBdVirt[0];\
	    		pShadowSendBd = &pDevice->ShadowSendBd[0];\
		}\
		else\
		{\
			pSendBd++;\
	    		pShadowSendBd++;\
		}\
\
	    	pSendBd->HostAddr.High =  0;\
    		pShadowSendBd->HostAddr.High = 0;\
\
	    	bcm_set_addr(&tempAddress, pci_map_single(pUmDevice->pdev,\
				(unsigned char *)bus_to_virt((unsigned long)payload_ptr),\
			       	payload_size, \
				PCI_DMA_TODEVICE));\
\
    		pSendBd->HostAddr.Low = tempAddress.Low;\
	    	Value32 = payload_size;\
\
    		Idx = (Idx + 1) & T3_SEND_RCB_ENTRY_COUNT_MASK;\
\
	    	pSendBd->u1.Len_Flags = (Value32 << 16) | SND_BD_FLAG_END;\
    		pShadowSendBd->u1.Len_Flags = (Value32 << 16) | SND_BD_FLAG_END;\
\
	    	REG_WR(pDevice, Mailbox.SendNicProdIdx[0].Low, Idx);\
\
    		if(T3_CHIP_REV(pDevice->ChipRevId) == T3_CHIP_REV_5700_BX)\
  			REG_WR(pDevice, Mailbox.SendNicProdIdx[0].Low, Idx);\
\
		pDevice->SendProdIdx = Idx;\
		break;\
	}\
}\
bcm5700_unlock(pUmDevice, flags);\
}\







#define tx_enq_pkt1(pkt_ptr,pkt_size)\
{\
long flags;\
PT3_SND_BD pSendBd;\
PT3_SND_BD pShadowSendBd;\
LM_UINT32 Value32;\
LM_UINT32 Idx;\
LM_PHYSICAL_ADDRESS tempAddress;\
\
flags = bcm5700_lock(pUmDevice);\
\
for (;;)\
{\
	if (pUmDevice->do_global_lock && pUmDevice->interrupt)\
	{\
		netif_stop_queue(gamma_device);\
		pUmDevice->tx_queued = 1;\
		if (!pUmDevice->interrupt)\
		{\
			netif_wake_queue(gamma_device);\
			pUmDevice->tx_queued = 0;\
		}\
		printk(KERN_WARNING "enq_pkt1(): ERROR! NIC locked\n");\
		break;\
	}\
	\
	if (atomic_read(&pDevice->SendBdLeft) < 1)\
	{\
		netif_stop_queue(gamma_device);\
		pUmDevice->tx_full = 1;\
    		if (atomic_read(&pDevice->SendBdLeft) >= 1) \
    		{\
    			netif_wake_queue(gamma_device);\
			pUmDevice->tx_full = 0;\
    		}\
    		printk(KERN_WARNING "enq_pkt1(): ERROR: no send bd left\n");\
		break;\
	}\
\
\
	atomic_sub(1, &pDevice->SendBdLeft);\
	Idx = pDevice->SendProdIdx;\
\
	pSendBd = &pDevice->pSendBdVirt[Idx];\
    	pShadowSendBd = &pDevice->ShadowSendBd[Idx];\
\
    	pSendBd->HostAddr.High =  0;\
    	pShadowSendBd->HostAddr.High = 0;\
\
    	bcm_set_addr(&tempAddress, pci_map_single(pUmDevice->pdev, pkt_ptr, pkt_size, PCI_DMA_TODEVICE));\
    	pSendBd->HostAddr.Low = tempAddress.Low;\
    	Value32 = pkt_size;\
\
    	Idx = (Idx + 1) & T3_SEND_RCB_ENTRY_COUNT_MASK;\
\
    	pSendBd->u1.Len_Flags = (Value32 << 16) | SND_BD_FLAG_END;\
    	pShadowSendBd->u1.Len_Flags = (Value32 << 16) | SND_BD_FLAG_END;\
\
    	REG_WR(pDevice, Mailbox.SendNicProdIdx[0].Low, Idx);\
\
    	if(T3_CHIP_REV(pDevice->ChipRevId) == T3_CHIP_REV_5700_BX)\
  		REG_WR(pDevice, Mailbox.SendNicProdIdx[0].Low, Idx);\
\
	pDevice->SendProdIdx = Idx;\
	break;\
}\
bcm5700_unlock(pUmDevice, flags);\
}\


#define tx_start()

#define tx_forw_tail()





#define tx_flush_ring(tx_desc_indx,n,i)\
{\
time_586 a,b;\
\
i = 1;\
gamma_time(a);\
\
for(;;)\
{\
	if(pDevice->pStatusBlkVirt->Idx[0].SendConIdx != tx_ring_head)\
	{\
		pDevice->SendConIdx = (pDevice->SendConIdx + 1) & T3_SEND_RCB_ENTRY_COUNT_MASK;\
		atomic_add(1, &pDevice->SendBdLeft);\
	}\
	barrier();\
	gamma_time(b);\
\
	if ((unsigned long)gamma_time_diff_ticks(b,a) > (n)*CLOCK)\
	{\
		printk(KERN_WARNING "tx_flush_ring(): ERROR! Forcing break\n");\
		/*Ignoring Failure*/\
		break;\
	}\
	if(tx_ring_head == tx_desc_indx)\
		break;\
\
}\
}\


#define tx_flush_ring_1(tx_desc_indx,n,i)\
{\
time_586 a,b;\
\
i = 1;\
gamma_time(a);\
\
for(;;)\
{\
	if(pDevice->pStatusBlkVirt->Idx[0].SendConIdx != tx_ring_head)\
	{\
		pDevice->SendConIdx = (pDevice->SendConIdx + 1) & T3_SEND_RCB_ENTRY_COUNT_MASK;\
		atomic_add(1, &pDevice->SendBdLeft);\
	}\
	barrier();\
	gamma_time(b);\
\
	if ((unsigned long)gamma_time_diff_ticks(b,a) > (n)*CLOCK)\
	{\
		printk(KERN_WARNING "tx_flush_ring_1(): ERROR! Forcing break\n");\
		i = 0;\
		break;\
	}\
	if(tx_ring_head == tx_desc_indx)\
		break;\
\
}\
}\



#define tx_wait_ring_room(n,i)\
{\
time_586 a,b;\
i = 1;\
\
\
gamma_time(a);\
\
while(!tx_ring_has_room())\
{\
	if(pDevice->pStatusBlkVirt->Idx[0].SendConIdx != tx_ring_head)\
	{\
		pDevice->SendConIdx = (pDevice->SendConIdx + 1) & T3_SEND_RCB_ENTRY_COUNT_MASK;\
		atomic_add(1, &pDevice->SendBdLeft);\
	}\
	barrier();\
	gamma_time(b);\
\
	if ((unsigned long)gamma_time_diff_ticks(b,a) > (n)*CLOCK)\
	{\
		printk(KERN_WARNING "tx_wait_ring_room(): ERROR! head(%d), tail(%d)\n",\
			       tx_ring_head,\
			       tx_ring_tail);\
		break;\
	}\
}\
}\


#define tx_close()\
{\
/*printk(KERN_WARNING "tx_close() called\n");*/\
}\



/*======================================================================*/
/*  RX MACROS                                             	        */
/*======================================================================*/

#define rx_common_decl_and_init \
PUM_DEVICE_BLOCK pUmDevice = ( PUM_DEVICE_BLOCK )gamma_device->priv;\
PLM_DEVICE_BLOCK pDevice   = ( PLM_DEVICE_BLOCK )pUmDevice;\
PLM_PACKET pPacket = NULL;\
PT3_RCV_BD pRcvBd = NULL;\
PT3_RCV_BD pRcvBd2 = NULL;\
LM_UINT32 add;\
LM_UINT32 SwRcvRetConIdx;\
unsigned int rx_status = 0\

#define poll_common_decl_and_init \
PUM_DEVICE_BLOCK pUmDevice = ( PUM_DEVICE_BLOCK )gamma_device->priv;\
PLM_DEVICE_BLOCK pDevice   = ( PLM_DEVICE_BLOCK )pUmDevice\

#define rx_status() rx_status 

#define rx_ring_tail  pDevice->RcvRetConIdx

#define rx_ring_head  pDevice->pStatusBlkVirt->Idx[0].RcvProdIdx 

#define read_rx_ring_tail(tail) tail = pDevice->RcvRetConIdx

#define write_rx_ring_tail(tail) pDevice->RcvRetConIdx = tail

#define forward_rx_ring_tail() \
pDevice->RcvRetConIdx = (pDevice->RcvRetConIdx + 1) & T3_RCV_RETURN_RCB_ENTRY_COUNT_MASK




#ifdef NO_IRQ_RECV

#define clear_rx_irqs()
#else
#define clear_rx_irqs()
#endif




#define rx_read_pkt_info()



#define rx_header(h)\
{\
\
rx_status = 0;\
(h) = (l2s *)0;\
\
SwRcvRetConIdx = pDevice->RcvRetConIdx;\
\
pRcvBd = &pDevice->pRcvRetBdVirt[SwRcvRetConIdx];\
add = pRcvBd->HostAddr.Low;\
(h) = (l2s *)(bus_to_virt(add));\
}



#define invoke_standard_rx(dev)\
{\
pPacket = (PLM_PACKET) (MM_UINT_PTR(pDevice->pPacketDescBase) +  MM_UINT_PTR(pRcvBd->Opaque));\
\
if(!(pRcvBd->ErrorFlag && pRcvBd->ErrorFlag != RCV_BD_ERR_ODD_NIBBLED_RCVD_MII))\
{\
    pPacket->PacketStatus = LM_STATUS_SUCCESS;\
    pPacket->PacketSize = pRcvBd->Len - 4;\
    pPacket->Flags = pRcvBd->Flags;\
    if(pRcvBd->Flags & RCV_BD_FLAG_VLAN_TAG)\
    {\
    	pPacket->VlanTag = pRcvBd->VlanTag;\
    }\
    pPacket->u.Rx.TcpUdpChecksum = pRcvBd->TcpUdpCksum;\
}\
else\
{\
    pPacket->PacketStatus = LM_STATUS_FAILURE;\
    rx_status = 1;\
    printk(KERN_WARNING "invoke_standard_rx():Rx ERROR!\n");\
}\
SwRcvRetConIdx = (SwRcvRetConIdx + 1) & T3_RCV_RETURN_RCB_ENTRY_COUNT_MASK;\
pDevice->RcvRetConIdx = SwRcvRetConIdx;\
REG_WR(pDevice, Mailbox.RcvRetConIdx[0].Low, SwRcvRetConIdx);\
\
QQ_PushTail(&pDevice->RxPacketReceivedQ.Container, pPacket);\
\
if(QQ_GetEntryCnt(&pDevice->RxPacketReceivedQ.Container))\
	MM_IndicateRxPackets(pDevice);\
if(QQ_GetEntryCnt(&pDevice->RxPacketFreeQ.Container))\
	LM_QueueRxPackets(pDevice);\
if(QQ_GetEntryCnt(&pUmDevice->rx_out_of_buf_q.Container))\
	replenish_rx_buffers(pUmDevice);\
}\




#define discard_pkt()\
{\
SwRcvRetConIdx = (SwRcvRetConIdx + 1) & T3_RCV_RETURN_RCB_ENTRY_COUNT_MASK;\
pDevice->RcvRetConIdx = SwRcvRetConIdx;\
REG_WR(pDevice, Mailbox.RcvRetConIdx[0].Low, SwRcvRetConIdx);\
\
pRcvBd2 = &pDevice->pRxStdBdVirt[pDevice->RxStdProdIdx];\
pRcvBd2->Flags = RCV_BD_FLAG_END;\
pRcvBd2->Len = MAX_STD_RCV_BUFFER_SIZE;\
pRcvBd2->HostAddr.Low = pRcvBd->HostAddr.Low;\
pRcvBd2->HostAddr.High = pRcvBd->HostAddr.High;\
pRcvBd2->Opaque = pRcvBd->Opaque;\
pDevice->RxStdProdIdx = (pDevice->RxStdProdIdx + 1) & T3_STD_RCV_RCB_ENTRY_COUNT_MASK;\
REG_WR(pDevice, Mailbox.RcvStdProdIdx.Low, pDevice->RxStdProdIdx);\
}







#define rx_err_generic() rx_status 

#define rx_err_pktnotavail() (rx_ring_head == rx_ring_tail)




#define rx_suspend_nic()\
{\
PUM_DEVICE_BLOCK pUmDevice = ( PUM_DEVICE_BLOCK )gamma_device->priv;\
PLM_DEVICE_BLOCK pDevice   = ( PLM_DEVICE_BLOCK )pUmDevice;\
LM_DisableInterrupt(pDevice);\
gamma_irq_disabled = 1; \
}

#define rx_resume_nic()\
{\
PUM_DEVICE_BLOCK pUmDevice = ( PUM_DEVICE_BLOCK )gamma_device->priv;\
PLM_DEVICE_BLOCK pDevice   = ( PLM_DEVICE_BLOCK )pUmDevice;\
LM_EnableInterrupt(pDevice);\
gamma_irq_disabled = 0; \
}


