/***************************************************************************
    COPYRIGHT NOTE

    GAMMA:		Genoa Active Message MAchine
			BIND2 layer:
			Basic Interface for Network Device Drivers.
			Derived from the Linux network driver acenic.c

    Copyright		(C) 1999-2001 Luca Landi, Giuseppe Ciaccio,
			Giovanni Chiola
    e-mail:		{landi,ciaccio,chiola}@disi.unige.it

    This program  or include  file  is free software,  either  original  or
    derived from other  software made available  under the terms of the GNU
    General Public License. You can redistribute and/or modify this program
    or include  file under the  terms of the GNU  General Public License as
    published  by the  Free Software Foundation;  either version 2  of  the
    License, or (at your option) any later version.

    THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR   IMPLIED
    WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
    MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
    NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT,  INDIRECT,
    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
    NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
    USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
    ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
    THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

    You should have received a copy of the  GNU General Public License along
    with this program; if not, write  to the Free Software Foundation, Inc.,
    675 Mass Ave, Cambridge, MA 02139, USA.

    END OF COPYRIGHT NOTE
****************************************************************************/

#define MAX_FLUSH_WAIT  100000	/* Max waiting (usec) for a Tx pkt to flush */
#define GAMMA_TX_RING_SIZE TX_RING_ENTRIES
				/*
				 *  Size of Tx descriptor ring.
				 *  We need a really large ring size for
				 *  really non-blocking send()
				 *  DO NOT CHANGE!
				 */
#define USE_JUMBO_FRAMES

#ifdef USE_JUMBO_FRAMES
#define GAMMA_RX_RING_SIZE RX_JUMBO_RING_ENTRIES
#else
#define GAMMA_RX_RING_SIZE RX_STD_RING_ENTRIES
#endif
				/*
				 *  Size of Rx descriptor ring.
				 *  We need a really large ring size due to
				 *  the static credit-based flow control.
				 *  DO NOT CHANGE!
				 */

#define MAX_CREDIT_SIZE 20
#define LOW_WATERMARK 20	/*  Must be > 0  */
				/*
				 *  The sum MAX_CREDIT_SIZE + LOW_WATERMARK
				 *  must NOT exceed the following quantity:
				 *  (GAMMA_RX_RING_SIZE-20)/(total_num_nodes-1)
				 */

#ifndef USE_JUMBO_FRAMES
#define FRAME_SIZ 1500		/*  Max payload chunk = 1476 bytes  */
#else
#if 0
/*  BATM Titan 5 switch does not support > 6KByte frames!  */
#define FRAME_SIZ 8215		/*  Max payload chunk = 8191 bytes  */
#endif
#if 0
/*  AceNIC has trouble with FRAME_SIZ = 4120 (payload = 4096 bytes)  */
#define FRAME_SIZ 4120		/*  Max payload chunk = 4 KBytes  */
#endif
#define FRAME_SIZ 5144		/*  Max payload chunk = 5 KBytes  */
#endif

/*======================================================================*/
/*  TX MACROS                                                           */
/*======================================================================*/

#define tx_optimal_frag(l) FRAME_SIZ

#define tx_common_decl_and_init \
struct ace_private *ap = (struct ace_private *)gamma_device->priv; \
struct ace_regs *regs = ap->regs; \
int i = 1

#define tx_ring_head ap->tx_ret_csm
#define tx_ring_tail ap->tx_prd
#define tx_ring_occupancy \
((ap->tx_ret_csm <= ap->tx_prd) ? \
ap->tx_prd - ap->tx_ret_csm : \
TX_RING_ENTRIES - (ap->tx_ret_csm - ap->tx_prd))

#define tx_status() (unsigned long)0

#define tx_reset_and_restart() ap->tx_ret_csm = ap->tx_prd

#define tx_entry_2_payloadptr(i) \
bus_to_virt(readl(&(ap->tx_ring[(i)].addr.addrlo)))

#define tx_entry_len(i) (readl(&(ap->tx_ring[(i)].flagsize))) >> 16

#define tx_enq_pkt(header_ptr,header_size,payload_ptr,payload_size)  \
for (;;) { /*  Allows easy escapes in fall-through programming style  */ \
if (header_size+payload_size <= HEADER_SIZE+MAX_1ST_PIECE_SIZE) { \
	struct tx_desc *descriptor = &ap->tx_ring[ap->tx_prd]; \
	memcpy(&(((unsigned char *)header_ptr)[header_size]), \
		(void *)bus_to_virt((unsigned long)(payload_ptr)), \
		payload_size); \
	descriptor->addr.addrlo = virt_to_bus(header_ptr); \
	descriptor->flagsize = \
		((header_size+payload_size) << 16) | BD_FLG_END; \
	break; \
} \
{ \
	struct tx_desc *descriptor = &ap->tx_ring[ap->tx_prd]; \
	descriptor->addr.addrlo = virt_to_bus(header_ptr); \
	descriptor->flagsize = ((header_size) << 16) | BD_FLG_MORE; \
	ap->tx_prd = (ap->tx_prd+1) % TX_RING_ENTRIES; \
	descriptor = &ap->tx_ring[ap->tx_prd]; \
	descriptor->addr.addrlo = (unsigned long)(payload_ptr); \
	descriptor->flagsize = ((payload_size) << 16) | BD_FLG_END; \
	break; \
} \
}
	
#define tx_enq_pkt1(pkt_ptr,pkt_size) \
{ \
	struct tx_desc *descriptor = &ap->tx_ring[ap->tx_prd]; \
	descriptor->addr.addrlo = virt_to_bus(pkt_ptr); \
	descriptor->flagsize = ((pkt_size) << 16) | BD_FLG_END; \
}

#define tx_enq_pktdummy(header_ptr,header_size,payload_ptr,payload_size) \
{ \
	struct tx_desc *descriptor = &ap->tx_ring[ap->tx_prd]; \
	descriptor->addr.addrlo = virt_to_bus(header_ptr); \
	descriptor->flagsize = \
		((header_size+payload_size) << 16) | BD_FLG_END; \
}

#define tx_start() \
{ \
mb(); \
ap->tx_prd = (ap->tx_prd+1) % TX_RING_ENTRIES;\
writel(ap->tx_prd, &regs->TxPrd); \
mb(); \
}

#define tx_forw_tail()

#define tx_flush_ring(tx_desc_indx,n,i) \
{ \
i = 1; \
for (;;) { \
	time_586 a,b; \
	if ((tx_ring_head>=(tx_desc_indx)) && (tx_ring_tail >= tx_ring_head)) \
		break; \
	if ((tx_ring_head>=(tx_desc_indx)) && (tx_ring_tail < (tx_desc_indx))) \
		break; \
	if ((tx_ring_tail>=tx_ring_head) && (tx_ring_tail < (tx_desc_indx))) \
		break; \
	gamma_time(a); \
	for (;;) { \
		mb(); \
		if (*(ap->tx_csm) != ap->tx_ret_csm) {  \
			ap->tx_ret_csm = *(ap->tx_csm); \
			break; \
		} \
		barrier(); \
		gamma_time(b); \
		if ( \
gamma_time_diff_ticks(b,a) > (unsigned long long)(n)*CLOCK \
		) { \
			i = 0; \
			break; \
		} \
	} \
	if (!i) { \
printk("tx_flush_ring(): failure at pos %d (tail is at %d).  Bug in NIC fw?  Ignoring...\n",ap->tx_ret_csm, ap->tx_prd); \
		/*  Ignoring failure  */ \
		i = 1; \
		ap->tx_ret_csm = (ap->tx_ret_csm + 1) % GAMMA_TX_RING_SIZE; \
		break; \
	} \
} \
}

#define tx_wait_ring_room(n,i) \
{ \
int guard = GAMMA_TX_RING_SIZE; \
i = 1; \
for (;guard;guard--) { \
	time_586 a,b; \
	gamma_time(a); \
	for (;;) { \
		mb(); \
		if (*(ap->tx_csm) != ap->tx_ret_csm) {  \
			ap->tx_ret_csm = *(ap->tx_csm) ; \
			break; \
		} \
		barrier(); \
		gamma_time(b); \
		if ( \
gamma_time_diff_ticks(b,a) > (unsigned long long)(n)*CLOCK \
		) { \
			i = 0; \
			break; \
		} \
	} \
	if (!i) { \
printk("tx_wait_ring_room(): failure at pos %d (tail is at %d).  Bug in NIC fw?  Ignoring...\n",ap->tx_ret_csm, ap->tx_prd); \
		/*  Ignoring failure  */ \
		i = 1; \
		if (guard) \
/* ap->tx_ret_csm = (ap->tx_ret_csm + 1) % GAMMA_TX_RING_SIZE; */ \
			ap->tx_ret_csm = ap->tx_prd; \
	} \
	if (tx_ring_has_room()) break; \
} \
if (!guard) { \
printk("tx_wait_ring_room(): Tx ring blocked at pos %d (tail is at %d).  Bug in NIC fw?  Resetting...\n",ap->tx_ret_csm, ap->tx_prd); \
	ap->tx_ret_csm = ap->tx_prd; \
} \
}

#define tx_close()

/*======================================================================*/
/*  RX MACROS                                                           */
/*======================================================================*/

#define rx_common_decl_and_init \
struct ace_private *ap = (struct ace_private *)gamma_device->priv; \
int entry; \
void *ptrentry = NULL

#define poll_common_decl_and_init \
struct ace_private *ap = (struct ace_private *)gamma_device->priv; \
int entry; \
void *ptrentry = NULL

#define rx_ring_head (*(ap->rx_ret_prd) % RX_RETURN_RING_ENTRIES)
#define rx_ring_tail ap->cur_rx
#define read_rx_ring_tail(tail) tail = ap->cur_rx
#define write_rx_ring_tail(tail) ap->cur_rx = tail
#define forward_rx_ring_tail() \
ap->cur_rx = (ap->cur_rx + 1) % RX_RETURN_RING_ENTRIES

#define rx_status() 0

#define clear_rx_irqs() \
/*writel(0,&(((struct ace_private *)(gamma_device->priv))->regs->Mb0Lo))*/

#define rx_read_pkt_info() \
for (;;) { \
if (ap->cur_rx == *(ap->rx_ret_prd)) { \
	ptrentry = NULL; \
	break; \
} \
switch( \
ap->rx_return_ring[entry = ap->cur_rx].flags & (BD_FLG_JUMBO | BD_FLG_MINI) \
) { \
	case 0: \
	ptrentry = bus_to_virt((unsigned long) \
	get_aceaddr(&ap->rx_std_ring[ap->rx_return_ring[entry].idx].addr)); \
	break; \
	case BD_FLG_MINI: \
	ptrentry = bus_to_virt((unsigned long) \
	get_aceaddr(&ap->rx_mini_ring[ap->rx_return_ring[entry].idx].addr)); \
	break; \
	case BD_FLG_JUMBO: \
	ptrentry = bus_to_virt((unsigned long) \
	get_aceaddr(&ap->rx_jumbo_ring[ap->rx_return_ring[entry].idx].addr)); \
	break; \
} \
break; \
}

#define rx_err_generic() 0

#define rx_err_pktnotavail()  (ptrentry == NULL)

#define discard_pkt() \
{ \
switch(ap->rx_return_ring[ap->cur_rx].flags & (BD_FLG_JUMBO | BD_FLG_MINI)) { \
	case 0: \
		atomic_dec(&ap->cur_rx_bufs); \
		break; \
	case BD_FLG_MINI: \
		atomic_dec(&ap->cur_mini_bufs); \
		break; \
	case BD_FLG_JUMBO: \
		atomic_dec(&ap->cur_jumbo_bufs); \
		break; \
} \
ap->cur_rx = (ap->cur_rx + 1) % RX_RETURN_RING_ENTRIES; \
}

#define rx_header(h) (h) = (l2s *)(ptrentry)

#define invoke_standard_rx(dev) ace_rx_int(dev, *(ap->rx_ret_prd), ap->cur_rx)

#define rx_suspend_nic() \
{ \
struct ace_private *ap = (struct ace_private *)gamma_device->priv; \
struct ace_regs *regs = ap->regs; \
gamma_irq_disabled = 1; \
/*writel(readl(&regs->HostCtrl)|MASK_INTS, &regs->HostCtrl);*/ \
writel(0, &regs->TuneStatTicks); \
/*init_timer(&ace_irq_timer); \
ace_irq_timer.expires = jiffies + (HZ >> 4); \
ace_irq_timer.function = ace_irq_timer_timeout; \
add_timer(&ace_irq_timer);*/ \
}

#define rx_resume_nic() \
{ \
struct ace_private *ap = (struct ace_private *)gamma_device->priv; \
struct ace_regs *regs = ap->regs; \
gamma_irq_disabled = 0; \
/*writel(readl(&regs->HostCtrl)&(~MASK_INTS), &regs->HostCtrl);*/ \
writel(DEF_STAT, &regs->TuneStatTicks); \
/*del_timer(&ace_irq_timer);*/ \
}
