/***************************************************************************
    COPYRIGHT NOTE

    GAMMA:		Genoa Active Message MAchine
			BIND2 layer:
			Basic Interface for Network Device Drivers.
			Derived from the Linux network driver acenic.c

    Copyright		(C) 1998-2001 Guido A. Guerreschi, Giuseppe Ciaccio,
			Luigi V. Mancini
    e-mail:		ciaccio@disi.unige.it, lv.mancini@dsi.uniroma1.it

    This program  or include  file  is free software,  either  original  or
    derived from other  software made available  under the terms of the GNU
    General Public License. You can redistribute and/or modify this program
    or include  file under the  terms of the GNU  General Public License as
    published  by the  Free Software Foundation;  either version 2  of  the
    License, or (at your option) any later version.

    THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR   IMPLIED
    WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
    MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
    NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT,  INDIRECT,
    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
    NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
    USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
    ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
    THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

    You should have received a copy of the  GNU General Public License along
    with this program; if not, write  to the Free Software Foundation, Inc.,
    675 Mass Ave, Cambridge, MA 02139, USA.

    END OF COPYRIGHT NOTE
****************************************************************************/

#define MAX_FLUSH_WAIT  100000	/* Max waiting (usec) for a Tx pkt to flush */
#define GAMMA_TX_RING_SIZE 128	/*
				 *  Size of Tx descriptor ring.
				 */
#define GAMMA_RX_RING_SIZE 1024	/*
				 *  Size of Rx descriptor ring.
				 *  We need a really large ring size due to
				 *  the static credit-based flow control.
				 */
#ifdef GAMMA_FULL_DUPLEX
#define MAX_CREDIT_SIZE 28 
#define LOW_WATERMARK 4		/*  Must be > 0  */
#else
#define MAX_CREDIT_SIZE 30
#define LOW_WATERMARK 2		/*  Must be > 0  */
#endif
				/*
				 *  The sum MAX_CREDIT_SIZE + LOW_WATERMARK
				 *  must NOT exceed the following quantity:
				 *  (GAMMA_RX_RING_SIZE-20)/(total_num_nodes-1)
				 */
#define FRAME_SIZ 1500		/*  Max payload chunk = 1476 bytes  */

/*======================================================================*/
/*  TX MACROS                                                           */
/*======================================================================*/

#define tx_optimal_frag(l) FRAME_SIZ

#define tx_common_decl_and_init \
struct speedo_private *sp = (struct speedo_private *)gamma_device->priv; \
unsigned long ioaddr = gamma_device->base_addr; \
int i

#define tx_ring_head (sp->dirty_tx % GAMMA_TX_RING_SIZE)
#define tx_ring_tail (sp->cur_tx % GAMMA_TX_RING_SIZE)
#define tx_ring_occupancy ((sp->cur_tx - sp->dirty_tx) % GAMMA_TX_RING_SIZE)

#define tx_status() \
(unsigned long)(sp->tx_ring[sp->dirty_tx % GAMMA_TX_RING_SIZE].status)

#define tx_reset_and_restart() sp->dirty_tx = sp->cur_tx

#define tx_enq_pkt(header_ptr,header_size,payload_ptr,payload_size) \
for (;;) { /*  Allows easy escapes in fall-through programming style  */ \
if (header_size+payload_size <= HEADER_SIZE+MAX_1ST_PIECE_SIZE) { \
struct TxFD *descriptor; \
memcpy((unsigned char *)header_ptr + header_size, \
	bus_to_virt((unsigned long)payload_ptr),payload_size); \
descriptor = &sp->tx_ring[sp->cur_tx % GAMMA_TX_RING_SIZE]; \
descriptor->status = cpu_to_le32(CmdSuspend | CmdTx | CmdTxFlex); \
descriptor->link = \
cpu_to_le32(TX_RING_ELEM_DMA(sp, (sp->cur_tx+1) % GAMMA_TX_RING_SIZE)); \
descriptor->tx_desc_addr = \
cpu_to_le32(TX_RING_ELEM_DMA(sp, sp->cur_tx % GAMMA_TX_RING_SIZE) \
+ TX_DESCR_BUF_OFFSET); \
descriptor->count = 0x01208000; \
descriptor->tx_buf_addr0 = \
	cpu_to_le32( \
pci_map_single(sp->pdev,header_ptr,header_size+payload_size,PCI_DMA_TODEVICE) \
	); \
descriptor->tx_buf_size0 = cpu_to_le32(header_size+payload_size); \
break; \
} \
{ \
struct TxFD *descriptor = &sp->tx_ring[sp->cur_tx % GAMMA_TX_RING_SIZE]; \
descriptor->status = cpu_to_le32(CmdSuspend | CmdTx | CmdTxFlex); \
descriptor->link = \
cpu_to_le32(TX_RING_ELEM_DMA(sp, (sp->cur_tx+1) % GAMMA_TX_RING_SIZE)); \
descriptor->tx_desc_addr = \
cpu_to_le32(TX_RING_ELEM_DMA(sp, sp->cur_tx % GAMMA_TX_RING_SIZE) \
+ TX_DESCR_BUF_OFFSET); \
/*descriptor->count = 0x01208000;*/ \
descriptor->count = 0x02208000; \
descriptor->tx_buf_addr0 = \
	cpu_to_le32( \
pci_map_single(sp->pdev,header_ptr,header_size,PCI_DMA_TODEVICE) \
	); \
descriptor->tx_buf_size0 = cpu_to_le32(header_size); \
descriptor->tx_buf_addr1 = cpu_to_le32(payload_ptr); \
descriptor->tx_buf_size1 = cpu_to_le32(payload_size); \
break; \
} \
}

#define tx_enq_pkt1(pkt_ptr,pkt_size) \
{ \
struct TxFD *descriptor = &sp->tx_ring[sp->cur_tx % GAMMA_TX_RING_SIZE]; \
descriptor->status = cpu_to_le32(CmdSuspend | CmdTx | CmdTxFlex); \
descriptor->link = \
cpu_to_le32(TX_RING_ELEM_DMA(sp, (sp->cur_tx+1) % GAMMA_TX_RING_SIZE)); \
descriptor->tx_desc_addr = \
cpu_to_le32(TX_RING_ELEM_DMA(sp, sp->cur_tx % GAMMA_TX_RING_SIZE) \
+ TX_DESCR_BUF_OFFSET); \
descriptor->count = 0x01208000; \
descriptor->tx_buf_addr0 = \
	cpu_to_le32( \
		pci_map_single(sp->pdev,pkt_ptr,pkt_size,PCI_DMA_TODEVICE) \
	); \
descriptor->tx_buf_size0 = cpu_to_le32(pkt_size); \
}

#define tx_start() \
{ \
wait_for_cmd_done(ioaddr + SCBCmd); \
clear_suspend(sp->last_cmd); \
outb(CUResume, ioaddr + SCBCmd); \
sp->last_cmd = \
	(struct descriptor *)&sp->tx_ring[sp->cur_tx % GAMMA_TX_RING_SIZE]; \
}

#define tx_forw_tail() sp->cur_tx++

#define tx_flush_ring(tx_desc_indx,n,i) \
{ \
i = 1; \
for (;;) { \
	time_586 a,b; \
	unsigned long status; \
	gamma_time(a); \
	for (;;) { \
		status = \
			le32_to_cpu( \
			sp->tx_ring[sp->dirty_tx % GAMMA_TX_RING_SIZE].status \
			); \
		if (status & StatusComplete) { \
			sp->dirty_tx++; \
			break; \
		} \
		barrier(); \
		gamma_time(b); \
		if ((unsigned long)gamma_time_diff_ticks(b,a) > (n)*CLOCK) { \
			i = 0; \
			break; \
		} \
	} \
	if (!i) { \
printk("tx_flush_ring(): failure at pos %d, tx_status %8.8lx (tail is at %d).\n",sp->dirty_tx % GAMMA_TX_RING_SIZE, status, sp->cur_tx % GAMMA_TX_RING_SIZE); \
		outl( \
			TX_RING_ELEM_DMA(sp, sp->dirty_tx % TX_RING_SIZE), \
			ioaddr + SCBPointer \
		); \
		outw(CUStart, ioaddr + SCBCmd); \
		reset_mii(gamma_device); \
		i = 1;  /*  Ignoring failure  */ \
		sp->dirty_tx++; \
	} \
	if (tx_ring_head == (tx_desc_indx)) break; \
} \
}

#define tx_flush_ring_1(tx_desc_indx,n,i) \
{ \
unsigned long entry; \
i = 1; \
for (;;) { \
	time_586 a,b; \
	unsigned long status; \
	entry = sp->dirty_tx % GAMMA_TX_RING_SIZE; \
	gamma_time(a); \
	for (;;) { \
		status = le32_to_cpu(sp->tx_ring[entry].status); \
		if (status & StatusComplete) { \
			sp->dirty_tx++; \
			break; \
		} \
		barrier(); \
		gamma_time(b); \
		if ((unsigned long)gamma_time_diff_ticks(b,a) > (n)*CLOCK) { \
			i = 0; \
			break; \
		} \
	} \
	if (!i) { \
printk("tx_flush_ring_1(): failure at pos %d, tx_status %8.8lx (tail is at %d, frame_no is %d).\n",sp->dirty_tx % GAMMA_TX_RING_SIZE, status, sp->cur_tx % GAMMA_TX_RING_SIZE, \
read_send_header_frameno((l2s *)bus_to_virt(sp->tx_ring[entry].tx_buf_addr0)) \
); \
		outl( \
			TX_RING_ELEM_DMA(sp, sp->dirty_tx % TX_RING_SIZE), \
			ioaddr + SCBPointer \
		); \
		outw(CUStart, ioaddr + SCBCmd); \
		reset_mii(gamma_device); \
		sp->dirty_tx++; \
	} \
	if (tx_ring_head == (tx_desc_indx)) break; \
} \
}

#define tx_wait_ring_room(n,i) \
{ \
int guard = GAMMA_TX_RING_SIZE; \
i = 1; \
for (;guard;guard--) { \
	time_586 a,b; \
	unsigned long status; \
	gamma_time(a); \
	for (;;) { \
		status = \
			le32_to_cpu( \
			sp->tx_ring[sp->dirty_tx % GAMMA_TX_RING_SIZE].status \
			); \
		if (status & StatusComplete) { \
			sp->dirty_tx++; \
			break; \
		} \
		barrier(); \
		gamma_time(b); \
		if ((unsigned long)gamma_time_diff_ticks(b,a) > (n)*CLOCK) { \
			i = 0; \
			break; \
		} \
	} \
	if (!i) { \
printk("tx_wait_ring_room(): failure at pos %d, tx_status %8.8lx (tail is at %d).\n",sp->dirty_tx % GAMMA_TX_RING_SIZE, status, sp->cur_tx % GAMMA_TX_RING_SIZE); \
		outl( \
			TX_RING_ELEM_DMA(sp, sp->dirty_tx % TX_RING_SIZE), \
			ioaddr + SCBPointer \
		); \
		outw(CUStart, ioaddr + SCBCmd); \
		reset_mii(gamma_device); \
		i = 1; /*  Ignoring failure  */ \
		if (guard) \
			sp->dirty_tx++; \
	} \
	if (tx_ring_has_room()) break; \
} \
if (!guard) { \
printk("tx_wait_ring_room(): Tx ring blocked at pos %d (tail is at %d).  Bug in NIC fw?  Resetting...\n",sp->dirty_tx, sp->cur_tx); \
	sp->dirty_tx = sp->cur_tx; \
} \
}

#define tx_close()

/*======================================================================*/
/*  RX MACROS                                                           */
/*======================================================================*/

#define rx_common_decl_and_init \
struct speedo_private *sp = (struct speedo_private *)gamma_device->priv; \
int entry; \
int rx_status

#define poll_common_decl_and_init \
struct speedo_private *sp = (struct speedo_private *)gamma_device->priv; \
unsigned long ioaddr = gamma_device->base_addr; \
int entry; \
int rx_status

#define rx_status() rx_status

#define rx_ring_head (sp->dirty_rx % GAMMA_RX_RING_SIZE)
#define rx_ring_tail (sp->cur_rx % GAMMA_RX_RING_SIZE)
#define read_rx_ring_tail(tail) tail = sp->cur_rx
#define write_rx_ring_tail(tail) sp->cur_rx = tail
#define forward_rx_ring_tail() sp->cur_rx++

#ifdef NO_IRQ_RECV
#define clear_rx_irqs()
#else
#define clear_rx_irqs() \
outw(inw(ioaddr + SCBStatus) & 0xfc00, ioaddr + SCBStatus)
#endif

#define rx_read_pkt_info() \
{ \
entry = sp->cur_rx % GAMMA_RX_RING_SIZE; \
rx_status = le32_to_cpu(sp->rx_ringp[entry]->status); \
}

#define rx_err_generic() (!(rx_status & RxOK))
#define rx_err_pktnotavail() (!(rx_status & RxComplete))

#define discard_pkt() \
{ \
struct RxFD *rxf = sp->rx_ringp[entry]; \
dma_addr_t rxf_dma = sp->rx_ring_dma[entry]; \
rxf->status = cpu_to_le32(0xC0000001); \
rxf->link = 0; \
rxf->count = cpu_to_le32(PKT_BUF_SZ << 16); \
sp->last_rxf->link = cpu_to_le32(rxf_dma); \
sp->last_rxf->status &= cpu_to_le32(~0xC0000000); \
sp->last_rxf = rxf; \
sp->cur_rx++; \
sp->dirty_rx++; \
}

#define rx_header(h) \
(h) = (l2s *)(((unsigned char *)(sp->rx_ringp[entry]))+sizeof(struct RxFD))

#define invoke_standard_rx(dev) speedo_rx(dev)

#define rx_suspend_nic() \
{ \
unsigned long ioaddr = gamma_device->base_addr; \
outw(SCBMaskAll, ioaddr + SCBCmd); \
/*  Ack all pending IRQs.  */ \
outw(inw(ioaddr + SCBStatus) & 0xfc00, ioaddr + SCBStatus); \
gamma_irq_disabled = 1; \
}

#define rx_resume_nic() \
{ \
unsigned long ioaddr = gamma_device->base_addr; \
outw(SCBMaskEarlyRx | SCBMaskFlowCtl, ioaddr + SCBCmd); \
gamma_irq_disabled = 0; \
}
