/***************************************************************************
    COPYRIGHT NOTE

    GAMMA:              Genoa Active Message MAchine
                        BIND2 layer:
                        Basic Interface for Network Device Drivers.
                        Derived from the Linux network drivers ``ga621'' and
                        ``ga622t''

    Copyright           (C) 2002 Marco Ehlert, Giuseppe Ciaccio
    e-mail:             mehlert@cs.uni-potsdam.de, ciaccio@disi.unige.it

    This program  or include  file  is free software,  either  original  or
    derived from other  software made available  under the terms of the GNU
    General Public License. You can redistribute and/or modify this program
    or include  file under the  terms of the GNU  General Public License as
    published  by the  Free Software Foundation;  either version 2  of  the
    License, or (at your option) any later version.

    THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR   IMPLIED
    WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
    MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
    NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT,  INDIRECT,
    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
    NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
    USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
    ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
    THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

    You should have received a copy of the  GNU General Public License along
    with this program; if not, write  to the Free Software Foundation, Inc.,
    675 Mass Ave, Cambridge, MA 02139, USA.

    END OF COPYRIGHT NOTE
****************************************************************************/

#define USE_JUMBO_FRAMES

#define MAX_FLUSH_WAIT  100000	/* Max waiting (usec) for a Tx pkt to flush */
#define GAMMA_TX_RING_SIZE 200
				/*
				 *  Size of Tx descriptor ring.
				 *  We need a really large ring size for
				 *  really non-blocking send()
				 *  DO NOT CHANGE!
				 */
#define GAMMA_RX_RING_SIZE 768
				/*
				 *  Size of Rx descriptor ring.
				 *  We need a really large ring size due to
				 *  the static credit-based flow control.
				 *  DO NOT CHANGE!
				 */



#define MAX_CREDIT_SIZE 20
#define LOW_WATERMARK 20	/*  Must be > 0  */
				/*
				 *  The sum MAX_CREDIT_SIZE + LOW_WATERMARK
				 *  must NOT exceed the following quantity:
				 *  (GAMMA_RX_RING_SIZE-20)/(total_num_nodes-1)
				 */

#ifdef USE_JUMBO_FRAMES
#define FRAME_SIZ 4120		/*
				 *  The GA621 seems to only work if this is
				 *  less than TxFIFO size (8192).
				 *  Max payload chunk = 4 KBytes (it could be
				 *  greater, but with no signific perf improve)
                                 */
#else
#define FRAME_SIZ 1500
#endif


/*======================================================================*/
/*  TX MACROS                                                           */
/*======================================================================*/

#define tx_debug \
{\
HsmTCB    	*pTcb=NULL;\
unsigned long   start_i;\
unsigned long   end_i;\
unsigned long   txdp;\
int 		ii;\
UINT		Pcr_val; \
UINT            TXDP_val; \
\
NsmRegRead32(pAdapter->pNsmContext, (pAdapter->RegAddr + TXDP), &TXDP_val); \
txdp=((unsigned long)TXDP_val - (unsigned long)(pAdapter->PriQue[0].pTxdArrayPa)) >> 4;\
printk("TXDP=%lu\n",txdp);\
\
pTcb=pAdapter->PriQue[0].pTcbArray;\
printk("tx_ring_head=%u tx_ring_tail=%u\n",tx_ring_head,tx_ring_tail);\
\
/* Checking for an active timing Pause Frame */\
NsmRegRead32(pAdapter->pNsmContext, (pAdapter->RegAddr + PCR), &Pcr_val);\
if(Pcr_val & PS_ACT )\
        printk("tx_debug(): Pause frame actively timing 0x%x\n", Pcr_val );\
\
   for (ii=0; ii<GAMMA_TX_RING_SIZE; ii++) {\
	int            more     = 9;\
	int            own      = 9;\
	int            intr     = 9;\
	unsigned long  link     = 999;\
\
	if (pTcb->pStartDesc){\
		start_i = ((unsigned long)pTcb->pStartDesc - (unsigned long)(pAdapter->PriQue[0].pTxdArray)) >> 4;\
	}else{\
		start_i = 999;\
	}\
\
	if (pTcb->pEndDesc){\
                end_i     = ((unsigned long)pTcb->pEndDesc - (unsigned long)(pAdapter->PriQue[0].pTxdArray)) >> 4;\
		more      = (pTcb->pEndDesc->CmdSts & MORE) ? 1 : 0;\
		own       = (pTcb->pEndDesc->CmdSts & OWN ) ? 1 : 0;\
		intr      = (pTcb->pEndDesc->CmdSts & INTR) ? 1 : 0;\
		link      = ((unsigned long)pTcb->pEndDesc->link - (unsigned long)(pAdapter->PriQue[0].pTxdArrayPa)) >> 4;\
        }else{\
                end_i = 999;\
        }\
	printk("tcb=%d s=%lu e=%lu, M=%1d, O=%1d, I=%1d L=%lu\n",ii,start_i,end_i,more,own,intr,link);\
\
	pTcb = pTcb->pNext;\
    }/* end for */\
}

#define tx_optimal_frag(l) FRAME_SIZ

#define tx_common_decl_and_init \
AdapterContext  *pAdapter = &((NsmContext*)gamma_device->priv)->AdpCtxt; \
int i = 1

/* divide by 16 because sizeof(DevDesc)==16 and descriptor array is page aligned */
#define tx_ring_head ( ((unsigned int)(pAdapter->PriQue[0].pTcbListHead->pStartDesc)- \
                        (unsigned int)(pAdapter->PriQue[0].pTxdArray)) >> 4 )

#define tx_ring_tail ( ((unsigned int)(pAdapter->PriQue[0].pTcbListNext->pStartDesc)- \
                        (unsigned int)(pAdapter->PriQue[0].pTxdArray)) >> 4 )

/* differences give correct values because head and tail are similarly scaled down unsigned ints */
#define tx_ring_occupancy (tx_ring_head <= tx_ring_tail ? \
                          (tx_ring_tail - tx_ring_head) : \
    GAMMA_TX_RING_SIZE - ((tx_ring_head - tx_ring_tail)  ))

#define tx_status() (unsigned long)0

#define tx_reset_and_restart() \
{\
/* only GAMMA packets are reset, that is why no deallocation is needed */ \
/* reset */\
	UCHAR stat;\
	printk("GAMMA: tx_reset_and_restart()\n");\
\
        if( ( stat =HsmClose(pAdapter) ) == FAILURE)\
		printk ("tx_reset_and_restart() HsmClose() failed \n");\
\
        stat = HsmReset(pAdapter, 0);\
        if(stat == SUCCESS)\
                printk("tx_reset_and_restart() HsmReset() complete\n");\
        else\
                printk("tx_reset_and_restart() HsmReset() failed \n");\
\
	if( ( stat =HsmOpen(pAdapter) ) == FAILURE)\
                printk ("tx_reset_and_restart() HsmOpen() failed \n");\
\
	HsmRxFilter(pAdapter, (ACCEPT_PERFECT_MATCH_ON | ACCEPT_ALL_BROADCAST_ON));\
\
	tx_start();\
\
	if ( !(pAdapter->interruptStatus & RXRCMP) )\
                HsmRxStart(pAdapter, 0);\
}

#define tx_entry_2_payloadptr(i)
//bus_to_virt(readl(&(ap->tx_ring[(i)].addr.addrlo)))

//#define tx_entry_len(i)

#define tx_enq_pkt(header_ptr,header_size,payload_ptr,payload_size)\
{\
        HsmTCB                  *pCurrTcb  =NULL; \
\
        volatile DevDesc        *currDesc  =NULL; \
        volatile DevDesc        *currDescPa=NULL; \
	volatile DevDesc        *descBase  =NULL;\
        volatile DevDesc        *descBasePa=NULL;\
\
/*printk(KERN_EMERG "tx_enq_pkt tx_head=%d tx_tail=%d\n",\
((unsigned int)tx_ring_head - (unsigned int)pAdapter->PriQue[0].pTxdArray) / sizeof(DevDesc),\
((unsigned int)tx_ring_tail - (unsigned int)pAdapter->PriQue[0].pTxdArray) / sizeof(DevDesc) );*/\
\
/*printk(KERN_EMERG "tx_enq_pkt header_size=%ld        payload_size=%ld\n",(long int)header_size,payload_size);*/ \
/*printk(KERN_EMERG "tx_enq_pkt HEADER_SIZE=%d MAX_1ST_PIECE_SIZE=%d\n",HEADER_SIZE,MAX_1ST_PIECE_SIZE);*/\
\
	if ( header_size + payload_size <= HEADER_SIZE+MAX_1ST_PIECE_SIZE) { \
\
		/* copy data fater header */\
		memcpy(&(((unsigned char *)header_ptr)[header_size]), \
			(void *)bus_to_virt((unsigned long)(payload_ptr)), \
							payload_size); \
\
	        pCurrTcb   = GET_TCB_NEXT(pAdapter, 0); \
\
	        currDesc   = pCurrTcb->pStartDesc; \
        	currDescPa = pCurrTcb->pStartDescPa; \
\
	        CLEAN_UP_DESC(currDesc); \
\
        	TX_DESC_SET_BUF(  currDesc, virt_to_bus(header_ptr) ); \
        	TX_DESC_SET_SIZE( currDesc, header_size+payload_size ); \
\
        	pCurrTcb->pEndDesc   = currDesc; \
        	pCurrTcb->pEndDescPa = currDescPa; \
\
        	TX_DESC_SET_OWN(currDesc); \
\
	}else{\
/* header */\
        	pCurrTcb   = GET_TCB_NEXT(pAdapter, 0); \
\
	        currDesc   = pCurrTcb->pStartDesc; \
        	currDescPa = pCurrTcb->pStartDescPa; \
\
        	CLEAN_UP_DESC(currDesc); \
\
        	TX_DESC_SET_BUF(  currDesc, virt_to_bus(header_ptr) ); \
        	TX_DESC_SET_SIZE( currDesc, header_size); \
\
        	TX_DESC_SET_OWN(  currDesc); \
		TX_DESC_SET_MORE( currDesc, 1); \
\
/* forward DevDesc */\
        	descBase   = pAdapter->PriQue[0].pTxdArray;\
        	descBasePa = pAdapter->PriQue[0].pTxdArrayPa;\
\
        	currDesc   = (DevDesc *)INC_CURR_DESC(pAdapter, 0, currDesc,  1, descBase);\
        	currDescPa = (DevDesc *)INC_CURR_DESC(pAdapter, 0, currDescPa,1, descBasePa);\
\
/*        	pCurrTcb               = pCurrTcb->pNext;\
        	pCurrTcb->pStartDesc   = currDesc;\
        	pCurrTcb->pStartDescPa = currDescPa;\
\
        	SET_TCB_NEXT(pAdapter, 0, pCurrTcb); */\
\
/* data */\
\
        	CLEAN_UP_DESC(currDesc); \
\
	        TX_DESC_SET_BUF(  currDesc, payload_ptr ); \
        	TX_DESC_SET_SIZE( currDesc, payload_size ); \
\
        	pCurrTcb->pEndDesc   = currDesc; \
        	pCurrTcb->pEndDescPa = currDescPa; \
\
        	TX_DESC_SET_OWN(currDesc); \
	} \
/*	printk(KERN_EMERG "tx_enq_pkt end \n");*/ \
}
	
#define tx_enq_pkt1(pkt_ptr,pkt_size) \
{ \
	HsmTCB          	*pCurrTcb  =NULL; \
\
        volatile DevDesc        *currDesc  =NULL; \
        volatile DevDesc        *currDescPa=NULL; \
\
/* GAMMA	NsmAquireLock(pAdapter->pNsmContext, pAdapter->TxLock); */ \
\
        pCurrTcb   = GET_TCB_NEXT(pAdapter, 0); \
\
        currDesc   = pCurrTcb->pStartDesc; \
        currDescPa = pCurrTcb->pStartDescPa; \
\
	CLEAN_UP_DESC(currDesc); \
\
        TX_DESC_SET_BUF(  currDesc, virt_to_bus(pkt_ptr) ); \
        TX_DESC_SET_SIZE( currDesc, pkt_size ); \
\
        pCurrTcb->pEndDesc   = currDesc; \
        pCurrTcb->pEndDescPa = currDescPa; \
\
        TX_DESC_SET_OWN(currDesc); \
\
/* GAMMA	NsmReleaseLock(pAdapter->pNsmContext, pAdapter->TxLock); */ \
}

/*
#define tx_enq_pktdummy(header_ptr,header_size,payload_ptr,payload_size) \
{ \
	struct tx_desc *descriptor = &ap->tx_ring[ap->tx_prd]; \
	writel(virt_to_bus(header_ptr),&descriptor->addr.addrlo); \
	writel(((header_size+payload_size) << 16) | (BD_FLG_END), \
		&descriptor->flagsize); \
}
*/

#define tx_start() \
{ \
        UINT    Cr_val; \
        NsmRegRead32(pAdapter->pNsmContext, (pAdapter->RegAddr + CR), &Cr_val); \
        Cr_val |= TXE; \
        Cr_val |= (pAdapter->TxDp[0][PRIQSEL]) ; \
        NsmRegWrite32(pAdapter->pNsmContext, (pAdapter->RegAddr + CR), Cr_val); \
}   

#define tx_forw_tail() \
{\
	HsmTCB                   *pCurrTcb  =NULL;\
        volatile DevDesc         *currDesc  =NULL;\
        volatile DevDesc         *currDescPa=NULL;\
	volatile DevDesc         *descBase  =NULL;\
        volatile DevDesc         *descBasePa=NULL;\
\
	pCurrTcb   = GET_TCB_NEXT(pAdapter, 0);\
	currDesc   = pCurrTcb->pEndDesc;\
        currDescPa = pCurrTcb->pEndDescPa; \
\
	if ( ! pCurrTcb->pEndDesc) {\
              printk(KERN_EMERG "GAMMA: tx_forw_tail() ERROR pCurrTcb->pEndDesc==NULL\n");\
	}\
\
        descBase   = pAdapter->PriQue[0].pTxdArray;\
        descBasePa = pAdapter->PriQue[0].pTxdArrayPa;\
\
        currDesc   = (DevDesc *)INC_CURR_DESC(pAdapter, 0, currDesc,  1, descBase);\
        currDescPa = (DevDesc *)INC_CURR_DESC(pAdapter, 0, currDescPa,1, descBasePa);\
\
/* set pStartDesc for next tcb !!! IMPORTANT */ \
        pCurrTcb               = pCurrTcb->pNext;\
        pCurrTcb->pStartDesc   = currDesc;\
        pCurrTcb->pStartDescPa = currDescPa;\
\
        SET_TCB_NEXT(pAdapter, 0, pCurrTcb);\
}

#define tx_flush_ring(tx_desc_indx,n,i) \
{\
        HsmTCB  *pTcbList;\
        time_586 a,b; \
\
	i = 1;\
	for (;;){\
\
		/* as long as there are tx packets enqueued \
                   and tx_desc_indx is not reached */ \
\
        	pTcbList   = GET_TCB_HEAD(pAdapter, 0);\
\
		/* break condition */\
        	if ((tx_ring_head>=(tx_desc_indx)) && (tx_ring_tail >= tx_ring_head)) \
                	break; \
        	if ((tx_ring_head>=(tx_desc_indx)) && (tx_ring_tail < (tx_desc_indx))) \
                	break; \
        	if ((tx_ring_tail>=tx_ring_head) && (tx_ring_tail < (tx_desc_indx))) \
                	break; \
\
        	gamma_time(a); \
\
		for (;;) {\
\
                	if( TX_COMPLETE(pTcbList->pEndDesc) ){\
\
                        	if(TX_ERRORED(pTcbList->pEndDesc)) {\
\
                           		pTcbList->TxStatus = TCB_ERROR; /* Revisit */\
                        	}else{\
                           		pTcbList->TxStatus = TCB_OK;\
                        	}\
\
                        	pTcbList->pEndDesc = NULL;\
                        	pTcbList = pTcbList->pNext;\
				SET_TCB_HEAD(pAdapter, 0, pTcbList);\
				break;\
			}\
			gamma_time(b);\
\
			if (gamma_time_diff_ticks(b,a) > (unsigned long long)(n)*CLOCK ){\
				i = 0;\
				break;\
			}\
                }\
\
        	if (!i) { \
			printk(KERN_EMERG "GAMMA: tx_flush_ring(): failure at pos %u (tail is at %u).  \
				Bug in NIC fw?\n",\
				tx_ring_head, \
				tx_ring_tail); \
                	break;\
        	} \
         }\
}

#define tx_wait_ring_room(n,ret) \
{\
        HsmTCB    *pTcbList;\
        time_586   a,b; \
\
	ret = 1;\
\
	for (;;){\
\
		/* as long as there are tx packets enqueued \
                   and tx_desc_indx is not reached */ \
\
        	pTcbList   = GET_TCB_HEAD(pAdapter, 0);\
\
		/* break condition */\
        	if ( tx_ring_has_room() )\
                	break; \
\
        	gamma_time(a); \
\
		for (;;) {\
\
                	if( TX_COMPLETE(pTcbList->pEndDesc) ){\
\
                        	if(TX_ERRORED(pTcbList->pEndDesc)) {\
\
					printk("tx_wait_ring_room() errored Packet head=%i\n",tx_ring_head);\
\
                           		pTcbList->TxStatus = TCB_ERROR; /* Revisit */\
                        	}else{\
                           		pTcbList->TxStatus = TCB_OK;\
                        	}\
\
                        	pTcbList->pEndDesc = NULL;\
                        	pTcbList = pTcbList->pNext;\
				SET_TCB_HEAD(pAdapter, 0, pTcbList);\
				break;\
			}\
			gamma_time(b);\
\
			if (gamma_time_diff_ticks(b,a) > (unsigned long long)(n)*CLOCK ){\
				ret = 0;\
				break;\
			}\
                }\
\
        	if (!ret) { \
			printk(KERN_EMERG "tx_wait_ring_room(): Tx ring blocked %lu ms at pos %u (tail is at %u).  \
				Bug in NIC fw?\n",\
				(unsigned long)n/1000,\
				tx_ring_head, \
				tx_ring_tail); \
\
                		break;\
        	} \
         }\
}

#define tx_close()

/*======================================================================*/
/*  RX MACROS                                                           */
/*======================================================================*/
#define rx_debug \
{\
HsmRCB          *pRcb=NULL;\
unsigned long   start_i;\
unsigned long   rxdp;\
int             ii;\
UINT            RXDP_val; \
\
NsmRegRead32(pAdapter->pNsmContext, (pAdapter->RegAddr + RXDP), &RXDP_val); \
rxdp=((unsigned long)RXDP_val - (unsigned long)(pAdapter->PriQue[0].pRxdArrayPa)) >> 4;\
printk("RXDP=%lu\n",rxdp);\
\
pRcb=pAdapter->PriQue[0].pRcbArray;\
printk("rx_ring_tail=%u\n",rx_ring_tail);\
\
   for (ii=0; ii<GAMMA_TX_RING_SIZE; ii++) {\
        int            more     = 9;\
        int            own      = 9;\
        int            intr     = 9;\
        unsigned long  link     = 999;\
\
        if (pRcb->pRxDp){\
                start_i   = ((unsigned long)pRcb->pRxDp - (unsigned long)(pAdapter->PriQue[0].pRxdArray)) >> 4;\
                more      = (pRcb->pRxDp->CmdSts & MORE) ? 1 : 0;\
                own       = (pRcb->pRxDp->CmdSts & OWN)  ? 1 : 0;\
                intr      = (pRcb->pRxDp->CmdSts & INTR) ? 1 : 0;\
                link      = ((unsigned long)pRcb->pRxDp->link - (unsigned long)(pAdapter->PriQue[0].pRxdArrayPa)) >> 4;\
        }else{\
                start_i = 999;\
        }\
\
        printk("rcb=%d s=%lu,  M=%1d, O=%1d, I=%1d L=%lu\n",\
                ii,    start_i,more,  own    ,intr,link);\
\
        pRcb = pRcb->pNext;\
    }/* end for */\
}

#define rx_common_decl_and_init \
AdapterContext  *pAdapter = &((NsmContext*)gamma_device->priv)->AdpCtxt; \
void  *ptrentry = NULL

#define poll_common_decl_and_init \
AdapterContext  *pAdapter = &((NsmContext*)gamma_device->priv)->AdpCtxt; \
void  *ptrentry = NULL

#define rx_ring_head 0

//#define rx_ring_tail pAdapter->PriQue[0].pRcbListNext
#define rx_ring_tail ( ((unsigned int)(pAdapter->PriQue[0].pRcbListNext->pRxDp)- \
                        (unsigned int)(pAdapter->PriQue[0].pRxdArray)) >> 4 )


#define read_rx_ring_tail(tail) \
tail = (unsigned long)(pAdapter->PriQue[0].pRcbListNext)

#define write_rx_ring_tail(tail) \
pAdapter->PriQue[0].pRcbListNext = (HsmRCB *)tail

#define forward_rx_ring_tail() \
/* MORE_INSTANCES_ON_SAME_NODE */\
{\
        HsmRCB *pRcb  = pAdapter->PriQue[0].pRcbListNext; \
\
        /* inc next */\
        SET_RCB_NEXT(pAdapter, 0 , pRcb->pNext); \
}

#define rx_status() 0

#ifdef NO_IRQ_RECV
#define clear_rx_irqs()
#else
#define clear_rx_irqs()\
{\
UINT interruptStatus;\
NsmRegRead32(pAdapter->pNsmContext, pAdapter->RegAddr + ISR, &interruptStatus );\
}
#endif

#define rx_read_pkt_info() \
{\
	HsmRCB	 *pRcbListNext = NULL;\
\
	pRcbListNext = GET_RCB_NEXT(pAdapter, 0);\
\
	if ( (RX_COMPLETE(pRcbListNext->pRxDp)) )\
		ptrentry = pRcbListNext->pDataBuf;\
	else\
		ptrentry = NULL; \
}

#define rx_err_generic() 0

#define rx_err_pktnotavail() (ptrentry == NULL)

#define discard_pkt() \
{\
	HsmRCB *pRcb = NULL;\
\
	pRcb=pAdapter->PriQue[0].pRcbListNext; \
\
	/* mark rcb free */\
	NsmMarkRcbFree(pRcb);\
\
	/* mark descriptor free */\
	DESC_CLEAR_OWN(pRcb->pRxDp);\
        pRcb->pRxDp->CmdSts = (pAdapter->MaxPktSize | INTR);\
\
	/* inc rcb pointer */\
	SET_RCB_NEXT(pAdapter, 0, pRcb->pNext); \
\
	/* rx start */\
	if ( !(pAdapter->interruptStatus & RXRCMP) )\
                HsmRxStart(pAdapter, 0); \
}

#define rx_header(h) (h) = (l2s *)(ptrentry)

#define invoke_standard_rx(dev) \
{\
	HsmRxPackets( &((NsmContext*)dev->priv)->AdpCtxt );\
\
	/* rx start */\
        if ( !( ((NsmContext*)dev->priv)->AdpCtxt.interruptStatus & RXRCMP) )\
                HsmRxStart(&((NsmContext*)dev->priv)->AdpCtxt, 0);\
}

#define rx_suspend_nic() \
{ \
UINT           Imr_val;\
AdapterContext  *pAdapter = NULL;\
\
pAdapter=&((NsmContext*)gamma_device->priv)->AdpCtxt;\
\
NsmRegRead32(pAdapter->pNsmContext, (pAdapter->RegAddr + IMR), &Imr_val);\
Imr_val &= ~(RXOK | RXERR | RXORN |RXSOVR | RXIDLE); \
NsmRegWrite32(pAdapter->pNsmContext,(pAdapter->RegAddr + IMR), Imr_val);\
gamma_irq_disabled = 1; \
}

#define rx_resume_nic() \
{ \
UINT           Imr_val;\
AdapterContext  *pAdapter = NULL;\
\
pAdapter=&((NsmContext*)gamma_device->priv)->AdpCtxt;\
\
NsmRegRead32(pAdapter->pNsmContext, (pAdapter->RegAddr + IMR), &Imr_val);\
Imr_val |= (RXOK | RXERR | RXORN |RXSOVR | RXIDLE); \
NsmRegWrite32(pAdapter->pNsmContext,(pAdapter->RegAddr + IMR), Imr_val);\
gamma_irq_disabled = 0; \
}
