/***************************************************************************
    COPYRIGHT NOTE

    GAMMA:		Genoa Active Message MAchine
			BIND2 layer:
			Basic Interface for Network Device Drivers.
			Derived from the Linux network driver ``tulip''

    Copyright		(C) 2001 Giuseppe Ciaccio
    e-mail:		ciaccio@disi.unige.it

    This program  or include  file  is free software,  either  original  or
    derived from other  software made available  under the terms of the GNU
    General Public License. You can redistribute and/or modify this program
    or include  file under the  terms of the GNU  General Public License as
    published  by the  Free Software Foundation;  either version 2  of  the
    License, or (at your option) any later version.

    THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR   IMPLIED
    WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
    MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
    NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT,  INDIRECT,
    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
    NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
    USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
    ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
    THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

    You should have received a copy of the  GNU General Public License along
    with this program; if not, write  to the Free Software Foundation, Inc.,
    675 Mass Ave, Cambridge, MA 02139, USA.

    END OF COPYRIGHT NOTE
****************************************************************************/

#define MAX_FLUSH_WAIT  100000	/* Max waiting (usec) for a Tx pkt to flush */
#define GAMMA_TX_RING_SIZE 128	/*
				 *  Size of Tx descriptor ring.
				 */
#define GAMMA_RX_RING_SIZE 1024	/*
				 *  Size of Rx descriptor ring.
				 *  We need a really large ring size due to
				 *  the static credit-based flow control.
				 */
#ifdef GAMMA_FULL_DUPLEX
#define MAX_CREDIT_SIZE 28
#define LOW_WATERMARK 4		/*  Must be > 0  */
#else
#define MAX_CREDIT_SIZE 30
#define LOW_WATERMARK 2		/*  Must be > 0  */
#endif
				/*
				 *  The sum MAX_CREDIT_SIZE + LOW_WATERMARK
				 *  must NOT exceed the following quantity:
				 *  (GAMMA_RX_RING_SIZE-20)/(total_num_nodes-1)
				 */
#define FRAME_SIZ 1500		/*  Max payload chunk = 1476 bytes  */

/*======================================================================*/
/*  TX MACROS                                                           */
/*======================================================================*/

#define tx_optimal_frag(l) FRAME_SIZ

#define tx_common_decl_and_init \
struct tulip_private *tp = (struct tulip_private *)gamma_device->priv; \
unsigned long ioaddr = gamma_device->base_addr; \
int i

#define tx_ring_head (tp->dirty_tx % GAMMA_TX_RING_SIZE)
#define tx_ring_tail (tp->cur_tx % GAMMA_TX_RING_SIZE)
#define tx_ring_occupancy ((tp->cur_tx - tp->dirty_tx) % GAMMA_TX_RING_SIZE)

#define tx_status() \
(unsigned long)(tp->tx_ring[tp->dirty_tx % GAMMA_TX_RING_SIZE].status)

#define tx_reset_and_restart() tp->dirty_tx = tp->cur_tx

/* peppino */
#define tx_enq_pkt(header_ptr,header_size,payload_ptr,payload_size) \
for (;;) { /*  Allows easy escapes in fall-through programming style  */ \
if (header_size+payload_size <= HEADER_SIZE+MAX_1ST_PIECE_SIZE) { \
unsigned long entry = tp->cur_tx % GAMMA_TX_RING_SIZE; \
memcpy( &(((unsigned char *)header_ptr)[header_size]), \
	(void *)bus_to_virt((unsigned long)(payload_ptr)), \
	payload_size ); \
tp->tx_ring[entry].buffer1 = cpu_to_le32(virt_to_bus(header_ptr)); \
tp->tx_ring[entry].length \
	= cpu_to_le32(TD_LS | TD_FS | (header_size+payload_size)); \
if (entry != GAMMA_TX_RING_SIZE - 1) { \
	tp->tx_ring[entry].status = cpu_to_le32(DescOwned); \
	wmb(); \
	break; \
} \
tp->tx_ring[entry].length |= DESC_RING_WRAP; \
tp->tx_ring[entry].status = cpu_to_le32(DescOwned); \
wmb(); \
break; \
} \
{ \
unsigned long entry = tp->cur_tx % GAMMA_TX_RING_SIZE; \
memcpy( &(((unsigned char *)header_ptr)[header_size]), \
	(void *)bus_to_virt((unsigned long)(payload_ptr)), \
	payload_size ); \
tp->tx_ring[entry].buffer1 = cpu_to_le32(virt_to_bus(header_ptr)); \
tp->tx_ring[entry].length \
	= cpu_to_le32(TD_LS | TD_FS | (header_size+payload_size)); \
if (entry != GAMMA_TX_RING_SIZE - 1) { \
	tp->tx_ring[entry].status = cpu_to_le32(DescOwned); \
	wmb(); \
	break; \
} \
tp->tx_ring[entry].length |= DESC_RING_WRAP; \
tp->tx_ring[entry].status = cpu_to_le32(DescOwned); \
wmb(); \
break; \
} \
}

#define tx_enq_pkt1(pkt_ptr,pkt_size) \
for (;;) { /*  Allows easy escapes in fall-through programming style  */ \
unsigned long entry = tp->cur_tx % GAMMA_TX_RING_SIZE; \
tp->tx_ring[entry].buffer1 = cpu_to_le32(virt_to_bus(pkt_ptr)); \
tp->tx_ring[entry].length = cpu_to_le32(TD_LS | TD_FS | (pkt_size)); \
if (entry != GAMMA_TX_RING_SIZE - 1) { \
	wmb(); \
	tp->tx_ring[entry].status = cpu_to_le32(DescOwned); \
	wmb(); \
	break; \
} \
tp->tx_ring[entry].length |= DESC_RING_WRAP; \
wmb(); \
tp->tx_ring[entry].status = cpu_to_le32(DescOwned); \
wmb(); \
break; \
}

#define tx_start() outl(0, ioaddr + CSR1)

#define tx_forw_tail() tp->cur_tx++

#define tx_flush_ring(tx_desc_indx,n,i) \
{ \
i = 1; \
for (;;) { \
	s32 status; \
	time_586 a,b; \
	gamma_time(a); \
	for (;;) { \
		status = (s32)le32_to_cpu(\
		tp->tx_ring[tp->dirty_tx % GAMMA_TX_RING_SIZE].status\
		); \
		if (status >= 0) { \
			tp->dirty_tx++; \
			break; \
		} \
		wmb(); \
		gamma_time(b); \
		if ((unsigned long)gamma_time_diff_ticks(b,a) > (n)*CLOCK) { \
			i = 0; \
			break; \
		} \
	} \
	if (!i) { \
printk("tx_flush_ring(): failure at pos %d, tx_status %8.8lx (tail is at %d).\n",tp->dirty_tx, (unsigned long)status, tp->cur_tx); \
		i = 1;  /* Ignoring failure */ \
		tp->dirty_tx++; \
	} \
	if (tx_ring_head == (tx_desc_indx)) break; \
} \
}

#define tx_flush_ring_1(tx_desc_indx,n,i) \
{ \
i = 1; \
for (;;) { \
	s32 status; \
	time_586 a,b; \
	gamma_time(a); \
	for (;;) { \
		status = (s32)le32_to_cpu(\
		tp->tx_ring[tp->dirty_tx % GAMMA_TX_RING_SIZE].status\
		); \
		if (status >= 0) { \
			tp->dirty_tx++; \
			break; \
		} \
		wmb(); \
		gamma_time(b); \
		if ((unsigned long)gamma_time_diff_ticks(b,a) > (n)*CLOCK) { \
			i = 0; \
			break; \
		} \
	} \
	if (!i) { \
printk("tx_flush_ring_1(): failure at pos %d, tx_status %8.8lx (tail is at %d).\n",tp->dirty_tx, (unsigned long)status, tp->cur_tx); \
		tp->dirty_tx++; \
	} \
	if (tx_ring_head == (tx_desc_indx)) break; \
} \
}

#define tx_wait_ring_room(n,i) \
{ \
int guard = GAMMA_TX_RING_SIZE; \
i = 1; \
for (;guard;guard--) { \
	s32 status; \
	time_586 a,b; \
	gamma_time(a); \
	for (;;) { \
		status = (s32)le32_to_cpu(\
		tp->tx_ring[tp->dirty_tx % GAMMA_TX_RING_SIZE].status\
		); \
		if (status >= 0){ \
			tp->dirty_tx++; \
			break; \
		} \
		wmb(); \
		gamma_time(b); \
		if ((unsigned long)gamma_time_diff_ticks(b,a) > (n)*CLOCK) { \
			i = 0; \
			break; \
		} \
	} \
	if (!i) { \
printk("tx_wait_ring_room(): failure at pos %d, tx_status %8.8lx (tail is at %d).\n",tp->dirty_tx, (unsigned long)status, tp->cur_tx); \
		i = 1;  /* Ignoring failure */ \
		if (guard) \
			tp->dirty_tx++; \
	} \
	if (tx_ring_has_room()) break; \
} \
if (!guard) { \
printk("tx_wait_ring_room(): Tx ring blocked at pos %d (tail is at %d).  Bug in NIC fw?  Resetting...\n",tp->dirty_tx, tp->cur_tx); \
	tp->dirty_tx = tp->cur_tx; \
} \
}

#define tx_close()

/*======================================================================*/
/*  RX MACROS                                                           */
/*======================================================================*/

#define rx_common_decl_and_init \
struct tulip_private *tp = (struct tulip_private *)dev->priv; \
int entry; \
s32 rx_status

#define poll_common_decl_and_init \
struct tulip_private *tp = (struct tulip_private *)gamma_device->priv; \
int entry; \
unsigned long ioaddr = gamma_device->base_addr; \
s32 rx_status

#define rx_status() rx_status

#define rx_ring_tail (tp->cur_rx % GAMMA_RX_RING_SIZE)
#define rx_ring_head (tp->dirty_rx % GAMMA_RX_RING_SIZE)
#define read_rx_ring_tail(tail) tail = tp->cur_rx
#define write_rx_ring_tail(tail) tp->cur_rx = tail
#define forward_rx_ring_tail() tp->cur_rx++

#ifdef NO_IRQ_RECV
#define clear_rx_irqs()
#else
#define clear_rx_irqs() outl(STS_RI,ioaddr + CSR5)
#endif

#define rx_read_pkt_info() \
{ \
entry = tp->cur_rx % GAMMA_RX_RING_SIZE; \
rx_status = (s32)cpu_to_le32(tp->rx_ring[entry].status); \
}

#define rx_err_generic() (rx_status & RxDescFatalErr)
#define rx_err_pktnotavail() (rx_status < 0)

#define discard_pkt() \
{ \
/*  Pop top Rx packet.  */ \
tp->rx_ring[entry].status = cpu_to_le32(DescOwned); wmb(); \
tp->cur_rx++; \
tp->dirty_rx++; \
}

#define rx_header(h) (h) = (l2s *)(bus_to_virt(tp->rx_ring[entry].buffer1))

#define invoke_standard_rx(dev) tulip_rx(dev)

#define rx_suspend_nic() \
{ \
unsigned long ioaddr = gamma_device->base_addr; \
s32 imr; \
imr = inl(ioaddr + CSR7); \
imr &= ~(IMR_NIM | IMR_AIM); \
outl(imr, ioaddr + CSR7);	/* Disable the IRQs */ \
imr = inl(ioaddr + CSR7); \
imr &= ~(IMR_RIM | IMR_UNM); \
outl(imr, ioaddr + CSR7);	/* Mask the IRQs */ \
imr = inl(ioaddr + CSR5);	/* Read IRQ status */ \
outl(imr, ioaddr + CSR5);	/* Reset the board interrupts */ \
gamma_irq_disabled = 1; \
}

#define rx_resume_nic() \
{ \
unsigned long ioaddr = gamma_device->base_addr; \
s32 imr; \
imr = inl(ioaddr + CSR7); \
imr |= (IMR_RIM | IMR_UNM); \
outl(imr, ioaddr + CSR7);	/* Unmask the IRQs */ \
imr = inl(ioaddr + CSR7); \
imr |= (IMR_NIM | IMR_AIM); \
outl(imr, ioaddr + CSR7);	/* Enable the IRQs */ \
gamma_irq_disabled = 0; \
}
