#undef DEFEAT_CACHE
#undef MEMCPY
#undef NOTIF
#undef SOURCE_IN_KERNEL
#undef CORRECTNESS

#ifdef MEMCPY
#undef SOURCE_IN_KERNEL
#endif

#define SEND_TIME
#define PERFORMANCE
#undef PERF_SHORT
#undef DELAY
#undef FREQUENCIES

#include <linux/damnick/libdamnick.h>
#include <stdio.h>
#include <stdlib.h>
#include "measure.h"

unsigned char *dest,*dest1;

int main(int argc, char **argv) {

#ifndef SOURCE_IN_KERNEL
	unsigned char *source,*source1;
#endif
	int len;
#ifdef NOTIF
	volatile unsigned long sem;
	unsigned long sem1;
#endif
#ifdef DEFEAT_CACHE
	unsigned char *dummy_buf;
#endif
#ifdef PERFORMANCE
#ifdef SEND_TIME
	double record_mc[NPROVE];
#endif
	double record_mm[NPROVE];
#endif
	int i, j;
	time_586
		t_send_s,
		t_rec;
#ifdef SEND_TIME
	time_586
		t_send_e;
#endif
#ifdef PERFORMANCE
	double time_overhead;
#endif

	damnick_init();
   	len = atoi(argv[1]);
	/*  Allocate data space (+4096 to allow page alignment)  */
#ifndef SOURCE_IN_KERNEL
	source1 = (unsigned char *)malloc(4096+len*sizeof(unsigned char));
#endif
	dest1 = (unsigned char *)malloc(4096+len*sizeof(unsigned char));
	/*  Align to page boundary  */
#ifndef SOURCE_IN_KERNEL
	source = (unsigned char *)(((unsigned long)source1+4096) & (~4095));
#endif
	dest = (unsigned char *)(((unsigned long)dest1+4096) & (~4095));
#ifdef DEFEAT_CACHE
	dummy_buf = (unsigned char *)malloc(1048576);
#endif

	/*  Pin data space in RAM.  */
#ifndef SOURCE_IN_KERNEL
	damnick_mlock(source,len);
#endif
	damnick_mlock(dest,len);
#ifdef NOTIF
	damnick_mlock((unsigned long *)&sem,sizeof(unsigned long));
#endif

#ifndef SOURCE_IN_KERNEL
	/*  Init the buffer content.  */
	for (j=0;j<len;j++) {
		source[j] = (unsigned char)j;
	}
#endif
#ifdef NOTIF
	sem = 0;
	sem1 = 0;
#endif

#ifdef PERFORMANCE
	/*  Computing the damnick_time() overhead...  */
	for (i=0;i<NPROVE;i++) {
		damnick_time(t_send_s);
 		damnick_time(t_rec);
		record_mm[i] = damnick_time_diff(t_rec, t_send_s);
	}
	{
	double Mmm,medmm,varmm;
	mean_variance(record_mm,&time_overhead,&Mmm,&medmm,&varmm);
	}
#endif

	for (i=0;i<NPROVE;i++) {
		for (j=0;j<len;j++) dest[j] = (unsigned char)0;

#ifdef DEFEAT_CACHE
		for (j=0;j<1048576;j++)
			dummy_buf[j] = j;
#endif
#ifdef PERFORMANCE
		damnick_time(t_send_s);
#endif

#ifdef NOTIF
#ifdef MEMCPY
		memcpy(dest,source,len);
		sem++;
#else
#ifdef SOURCE_IN_KERNEL
		damnick_icopy_k2u_notif(dest,len,(unsigned long *)&sem,sem+1);
#else
		damnick_icopy_notif(
			dest,source,len,(unsigned long *)&sem,sem+1
		);
#endif
#endif
#else
#ifdef MEMCPY
		memcpy(dest,source,len);
#else
#ifdef SOURCE_IN_KERNEL
		damnick_icopy_k2u(dest,len);
#else
		damnick_icopy(dest,source,len);
#endif
#endif
#endif

#ifdef PERFORMANCE
#ifdef SEND_TIME
		damnick_time(t_send_e);
#endif
#endif
		/*  This for debug only!
		copy_wait();
		*/
#ifdef NOTIF
		for (;;) {
			if (sem != sem1)
				break;
			/*
			damnick_sched_yield();
			*/
		}
		sem1++;
#else
		for (;;) {
			volatile unsigned char flag;
			flag = dest[len-1];
			if (flag)
				break;
		}
#endif
#ifdef PERFORMANCE
		damnick_time(t_rec);
#endif

#ifdef CORRECTNESS
		/*  Check data correctness.  */
#ifdef NOTIF
		for (j=0;j<len;j++)
#else
		for (j=0;j<len-1;j++)
#endif
			if (dest[j] != (unsigned char)j) {
				int k;
				printf("data error\n");
				for (k=j-5;k<j+10;k++)
					printf("%d:%d ",k,dest[k]);
				printf("\n");
			}
#endif

#ifdef PERFORMANCE
#ifdef SEND_TIME
		record_mc[i] =
		damnick_time_diff(t_send_e, t_send_s)-time_overhead;
#endif
		record_mm[i] =
		damnick_time_diff(t_rec, t_send_s)-time_overhead;
#endif
	}
	/*
	 *  Force printout of kernel-level time measurements,
	 *  if enabled.  Otherwise nothing happens.
	 */
	damnick_timers(t_send_s,t_rec);

	damnick_munlockall();
#ifndef SOURCE_IN_KERNEL
	free(source1);
#endif
	free(dest1);
#ifdef PERFORMANCE
	{
#ifdef PERF_SHORT
#ifdef SEND_TIME
	double mmc,Mmc,medmc,varmc;
	mean_variance(record_mc,&mmc,&Mmc,&medmc,&varmc);
#ifdef DELAY
        printf("%d %g\n",len,medmc);
#else
	if (!len)
	        printf("%d %g\n",len,medmc);
	else
	        printf("%d %g\n",len,len/medmc);
#endif
#else
	double mmm,Mmm,medmm,varmm;
	mean_variance(record_mm,&mmm,&Mmm,&medmm,&varmm);
#ifdef DELAY
        printf("%d %g\n",len,medmm);
#else
	if (!len)
	        printf("%d %g\n",len,medmm);
	else
	        printf("%d %g\n",len,len/medmm);
#endif
#endif
#else
#ifdef SEND_TIME
	double mmc,Mmc,medmc,varmc;
#endif
	double mmm,Mmm,medmm,varmm;
	printf("\n\n");
	printf("PERFORMANCE (microsec, MB/sec) (lowest < average < highest):\n\n");
#ifdef SEND_TIME
	mean_variance(record_mc,&mmc,&Mmc,&medmc,&varmc);
        printf("total send time        %g < %g < %g\n",
		mmc,medmc,Mmc);
	if (len)
	        printf("send throughput        %g < %g < %g\n",
			len/Mmc,len/medmc,
			len/mmc);
#endif
	mean_variance(record_mm,&mmm,&Mmm,&medmm,&varmm);
        printf("total delay            %g < %g < %g\n",
		mmm,medmm,Mmm);
	if (len)
	        printf("end-to-end throughput  %g < %g < %g\n",
			len/Mmm,len/medmm,
			len/mmm);
#endif
	printf("\n\n");

        printf("VARIANCE:  \n");
#ifdef SEND_TIME
        printf("send time        %g\n", varmc);
#endif
        printf("end-to-end delay %g\n", varmm);
	printf("\n\n");

#ifdef FREQUENCIES
        printf("FREQUENCIES:\n");
#ifdef SEND_TIME
	printf("send time:\n");
	frequencies(record_mc,mmc,Mmc);
#endif
	printf("end-to-end delay:\n");
	frequencies(record_mm,mmm,Mmm);
#endif
	fflush(stdout);
	}
#endif
	return 0;
}
