/***************************************************************************
    COPYRIGHT NOTE

    DAMNICK:		DAta Moving through a NIC connected in loopbacK
			BIND2 layer:
			Basic Interface for Network Device Drivers
			(from the GAMMA project).
			Derived from the Linux network driver acenic.c

    Copyright		(C) 1999-2001 Luca Landi, Giuseppe Ciaccio
    e-mail:		ciaccio@disi.unige.it

    This program  or include  file  is free software,  either  original  or
    derived from other  software made available  under the terms of the GNU
    General Public License. You can redistribute and/or modify this program
    or include  file under the  terms of the GNU  General Public License as
    published  by the  Free Software Foundation;  either version 2  of  the
    License, or (at your option) any later version.

    THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR   IMPLIED
    WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
    MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
    NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT,  INDIRECT,
    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
    NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
    USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
    ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
    THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

    You should have received a copy of the  GNU General Public License along
    with this program; if not, write  to the Free Software Foundation, Inc.,
    675 Mass Ave, Cambridge, MA 02139, USA.

    END OF COPYRIGHT NOTE
****************************************************************************/

#define MAX_FLUSH_WAIT  100000	/* Max waiting (usec) for a Tx pkt to flush */
#define DAMNICK_TX_RING_SIZE TX_RING_ENTRIES

#define MIN_FRAME_SIZ 101
	/*
	 *  60 would also work, but there is little difference with 101 and
	 *  this way we do not use "mini frames"...
	 */

#define USE_JUMBO_FRAMES

#ifndef USE_JUMBO_FRAMES
#define FRAME_SIZ 1500
#else
#define FRAME_SIZ 9000
#endif

/*======================================================================*/
/*  TX MACROS                                                           */
/*======================================================================*/

#define tx_optimal_frag(l) FRAME_SIZ

#define tx_common_decl_and_init \
struct ace_private *ap = (struct ace_private *)damnick_device->priv; \
struct ace_regs *regs = ap->regs; \
int i = 1

#define tx_ring_head ap->tx_ret_csm
#define tx_ring_tail ap->tx_prd
#define tx_ring_occupancy \
((ap->tx_ret_csm <= ap->tx_prd) ? \
ap->tx_prd - ap->tx_ret_csm : \
TX_RING_ENTRIES - (ap->tx_ret_csm - ap->tx_prd))

#define tx_status() (unsigned long)0

#define tx_reset_and_restart() ap->tx_ret_csm = ap->tx_prd

#define tx_enq_pkt_noheader(payload_ptr,payload_size) \
{ \
	struct tx_desc *descriptor = &ap->tx_ring[ap->tx_prd]; \
	writel((unsigned long)(payload_ptr) & 0xffffffff, \
		&descriptor->addr.addrlo); \
	writel(((payload_size) << 16) | (BD_FLG_END), &descriptor->flagsize); \
}

#define tx_enq_pkt2_noheader(payload1_ptr,payload1_size,payload2_ptr,payload2_size) \
{ \
	struct tx_desc *descriptor = &ap->tx_ring[ap->tx_prd]; \
	writel((unsigned long)(payload1_ptr) & 0xffffffff, \
		&descriptor->addr.addrlo); \
	writel(((payload1_size) << 16) | (BD_FLG_MORE), \
		&descriptor->flagsize); \
	ap->tx_prd = (ap->tx_prd+1) % TX_RING_ENTRIES; \
	descriptor = &ap->tx_ring[ap->tx_prd]; \
	writel((virt_to_bus(payload2_ptr)) & 0xffffffff, \
		&descriptor->addr.addrlo); \
	writel(((payload2_size) << 16) | (BD_FLG_END), \
		&descriptor->flagsize); \
}

#define tx_start() \
{ \
mb(); \
ap->tx_prd = (ap->tx_prd+1) % TX_RING_ENTRIES;\
writel(ap->tx_prd, &regs->TxPrd); \
mb(); \
}

#define tx_forw_tail()

#define tx_flush_ring(tx_desc_indx,n,i) \
{ \
i = 1; \
for (;;) { \
	time_586 a,b; \
	if ((tx_ring_head>=(tx_desc_indx)) && (tx_ring_tail >= tx_ring_head)) \
		break; \
	if ((tx_ring_head>=(tx_desc_indx)) && (tx_ring_tail < (tx_desc_indx))) \
		break; \
	if ((tx_ring_tail>=tx_ring_head) && (tx_ring_tail < (tx_desc_indx))) \
		break; \
	gamma_time(a); \
	for (;;) { \
		mb(); \
		if (*(ap->tx_csm) != ap->tx_ret_csm) {  \
			ap->tx_ret_csm = *(ap->tx_csm); \
			break; \
		} \
		barrier(); \
		gamma_time(b); \
		if ( \
gamma_time_diff_ticks(b,a) > (unsigned long long)(n)*CLOCK \
		) { \
			i = 0; \
			break; \
		} \
	} \
	if (!i) { \
printk("tx_flush_ring(): failure at pos %d (tail is at %d).  Bug in NIC fw?  Ignoring...\n",ap->tx_ret_csm, ap->tx_prd); \
		/*  Ignoring failure  */ \
		i = 1; \
		ap->tx_ret_csm = (ap->tx_ret_csm + 1) % TX_RING_ENTRIES; \
		break; \
	} \
} \
}

#define tx_wait_ring_room(n,i) \
{ \
int guard = DAMNICK_TX_RING_SIZE; \
i = 1; \
for (;guard;guard--) { \
	time_586 a,b; \
	gamma_time(a); \
	for (;;) { \
		mb(); \
		if (*(ap->tx_csm) != ap->tx_ret_csm) {  \
			ap->tx_ret_csm = *(ap->tx_csm) ; \
			break; \
		} \
		barrier(); \
		gamma_time(b); \
		if ( \
gamma_time_diff_ticks(b,a) > (unsigned long long)(n)*CLOCK \
		) { \
			i = 0; \
			break; \
		} \
	} \
	if (!i) { \
printk("tx_wait_ring_room(): failure at pos %d (tail is at %d).  Bug in NIC fw?  Ignoring...\n",ap->tx_ret_csm, ap->tx_prd); \
		/*  Ignoring failure  */ \
		i = 1; \
		if (guard) \
/* ap->tx_ret_csm = (ap->tx_ret_csm + 1) % GAMMA_TX_RING_SIZE; */ \
			ap->tx_ret_csm = ap->tx_prd; \
	} \
	if (tx_ring_has_room()) break; \
} \
if (!guard) { \
printk("tx_wait_ring_room(): Tx ring blocked at pos %d (tail is at %d).  Bug in NIC fw?  Resetting...\n",ap->tx_ret_csm, ap->tx_prd); \
	ap->tx_ret_csm = ap->tx_prd; \
} \
}

#define tx_close()

/*======================================================================*/
/*  RX MACROS                                                           */
/*======================================================================*/

#define rx_common_decl_and_init \
struct ace_private *ap = (struct ace_private *)damnick_device->priv; \
int entry; \
void *ptrentry = NULL

#define poll_common_decl_and_init \
struct ace_private *ap = (struct ace_private *)damnick_device->priv; \
int entry; \
void *ptrentry = NULL

#define rx_ring_head (*(ap->rx_ret_prd) % RX_RETURN_RING_ENTRIES)
#define rx_ring_tail ap->cur_rx

#define rx_status() 0

#define clear_rx_irqs() \
/* writel(1,&(((struct ace_private *)(damnick_device->priv))->regs->Mb0Lo)) */

#define rx_read_pkt_info() \
for (;;) { \
if (ap->cur_rx == *(ap->rx_ret_prd)) { \
	ptrentry = NULL; \
	break; \
} \
switch( \
ap->rx_return_ring[entry = ap->cur_rx].flags & (BD_FLG_JUMBO | BD_FLG_MINI) \
) { \
	case 0: \
	ptrentry = bus_to_virt((unsigned long) \
	get_aceaddr(&ap->rx_std_ring[ap->rx_return_ring[entry].idx].addr)); \
	break; \
	case BD_FLG_MINI: \
	ptrentry = bus_to_virt((unsigned long) \
	get_aceaddr(&ap->rx_mini_ring[ap->rx_return_ring[entry].idx].addr)); \
	break; \
	case BD_FLG_JUMBO: \
	ptrentry = bus_to_virt((unsigned long) \
	get_aceaddr(&ap->rx_jumbo_ring[ap->rx_return_ring[entry].idx].addr)); \
	break; \
} \
break; \
}

#define rx_err_generic() 0

#define rx_err_pktnotavail()  (ptrentry == NULL)

#define discard_pkt() \
ap->cur_rx = (ap->cur_rx + 1) % RX_RETURN_RING_ENTRIES

#define rx_header(h) (h) = (l2s *)(ptrentry)

#define rx_enq_desc1(payload_ptr,payload_size) \
{ \
short idx; \
struct rx_desc *desc; \
for (;;) { \
	if ((payload_size) > 1518) { \
		idx = ap->cur_j_rx; \
		desc = &ap->rx_jumbo_ring[idx]; \
		desc->addr.addrlo = payload_ptr; \
		desc->size = (u16)(payload_size); \
		desc->idx = idx; \
		idx = (idx+1) % RX_JUMBO_RING_ENTRIES; \
		mb(); \
		writel(idx, &regs->RxJumboPrd); \
		mb(); \
		ap->cur_j_rx = idx; \
		break; \
	} \
/*  Don't use Mini-frames  \
	if ((payload_size) <= 100) { \
		idx = ap->cur_m_rx; \
		desc = &ap->rx_mini_ring[idx]; \
		desc->addr.addrlo = payload_ptr; \
		desc->size = (u16)(payload_size); \
		desc->idx = idx; \
		idx = (idx+1) % RX_MINI_RING_ENTRIES; \
		mb(); \
		writel(idx, &regs->RxMiniPrd); \
		mb(); \
		ap->cur_m_rx = idx; \
		break; \
	} */\
	idx = ap->cur_s_rx; \
	desc = &ap->rx_std_ring[idx]; \
	desc->addr.addrlo = payload_ptr; \
	desc->size = (u16)(payload_size); \
	desc->idx = idx; \
	idx = (idx+1) % RX_STD_RING_ENTRIES; \
	mb(); \
	writel(idx, &regs->RxStdPrd); \
	mb(); \
	ap->cur_s_rx = idx; \
	break; \
} \
}

#define rx_enq_desc2(payload1_ptr,payload1_size,payload2_ptr,payload2_size) \
{ \
short idx; \
struct rx_desc *desc; \
for (;;) { \
	if (((payload1_size)+(payload2_size)) > 1518) { \
		idx = ap->cur_j_rx; \
		desc = &ap->rx_jumbo_ring[idx]; \
		desc->addr.addrlo = payload1_ptr; \
		desc->size = (u16)(payload1_size); \
		idx = (idx+1) % RX_JUMBO_RING_ENTRIES; \
		desc->idx = idx; \
		desc = &ap->rx_jumbo_ring[idx]; \
		desc->addr.addrlo = payload2_ptr; \
		desc->size = (u16)(payload2_size); \
		desc->idx = idx; \
		idx = (idx+1) % RX_JUMBO_RING_ENTRIES; \
		mb(); \
		writel(idx, &regs->RxJumboPrd); \
		mb(); \
		ap->cur_j_rx = idx; \
		break; \
	} \
/*  Don't use Mini-frames  \
	if (((payload1_size)+(payload2_size)) <= 100) { \
		idx = ap->cur_m_rx; \
		desc = &ap->rx_mini_ring[idx]; \
		desc->addr.addrlo = payload1_ptr; \
		desc->size = (u16)(payload1_size); \
		idx = (idx+1) % RX_MINI_RING_ENTRIES; \
		desc->idx = idx; \
		desc = &ap->rx_mini_ring[idx]; \
		desc->addr.addrlo = payload2_ptr; \
		desc->size = (u16)(payload2_size); \
		desc->idx = idx; \
		idx = (idx+1) % RX_MINI_RING_ENTRIES; \
		mb(); \
		writel(idx, &regs->RxMiniPrd); \
		mb(); \
		ap->cur_m_rx = idx; \
		break; \
	} */\
	idx = ap->cur_s_rx; \
	desc = &ap->rx_std_ring[idx]; \
	desc->addr.addrlo = payload1_ptr; \
	desc->size = (u16)(payload1_size); \
	idx = (idx+1) % RX_STD_RING_ENTRIES; \
	desc->idx = idx; \
	desc = &ap->rx_std_ring[idx]; \
	desc->addr.addrlo = payload2_ptr; \
	desc->size = (u16)(payload2_size); \
	desc->idx = idx; \
	idx = (idx+1) % RX_STD_RING_ENTRIES; \
	mb(); \
	writel(idx, &regs->RxStdPrd); \
	mb(); \
	ap->cur_s_rx = idx; \
	break; \
} \
}

#define rx_enq_desc3(payload1_ptr,payload1_size,payload2_ptr,payload2_size,payload3_ptr,payload3_size) \
{ \
short idx; \
struct rx_desc *desc; \
for (;;) { \
	if (((payload1_size)+(payload2_size)+(payload3_size)) > 1518) { \
		idx = ap->cur_j_rx; \
		desc = &ap->rx_jumbo_ring[idx]; \
		desc->addr.addrlo = payload1_ptr; \
		desc->size = (u16)(payload1_size); \
		idx = (idx+1) % RX_JUMBO_RING_ENTRIES; \
		desc->idx = idx; \
		desc = &ap->rx_jumbo_ring[idx]; \
		desc->addr.addrlo = payload2_ptr; \
		desc->size = (u16)(payload2_size); \
		idx = (idx+1) % RX_JUMBO_RING_ENTRIES; \
		desc->idx = idx; \
		desc = &ap->rx_jumbo_ring[idx]; \
		desc->addr.addrlo = payload3_ptr; \
		desc->size = (u16)(payload3_size); \
		desc->idx = idx; \
		idx = (idx+1) % RX_JUMBO_RING_ENTRIES; \
		mb(); \
		writel(idx, &regs->RxJumboPrd); \
		mb(); \
		ap->cur_j_rx = idx; \
		break; \
	} \
/*  Don't use Mini-frames  \
	if (((payload1_size)+(payload2_size)+(payload3_size)) <= 100) { \
		idx = ap->cur_m_rx; \
		desc = &ap->rx_mini_ring[idx]; \
		desc->addr.addrlo = payload1_ptr; \
		desc->size = (u16)(payload1_size); \
		idx = (idx+1) % RX_MINI_RING_ENTRIES; \
		desc->idx = idx; \
		desc = &ap->rx_mini_ring[idx]; \
		desc->addr.addrlo = payload2_ptr; \
		desc->size = (u16)(payload2_size); \
		idx = (idx+1) % RX_MINI_RING_ENTRIES; \
		desc->idx = idx; \
		desc = &ap->rx_mini_ring[idx]; \
		desc->addr.addrlo = payload3_ptr; \
		desc->size = (u16)(payload3_size); \
		desc->idx = idx; \
		idx = (idx+1) % RX_MINI_RING_ENTRIES; \
		mb(); \
		writel(idx, &regs->RxMiniPrd); \
		mb(); \
		ap->cur_m_rx = idx; \
		break; \
	} */\
	idx = ap->cur_s_rx; \
	desc = &ap->rx_std_ring[idx]; \
	desc->addr.addrlo = payload1_ptr; \
	desc->size = (u16)(payload1_size); \
	idx = (idx+1) % RX_STD_RING_ENTRIES; \
	desc->idx = idx; \
	desc = &ap->rx_std_ring[idx]; \
	desc->addr.addrlo = payload2_ptr; \
	desc->size = (u16)(payload2_size); \
	idx = (idx+1) % RX_STD_RING_ENTRIES; \
	desc->idx = idx; \
	desc = &ap->rx_std_ring[idx]; \
	desc->addr.addrlo = payload3_ptr; \
	desc->size = (u16)(payload3_size); \
	desc->idx = idx; \
	idx = (idx+1) % RX_STD_RING_ENTRIES; \
	mb(); \
	writel(idx, &regs->RxStdPrd); \
	mb(); \
	ap->cur_s_rx = idx; \
	break; \
} \
}
