/***************************************************************************)
    COPYRIGHT NOTE

    DAMNICK:		DAta Moving through a NIC connected in loopbacK

    Copyright		(C) 2001 Giuseppe Ciaccio
    Institution:	DISI, Universita' di Genova
    Address:		via Dodecaneso 35, I-16146 Genova, Italy
    e-mail:		ciaccio@disi.unige.it

    This program  or include  file  is free software,  either  original  or
    derived from other  software made available  under the terms of the GNU
    General Public License. You can redistribute and/or modify this program
    or include  file under the  terms of the GNU  General Public License as
    published  by the  Free Software Foundation;  either version 2  of  the
    License, or (at your option) any later version.

    THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR   IMPLIED
    WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
    MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
    NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT,  INDIRECT,
    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
    NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
    USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
    ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
    THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

    You should have received a copy of the  GNU General Public License along
    with this program; if not, write  to the Free Software Foundation, Inc.,
    675 Mass Ave, Cambridge, MA 02139, USA.

    END OF COPYRIGHT NOTE
****************************************************************************/

#include "damnick.h"
#include "damnick_leading.h"
static char *damnick_authors = "ciaccio@disi.unige.it\n";


/************************************************************************
 *  DAMNICK Data structures						*
 ************************************************************************/

struct net_device *damnick_device;

void *d_next_dest = NULL;
unsigned long d_next_d_paddr = 0;

unsigned long d_notif_ring[DAMNICK_TX_RING_SIZE];
unsigned long d_notif_ring_tail;

unsigned char d_thrash[128];

#ifdef MEASURE_ICOPY
unsigned long icopy_counter = 0;
unsigned long icopy_amount = 0;
#endif

#ifdef MEASURE_COPYCOMPL
unsigned long ccopy_counter = 0;
unsigned long ccopy_amount = 0;
unsigned long ccopy_bytes = 0;
#endif

static char *damnick_version =
"\nDAMNICK  DAta Moving through a NIC connected in loopbacK  7 May 2002\n";
static char *damnick_debuglevel =
#ifdef DRIVER_DEBUG
"         Debug level: High\n";
#else
#ifdef DRIVER_DEBUG_LIGHT
"         Debug level: Low\n";
#else
"         Debug level: Null\n";
#endif
#endif

/************************************************************************
 *	Some function protos						*
 ************************************************************************/


/************************************************************************
 *	Init routines							*
 ************************************************************************/


void damnick_start
(struct net_device *dev)
{

	int i;


	damnick_device = dev;

	for (i=0;i<DAMNICK_TX_RING_SIZE;i++)
		d_notif_ring[i] = 0;
	d_notif_ring_tail = 0;

/*  Set the entry points for the damnick calls (trap 0x82)  */
	set_system_gate(0x82,&damnick_system_call);

	printk(damnick_version);
	printk(damnick_debuglevel);
#ifdef USE_JUMBO_FRAMES
	printk("         Jumbo frame support enabled:  MTU %d\n",FRAME_SIZ);
#endif
	printk("         DAMNICK protocol: ");
	printk(damnick_authors);
	printk("         DAMNICK driver  : ");
	printk(bind2_authors);
	printk("\n");

	return;

}


int
sys_damnick_init
(void)
{
/*  Return the CPU clock freq. in MHz  */
        return cpu_khz/1000;
}


/************************************************************************
 *	Init-Copy routines (non-blocking)				*
 ************************************************************************/


int sys_damnick_icopy_u2u
(
	void *dest,
	void *source,
	unsigned long len
)
{

	unsigned long frame_size;
	unsigned long cnk_len;
	unsigned long s_paddr;
	unsigned long d_paddr;
	unsigned long s_next_paddr;
	unsigned long d_next_paddr;
	tx_common_decl_and_init;
#ifdef ALLOW_IRQS
	time_586 t_start, t_elaps;
	unsigned long flags;
#endif


DEBUG(printk("sys_damnick_icopy_u2u(): Entering\n"));

	if (len < MIN_FRAME_SIZ)
		goto short_len;
	frame_size = tx_optimal_frag(len);
	cnk_len = lmin(len,frame_size);
	s_paddr = user_space_to_phys(current,source);
	d_paddr = user_space_to_phys(current,dest);
	s_next_paddr = user_space_to_phys(current,(char *)source+cnk_len);
	if (s_next_paddr != s_paddr+cnk_len) {
		cnk_len = (unsigned short)(PAGE_SIZE-(s_paddr & 0x00000fff));
		s_next_paddr =
			user_space_to_phys(current,(char *)source+cnk_len);
	}
	d_next_paddr = user_space_to_phys(current,(char *)dest+cnk_len);
	if (d_next_paddr != d_paddr+cnk_len) {
		cnk_len = (unsigned short)(PAGE_SIZE-(d_paddr & 0x00000fff));
		d_next_paddr =
			user_space_to_phys(current,(char *)dest+cnk_len);
		s_next_paddr =
			user_space_to_phys(current,(char *)source+cnk_len);
	}

#ifdef ALLOW_IRQS
	damnick_time(t_start);
#endif

	for (;;) {

		if (cnk_len < MIN_FRAME_SIZ)
			goto short_cnk;

		if (!tx_ring_has_room())
			goto no_room;

room_in_tx_ring:

DEBUG(printk(
"sys_damnick_icopy_u2u():  cnk_len %ld, s_paddr %8.8lx, d_paddr %8.8lx\n",
cnk_len,s_paddr,d_paddr
));

		rx_enq_desc1(d_paddr,cnk_len);
		tx_enq_pkt_noheader(s_paddr,cnk_len);
		tx_start();
		tx_forw_tail();
#ifdef SAFE_TX
		if (tx_ring_occupancy >= STEPS_THEN_STOP)
			tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif

pkt_sent:

		len -= cnk_len;
		if (!len)
			break;

#ifdef ALLOW_IRQS
		damnick_time(t_elaps);
		IRQ_enable_for_5usec(t_elaps,t_start);
#endif

/*  Prepare next frame  */
		source = (void *)((char *)source+cnk_len);
		dest = (void *)((char *)dest+cnk_len);
		cnk_len = lmin(len,frame_size);
		s_paddr = s_next_paddr;
		d_paddr = d_next_paddr;
		s_next_paddr =
			user_space_to_phys(current,(char *)source+cnk_len);
		if (s_next_paddr != s_paddr+cnk_len) {
			cnk_len = (unsigned short)
				(PAGE_SIZE-(s_paddr & 0x00000fff));
			s_next_paddr = user_space_to_phys(
				current,(char *)source+cnk_len
			);
		}
		d_next_paddr =
			user_space_to_phys(current,(char *)dest+cnk_len);
		if (d_next_paddr != d_paddr+cnk_len) {
			cnk_len = (unsigned short)
				(PAGE_SIZE-(d_paddr & 0x00000fff));
			d_next_paddr = user_space_to_phys(
				current,(char *)dest+cnk_len
			);
			s_next_paddr = user_space_to_phys(
				current,(char *)source+cnk_len
			);
		}

		continue;

short_cnk:

DEBUG(printk(
"sys_damnick_icopy_u2u(): data chunk is short (%ld bytes).  Copying from %8.8lx to %8.8lx\n",
cnk_len,s_paddr,d_paddr
));

		memcpy(bus_to_virt(d_paddr),bus_to_virt(s_paddr),cnk_len);
		goto pkt_sent;

	}

	tx_close();

DEBUG(printk("sys_damnick_icopy_u2u(): Done!\n"));

	return 0;

no_room:

	tx_wait_ring_room(MAX_FLUSH_WAIT,i);
	if (!i)
        	goto tx_failure;
	goto room_in_tx_ring;

tx_failure:

DEBUG_L(printk(
"sys_damnick_icopy_u2u(): Failure (ring pos %d, tx_status %8.8lx).\n",
tx_ring_head,tx_status()
));

	tx_reset_and_restart();
	return -DAMNICK_ERR_NICFAILURE;

short_len:

DEBUG(printk(
"sys_damnick_icopy_u2u(): data buffer is short (%ld bytes).  Copying...\n",
len
));

	cnk_len = len;
	d_paddr = user_space_to_phys(current,dest);
	d_next_paddr = user_space_to_phys(current,(char *)dest+cnk_len);
	if (d_next_paddr != d_paddr+cnk_len) {
		cnk_len = (unsigned short)(PAGE_SIZE-(d_paddr & 0x00000fff));
		d_next_paddr =
			user_space_to_phys(current,(char *)dest+cnk_len);
	}
	copy_from_user(bus_to_virt(d_paddr),source,cnk_len);
	len -= cnk_len;
	if (len)
		copy_from_user(
			bus_to_virt(d_next_paddr),(char *)source+cnk_len,len
		);
	return 0;

}


int sys_damnick_icopy_u2u_nu
(
	void *dest,
	void *source,
	unsigned long len,
	unsigned long *notif,
	unsigned long val
)
{

	unsigned long *tmp_notif;
	unsigned long n_paddr;
	unsigned long frame_size;
	unsigned long cnk_len;
	unsigned long s_paddr;
	unsigned long d_paddr;
	unsigned long s_next_paddr;
	unsigned long d_next_paddr;
	tx_common_decl_and_init;
#ifdef ALLOW_IRQS
	time_586 t_start, t_elaps;
	unsigned long flags;
#endif


DEBUG(printk("sys_damnick_icopy_u2u_nu(): Entering\n"));

	tmp_notif = &d_notif_ring[d_notif_ring_tail];
	d_notif_ring_tail = (d_notif_ring_tail+1) % DAMNICK_TX_RING_SIZE;
	*tmp_notif = val;

	n_paddr = user_space_to_phys(current,notif);
	if (len < MIN_FRAME_SIZ)
		goto short_len;
	frame_size = tx_optimal_frag(len);
	cnk_len = lmin(len,frame_size);
	s_paddr = user_space_to_phys(current,source);
	d_paddr = user_space_to_phys(current,dest);
	s_next_paddr = user_space_to_phys(current,(char *)source+cnk_len);
	if (s_next_paddr != s_paddr+cnk_len) {
		cnk_len = (unsigned short)(PAGE_SIZE-(s_paddr & 0x00000fff));
		s_next_paddr =
			user_space_to_phys(current,(char *)source+cnk_len);
	}
	d_next_paddr = user_space_to_phys(current,(char *)dest+cnk_len);
	if (d_next_paddr != d_paddr+cnk_len) {
		cnk_len = (unsigned short)(PAGE_SIZE-(d_paddr & 0x00000fff));
		d_next_paddr =
			user_space_to_phys(current,(char *)dest+cnk_len);
		s_next_paddr =
			user_space_to_phys(current,(char *)source+cnk_len);
	}

#ifdef ALLOW_IRQS
	damnick_time(t_start);
#endif

	for (;;) {

		if (!tx_ring_has_room())
			goto no_room;

room_in_tx_ring:

DEBUG(printk(
"sys_damnick_icopy_u2u_nu():  cnk_len %ld, s_paddr %8.8lx, d_paddr %8.8lx\n",
cnk_len,s_paddr,d_paddr
));

		len -= cnk_len;
		if (!len) {
			/*  Last data chunk to move. Perform notification.  */

DEBUG(printk("sys_damnick_icopy_u2u_nu():  n_paddr %8.8lx\n",n_paddr));

			if (cnk_len < MIN_FRAME_SIZ) {
/*  DMA + notification with padding to MIN_FRAME_SIZ byte  */
				rx_enq_desc3(
					d_paddr,cnk_len,
					(unsigned long)virt_to_bus(d_thrash),
					MIN_FRAME_SIZ-cnk_len,
					n_paddr,sizeof(unsigned long)
				);
				tx_enq_pkt2_noheader(
					s_paddr,MIN_FRAME_SIZ,
					tmp_notif,sizeof(unsigned long)
				);
			}
			else {
				rx_enq_desc2(
					d_paddr,cnk_len,
					n_paddr,sizeof(unsigned long)
				);
				tx_enq_pkt2_noheader(
					s_paddr,cnk_len,
					tmp_notif,sizeof(unsigned long)
				);
			}
		}
		else {
			if (cnk_len < MIN_FRAME_SIZ)
				goto do_memcpy;
			rx_enq_desc1(d_paddr,cnk_len);
			tx_enq_pkt_noheader(s_paddr,cnk_len);
		}
		tx_start();
		tx_forw_tail();
#ifdef SAFE_TX
		if (tx_ring_occupancy >= STEPS_THEN_STOP)
			tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif

after_enq:

		if (!len)
			break;

#ifdef ALLOW_IRQS
		damnick_time(t_elaps);
		IRQ_enable_for_5usec(t_elaps,t_start);
#endif

/*  Prepare next frame  */
		source = (void *)((char *)source+cnk_len);
		dest = (void *)((char *)dest+cnk_len);
		cnk_len = lmin(len,frame_size);
		s_paddr = s_next_paddr;
		d_paddr = d_next_paddr;
		s_next_paddr =
			user_space_to_phys(current,(char *)source+cnk_len);
		if (s_next_paddr != s_paddr+cnk_len) {
			cnk_len = (unsigned short)
				(PAGE_SIZE-(s_paddr & 0x00000fff));
			s_next_paddr = user_space_to_phys(
				current,(char *)source+cnk_len
			);
		}
		d_next_paddr =
			user_space_to_phys(current,(char *)dest+cnk_len);
		if (d_next_paddr != d_paddr+cnk_len) {
			cnk_len = (unsigned short)
				(PAGE_SIZE-(d_paddr & 0x00000fff));
			d_next_paddr = user_space_to_phys(
				current,(char *)dest+cnk_len
			);
			s_next_paddr = user_space_to_phys(
				current,(char *)source+cnk_len
			);
		}

		continue;

do_memcpy:

		memcpy(bus_to_virt(d_paddr),bus_to_virt(s_paddr),cnk_len);
		goto after_enq;

	}

	tx_close();

DEBUG(printk("sys_damnick_icopy_u2u_nu(): Done!\n"));

	return 0;

no_room:

	tx_wait_ring_room(MAX_FLUSH_WAIT,i);
	if (!i)
        	goto tx_failure;
	goto room_in_tx_ring;

tx_failure:

DEBUG_L(printk(
"sys_damnick_icopy_u2u_nu(): Failure (ring pos %d, tx_status %8.8lx).\n",
tx_ring_head,tx_status()
));

	tx_reset_and_restart();
	return -DAMNICK_ERR_NICFAILURE;

short_len:

DEBUG(printk(
"sys_damnick_icopy_u2u_nu(): data buffer is short (%ld bytes).  Copying...\n",
len
));

	cnk_len = len;
	d_paddr = user_space_to_phys(current,dest);
	d_next_paddr = user_space_to_phys(current,(char *)dest+cnk_len);
	if (d_next_paddr != d_paddr+cnk_len) {
		cnk_len = (unsigned short)(PAGE_SIZE-(d_paddr & 0x00000fff));
		d_next_paddr =
			user_space_to_phys(current,(char *)dest+cnk_len);
	}
	copy_from_user(bus_to_virt(d_paddr),source,cnk_len);
	len -= cnk_len;
	if (len) {

DEBUG(printk(
"sys_damnick_icopy_u2u_nu(): residual chunk (%ld bytes).  Copying...\n",
len
));

		copy_from_user(
			bus_to_virt(d_next_paddr),(char *)source+cnk_len,len
		);
	}

/*  Post notification with padding to MIN_FRAME_SIZ byte  */
	if (!tx_ring_has_room())
		goto no_room1;

room_in_tx_ring1:

	rx_enq_desc2(
		n_paddr,sizeof(unsigned long),
		(unsigned long)virt_to_bus(d_thrash),
		MIN_FRAME_SIZ-sizeof(unsigned long)
	);
	tx_enq_pkt_noheader(
		(unsigned long)virt_to_bus(tmp_notif),
		MIN_FRAME_SIZ
	);
	tx_start();
	tx_forw_tail();
#ifdef SAFE_TX
	if (tx_ring_occupancy >= STEPS_THEN_STOP)
		tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif
	return 0;

no_room1:

	tx_wait_ring_room(MAX_FLUSH_WAIT,i);
	if (!i)
		goto tx_failure;
	goto room_in_tx_ring1;

}


int damnick_icopy_k2k
(
	void *dest,
	void *source,
	unsigned long len
)
{

	unsigned long frame_size;
	unsigned long cnk_len;
	unsigned long s_paddr;
	unsigned long d_paddr;
	unsigned long s_next_paddr;
	unsigned long d_next_paddr;
	tx_common_decl_and_init;
#ifdef ALLOW_IRQS
	time_586 t_start, t_elaps;
	unsigned long flags;
#endif


DEBUG(printk("damnick_icopy_k2k(): Entering\n"));

	if (len < MIN_FRAME_SIZ)
		goto short_len;
	frame_size = tx_optimal_frag(len);
	cnk_len = lmin(len,frame_size);
	s_paddr = virt_to_bus(source);
	d_paddr = virt_to_bus(dest);
	s_next_paddr = virt_to_bus((char *)source+cnk_len);
	d_next_paddr = virt_to_bus((char *)dest+cnk_len);

#ifdef ALLOW_IRQS
	damnick_time(t_start);
#endif

	for (;;) {

		if (cnk_len < MIN_FRAME_SIZ)
			goto short_cnk;

		if (!tx_ring_has_room())
			goto no_room;

room_in_tx_ring:

DEBUG(printk(
"damnick_icopy_k2k():  cnk_len %ld, s_paddr %8.8lx, d_paddr %8.8lx\n",
cnk_len,s_paddr,d_paddr
));

		rx_enq_desc1(d_paddr,cnk_len);
		tx_enq_pkt_noheader(s_paddr,cnk_len);
		tx_start();
		tx_forw_tail();
#ifdef SAFE_TX
		if (tx_ring_occupancy >= STEPS_THEN_STOP)
			tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif

pkt_sent:

		len -= cnk_len;
		if (!len)
			break;

#ifdef ALLOW_IRQS
		damnick_time(t_elaps);
		IRQ_enable_for_5usec(t_elaps,t_start);
#endif

/*  Prepare next frame  */
		source = (void *)((char *)source+cnk_len);
		dest = (void *)((char *)dest+cnk_len);
		cnk_len = lmin(len,frame_size);
		s_paddr = s_next_paddr;
		d_paddr = d_next_paddr;
		s_next_paddr = virt_to_bus((char *)source+cnk_len);
		d_next_paddr = virt_to_bus((char *)dest+cnk_len);

		continue;

short_cnk:

DEBUG(printk(
"damnick_icopy_k2k(): data chunk is short (%ld bytes).  Copying from %8.8lx to %8.8lx\n",
cnk_len,s_paddr,d_paddr
));

		memcpy(dest,source,cnk_len);
		goto pkt_sent;

	}

	tx_close();

DEBUG(printk("damnick_icopy_k2k(): Done!\n"));

	return 0;

no_room:

	tx_wait_ring_room(MAX_FLUSH_WAIT,i);
	if (!i)
        	goto tx_failure;
	goto room_in_tx_ring;

tx_failure:

DEBUG_L(printk(
"damnick_icopy_k2k(): Failure (ring pos %d, tx_status %8.8lx).\n",
tx_ring_head,tx_status()
));

	tx_reset_and_restart();
	return -DAMNICK_ERR_NICFAILURE;

short_len:

DEBUG(printk(
"damnick_icopy_k2k(): data buffer is short (%ld bytes).  Copying...\n",
len
));

	memcpy(dest,source,len);
	return 0;

}


int damnick_icopy_k2k_nk
(
	void *dest,
	void *source,
	unsigned long len,
	unsigned long *notif,
	unsigned long val
)
{

	unsigned long *tmp_notif;
	unsigned long n_paddr;
	unsigned long frame_size;
	unsigned long cnk_len;
	unsigned long s_paddr;
	unsigned long d_paddr;
	unsigned long s_next_paddr;
	unsigned long d_next_paddr;
	tx_common_decl_and_init;
#ifdef ALLOW_IRQS
	time_586 t_start, t_elaps;
	unsigned long flags;
#endif


DEBUG(printk("damnick_icopy_k2k_nk(): Entering\n"));

	tmp_notif = &d_notif_ring[d_notif_ring_tail];
	d_notif_ring_tail = (d_notif_ring_tail+1) % DAMNICK_TX_RING_SIZE;
	*tmp_notif = val;

	n_paddr = virt_to_bus(notif);
	if (len < MIN_FRAME_SIZ)
		goto short_len;
	frame_size = tx_optimal_frag(len);
	cnk_len = lmin(len,frame_size);
	s_paddr = virt_to_bus(source);
	d_paddr = virt_to_bus(dest);
	s_next_paddr = virt_to_bus((char *)source+cnk_len);
	d_next_paddr = virt_to_bus((char *)dest+cnk_len);

#ifdef ALLOW_IRQS
	damnick_time(t_start);
#endif

	for (;;) {

		if (!tx_ring_has_room())
			goto no_room;

room_in_tx_ring:

DEBUG(printk(
"damnick_icopy_k2k_nk():  cnk_len %ld, s_paddr %8.8lx, d_paddr %8.8lx\n",
cnk_len,s_paddr,d_paddr
));

		len -= cnk_len;
		if (!len) {
			/*  Last data chunk to move. Perform notification.  */

DEBUG(printk("damnick_icopy_k2k_nk():  n_paddr %8.8lx\n",n_paddr));

			if (cnk_len < MIN_FRAME_SIZ) {
/*  DMA + notification with padding to MIN_FRAME_SIZ byte  */
				rx_enq_desc3(
					d_paddr,cnk_len,
					(unsigned long)virt_to_bus(d_thrash),
					MIN_FRAME_SIZ-cnk_len,
					n_paddr,sizeof(unsigned long)
				);
				tx_enq_pkt2_noheader(
					s_paddr,MIN_FRAME_SIZ,
					tmp_notif,sizeof(unsigned long)
				);
			}
			else {
				rx_enq_desc2(
					d_paddr,cnk_len,
					n_paddr,sizeof(unsigned long)
				);
				tx_enq_pkt2_noheader(
					s_paddr,cnk_len,
					tmp_notif,sizeof(unsigned long)
				);
			}
		}
		else {
			if (cnk_len < MIN_FRAME_SIZ)
				goto do_memcpy;
			rx_enq_desc1(d_paddr,cnk_len);
			tx_enq_pkt_noheader(s_paddr,cnk_len);
		}
		tx_start();
		tx_forw_tail();
#ifdef SAFE_TX
		if (tx_ring_occupancy >= STEPS_THEN_STOP)
			tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif

after_enq:

		if (!len)
			break;

#ifdef ALLOW_IRQS
		damnick_time(t_elaps);
		IRQ_enable_for_5usec(t_elaps,t_start);
#endif

/*  Prepare next frame  */
		source = (void *)((char *)source+cnk_len);
		dest = (void *)((char *)dest+cnk_len);
		cnk_len = lmin(len,frame_size);
		s_paddr = s_next_paddr;
		d_paddr = d_next_paddr;
		s_next_paddr = virt_to_bus((char *)source+cnk_len);
		d_next_paddr = virt_to_bus((char *)dest+cnk_len);

		continue;

do_memcpy:

		memcpy(bus_to_virt(d_paddr),bus_to_virt(s_paddr),cnk_len);
		goto after_enq;

	}

	tx_close();

DEBUG(printk("damnick_icopy_k2k_nk(): Done!\n"));

	return 0;

no_room:

	tx_wait_ring_room(MAX_FLUSH_WAIT,i);
	if (!i)
        	goto tx_failure;
	goto room_in_tx_ring;

tx_failure:

DEBUG_L(printk(
"damnick_icopy_k2k_nk(): Failure (ring pos %d, tx_status %8.8lx).\n",
tx_ring_head,tx_status()
));

	tx_reset_and_restart();
	return -DAMNICK_ERR_NICFAILURE;

short_len:

DEBUG(printk(
"damnick_icopy_k2k_nk(): data buffer is short (%ld bytes).  Copying...\n",
len
));

	memcpy(dest,source,len);

/*  Post notification with padding to MIN_FRAME_SIZ byte  */
	if (!tx_ring_has_room())
		goto no_room1;

room_in_tx_ring1:

	rx_enq_desc2(
		n_paddr,sizeof(unsigned long),
		(unsigned long)virt_to_bus(d_thrash),
		MIN_FRAME_SIZ-sizeof(unsigned long)
	);
	tx_enq_pkt_noheader(
		(unsigned long)virt_to_bus(tmp_notif),
		MIN_FRAME_SIZ
	);
	tx_start();
	tx_forw_tail();
#ifdef SAFE_TX
	if (tx_ring_occupancy >= STEPS_THEN_STOP)
		tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif
	return 0;

no_room1:

	tx_wait_ring_room(MAX_FLUSH_WAIT,i);
	if (!i)
		goto tx_failure;
	goto room_in_tx_ring1;

}


int damnick_icopy_k2u
(
	void *dest,
	void *source,
	unsigned long len
)
{

	unsigned long frame_size;
	unsigned long cnk_len;
	unsigned long s_paddr;
	unsigned long d_paddr;
	unsigned long d_next_paddr;
	tx_common_decl_and_init;
#ifdef ALLOW_IRQS
	time_586 t_start, t_elaps;
	unsigned long flags;
#endif


DEBUG(printk("damnick_icopy_k2u(): Entering\n"));

	if (len < MIN_FRAME_SIZ)
		goto short_len;
	frame_size = tx_optimal_frag(len);
	cnk_len = lmin(len,frame_size);
	s_paddr = virt_to_bus(source);
	if (dest == d_next_dest)
		d_paddr = d_next_d_paddr;
	else
		d_paddr = user_space_to_phys(current,dest);
	if (cnk_len >= (PAGE_SIZE-(d_paddr & 0x00000fff)))
		goto next_page1;
	d_next_paddr = d_paddr+cnk_len;

next_page_back1:

#ifdef ALLOW_IRQS
	damnick_time(t_start);
#endif

	for (;;) {

		if (cnk_len < MIN_FRAME_SIZ)
			goto do_memcpy;

		if (!tx_ring_has_room())
			goto no_room;

room_in_tx_ring:

DEBUG(printk(
"damnick_icopy_k2u():  cnk_len %ld, s_paddr %8.8lx, d_paddr %8.8lx\n",
cnk_len,s_paddr,d_paddr
));

		rx_enq_desc1(d_paddr,cnk_len);
		tx_enq_pkt_noheader(s_paddr,cnk_len);
		tx_start();
		tx_forw_tail();
#ifdef SAFE_TX
		if (tx_ring_occupancy >= STEPS_THEN_STOP)
			tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif

after_enq:

		len -= cnk_len;
		if (!len)
			break;

#ifdef ALLOW_IRQS
		damnick_time(t_elaps);
		IRQ_enable_for_5usec(t_elaps,t_start);
#endif

/*  Prepare next frame  */
		s_paddr += cnk_len;
		dest = (void *)((char *)dest+cnk_len);
		cnk_len = lmin(len,frame_size);
		d_paddr = d_next_paddr;
		if (cnk_len >= (PAGE_SIZE-(d_paddr & 0x00000fff)))
			goto next_page;
		d_next_paddr += cnk_len;
		continue;

next_page:

		cnk_len = (unsigned short)(PAGE_SIZE-(d_paddr & 0x00000fff));
		d_next_paddr =
			user_space_to_phys(current,(char *)dest+cnk_len);
		continue;

do_memcpy:

DEBUG(printk(
"damnick_icopy_k2u(): data chunk is short (%ld bytes).  Copying from %8.8lx to %8.8lx\n",
cnk_len,s_paddr,d_paddr
));

		memcpy(bus_to_virt(d_paddr),bus_to_virt(s_paddr),cnk_len);
		goto after_enq;

	}

	tx_close();

DEBUG(printk("damnick_icopy_k2u(): Done!\n"));

	d_next_dest = (void *)((char *)dest+cnk_len);
	d_next_d_paddr = d_next_paddr;
	return 0;

next_page1:

	cnk_len = (unsigned short)(PAGE_SIZE-(d_paddr & 0x00000fff));
	d_next_paddr = user_space_to_phys(current,(char *)dest+cnk_len);
	goto next_page_back1;

no_room:

	tx_wait_ring_room(MAX_FLUSH_WAIT,i);
	if (!i)
        	goto tx_failure;
	goto room_in_tx_ring;

tx_failure:

DEBUG_L(printk(
"damnick_icopy_k2u(): Failure (ring pos %d, tx_status %8.8lx).\n",
tx_ring_head,tx_status()
));

	tx_reset_and_restart();
	return -DAMNICK_ERR_NICFAILURE;

short_len:

DEBUG(printk(
"damnick_icopy_k2u(): data buffer is short (%ld bytes).  Copying...\n",
len
));

	copy_to_user(dest,source,len);
	return 0;

}


int damnick_icopy_k2u_nu
(
	void *dest,
	void *source,
	unsigned long len,
	unsigned long *notif,
	unsigned long val
)
{

	unsigned long *tmp_notif;
	unsigned long n_paddr;
	unsigned long frame_size;
	unsigned long cnk_len;
	unsigned long s_paddr;
	unsigned long d_paddr;
	unsigned long d_next_paddr;
	tx_common_decl_and_init;
#ifdef ALLOW_IRQS
	time_586 t_start, t_elaps;
	unsigned long flags;
#endif


DEBUG(printk("damnick_icopy_k2u_nu(): Entering\n"));

	tmp_notif = &d_notif_ring[d_notif_ring_tail];
	d_notif_ring_tail = (d_notif_ring_tail+1) % DAMNICK_TX_RING_SIZE;
	*tmp_notif = val;

	n_paddr = user_space_to_phys(current,notif);
	if (len < MIN_FRAME_SIZ)
		goto short_len;
	frame_size = tx_optimal_frag(len);
	cnk_len = lmin(len,frame_size);
	s_paddr = virt_to_bus(source);
	if (dest == d_next_dest)
		d_paddr = d_next_d_paddr;
	else
		d_paddr = user_space_to_phys(current,dest);
	if (cnk_len >= (PAGE_SIZE-(d_paddr & 0x00000fff)))
		goto next_page1;
	d_next_paddr = d_paddr+cnk_len;

next_page_back1:

#ifdef ALLOW_IRQS
	damnick_time(t_start);
#endif

	for (;;) {

		if (!tx_ring_has_room())
			goto no_room;

room_in_tx_ring:

DEBUG(printk(
"damnick_icopy_k2u_nu():  cnk_len %ld, s_paddr %8.8lx, d_paddr %8.8lx\n",
cnk_len,s_paddr,d_paddr
));

		len -= cnk_len;
		if (!len) {
			/*  Last data chunk to move. Perform notification.  */

DEBUG(printk("damnick_icopy_k2u_nu():  n_paddr %8.8lx\n",n_paddr));

			if (cnk_len < MIN_FRAME_SIZ) {
/*  DMA + notification with padding to MIN_FRAME_SIZ byte  */
				rx_enq_desc3(
					d_paddr,cnk_len,
					(unsigned long)virt_to_bus(d_thrash),
					MIN_FRAME_SIZ-cnk_len,
					n_paddr,sizeof(unsigned long)
				);
				tx_enq_pkt2_noheader(
					s_paddr,MIN_FRAME_SIZ,
					tmp_notif,sizeof(unsigned long)
				);
			}
			else {
				rx_enq_desc2(
					d_paddr,cnk_len,
					n_paddr,sizeof(unsigned long)
				);
				tx_enq_pkt2_noheader(
					s_paddr,cnk_len,
					tmp_notif,sizeof(unsigned long)
				);
			}
		}
		else {
			if (cnk_len < MIN_FRAME_SIZ)
				goto do_memcpy;
			rx_enq_desc1(d_paddr,cnk_len);
			tx_enq_pkt_noheader(s_paddr,cnk_len);
		}
		tx_start();
		tx_forw_tail();
#ifdef SAFE_TX
		if (tx_ring_occupancy >= STEPS_THEN_STOP)
			tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif

after_enq:

		if (!len)
			break;

#ifdef ALLOW_IRQS
		damnick_time(t_elaps);
		IRQ_enable_for_5usec(t_elaps,t_start);
#endif

/*  Prepare next frame  */
		s_paddr += cnk_len;
		dest = (void *)((char *)dest+cnk_len);
		cnk_len = lmin(len,frame_size);
		d_paddr = d_next_paddr;
		if (cnk_len >= (PAGE_SIZE-(d_paddr & 0x00000fff)))
			goto next_page;
		d_next_paddr += cnk_len;
		continue;

next_page:

		cnk_len = (unsigned short)(PAGE_SIZE-(d_paddr & 0x00000fff));
		d_next_paddr =
			user_space_to_phys(current,(char *)dest+cnk_len);
		continue;

do_memcpy:

		memcpy(bus_to_virt(d_paddr),bus_to_virt(s_paddr),cnk_len);
		goto after_enq;

	}

	tx_close();

DEBUG(printk("damnick_icopy_k2u_nu(): Done!\n"));

	d_next_dest = (void *)((char *)dest+cnk_len);
	d_next_d_paddr = d_next_paddr;
	return 0;

next_page1:

	cnk_len = (unsigned short)(PAGE_SIZE-(d_paddr & 0x00000fff));
	d_next_paddr = user_space_to_phys(current,(char *)dest+cnk_len);
	goto next_page_back1;

no_room:

	tx_wait_ring_room(MAX_FLUSH_WAIT,i);
	if (!i)
		goto tx_failure;
	goto room_in_tx_ring;

tx_failure:

DEBUG_L(printk(
"damnick_icopy_k2u_nu(): Failure (ring pos %d, tx_status %8.8lx).\n",
tx_ring_head,tx_status()
));

	tx_reset_and_restart();
	return -DAMNICK_ERR_NICFAILURE;

short_len:

DEBUG(printk(
"damnick_icopy_k2u_nu(): data buffer is short (%ld bytes).  Copying...\n",
len
));

	copy_to_user(dest,source,len);

/*  Post notification with padding to MIN_FRAME_SIZ byte  */
	if (!tx_ring_has_room())
		goto no_room1;

room_in_tx_ring1:

	rx_enq_desc2(
		n_paddr,sizeof(unsigned long),
		(unsigned long)virt_to_bus(d_thrash),
		MIN_FRAME_SIZ-sizeof(unsigned long)
	);
	tx_enq_pkt_noheader(
		(unsigned long)virt_to_bus(tmp_notif),
		MIN_FRAME_SIZ
	);
	tx_start();
	tx_forw_tail();
#ifdef SAFE_TX
	if (tx_ring_occupancy >= STEPS_THEN_STOP)
		tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif
	return 0;

no_room1:

	tx_wait_ring_room(MAX_FLUSH_WAIT,i);
	if (!i)
		goto tx_failure;
	goto room_in_tx_ring1;

}


/*  This only performs a notification, no memcpy.  */
int damnick_writel_u
(
	unsigned long *notif,
	unsigned long val
)
{

	unsigned long *tmp_notif;
	unsigned long n_paddr;
	tx_common_decl_and_init;


DEBUG(printk("damnick_writel_u(): Entering\n"));

	tmp_notif = &d_notif_ring[d_notif_ring_tail];
	d_notif_ring_tail = (d_notif_ring_tail+1) % DAMNICK_TX_RING_SIZE;
	*tmp_notif = val;

	n_paddr = user_space_to_phys(current,notif);

/*  Post notification with padding to MIN_FRAME_SIZ byte  */
	if (!tx_ring_has_room())
		goto no_room;

room_in_tx_ring:

	rx_enq_desc2(
		n_paddr,sizeof(unsigned long),
		(unsigned long)virt_to_bus(d_thrash),
		MIN_FRAME_SIZ-sizeof(unsigned long)
	);
	tx_enq_pkt_noheader(
		(unsigned long)virt_to_bus(tmp_notif),
		MIN_FRAME_SIZ
	);
	tx_start();
	tx_forw_tail();
#ifdef SAFE_TX
	if (tx_ring_occupancy >= STEPS_THEN_STOP)
		tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif
	return 0;

no_room:

	tx_wait_ring_room(MAX_FLUSH_WAIT,i);
	if (!i)
		goto tx_failure;
	goto room_in_tx_ring;

tx_failure:

DEBUG_L(printk(
"damnick_writel_u(): Failure (ring pos %d, tx_status %8.8lx).\n",
tx_ring_head,tx_status()
));

	tx_reset_and_restart();
	return -DAMNICK_ERR_NICFAILURE;

}


/************************************************************************
 *	Test routines							*
 ************************************************************************/


/*  This is to allow invocation of kernel-to-user copy from a user program  */
/*
 *  NOTE:
 *  copying more than DUMMY_RING_SIZE*4096 bytes seemingly causes the DAMNICK
 *  device to miss some data (as a result of overrun on dummy_ring[]??).
 */
#undef USE_MEMCPY
#define DUMMY_RING_SIZE 256
unsigned long dummy_ring_tail;
unsigned char *dummy_ring[DUMMY_RING_SIZE];

int sys_damnick_icopy_k2u
(
	void *dest,
	unsigned long len
)
{

	int i,j,k=0;
#ifdef MEASURE_ICOPY
	time_586 a,b;
#endif


	if (dummy_ring[0] == NULL) {
		for (i=0;i<DUMMY_RING_SIZE;i++) {
			dummy_ring[i] =
				(unsigned char *)kmalloc(4096,GFP_KERNEL);
			if (dummy_ring[i])
				for (j=0;j<4096;j++) dummy_ring[i][j] = k++;
			else {
				printk("KMALLOC FAILED!!!!!!!!!!!!!!!!\n");
				return 0;
			}
		}
		dummy_ring_tail = 0;
	}

#ifdef MEASURE_ICOPY
	damnick_time(a);
#endif
	for (i=0;i<len/4096;i++,dest+=4096) {
#ifdef USE_MEMCPY
		copy_to_user(dest,dummy_ring[dummy_ring_tail],4096);
#else
		damnick_icopy_k2u(dest,dummy_ring[dummy_ring_tail],4096);
#endif
		dummy_ring_tail = (dummy_ring_tail+1) % DUMMY_RING_SIZE;
	}
	damnick_icopy_k2u(dest,dummy_ring[dummy_ring_tail],len%4096);
	dummy_ring_tail = (dummy_ring_tail+1) % DUMMY_RING_SIZE;
#ifdef MEASURE_ICOPY
	damnick_time(b);
	icopy_counter++;
	icopy_amount += damnick_time_diff_ticks(b,a);
#endif
	return 0;
}


int sys_damnick_icopy_k2u_nu
(
	void *dest,
	unsigned long len,
	unsigned long *notif,
	unsigned long val
)
{

	int i,j,k=0;
#ifdef MEASURE_ICOPY
	time_586 a,b;
#endif
#ifdef MEASURE_COPYCOMPL
	unsigned long tmp;
	int guard;
	time_586 c;
#endif


	if (dummy_ring[0] == NULL) {
		for (i=0;i<DUMMY_RING_SIZE;i++) {
			dummy_ring[i] =
				(unsigned char *)kmalloc(4096,GFP_KERNEL);
			if (dummy_ring[i])
				for (j=0;j<4096;j++) dummy_ring[i][j] = k++;
			else {
				printk("KMALLOC FAILED!!!!!!!!!!!!!!!!\n");
				return 0;
			}
		}
		dummy_ring_tail = 0;
	}
#ifdef MEASURE_ICOPY
	damnick_time(a);
#endif
	for (i=0;i<len/4096;i++,dest+=4096) {
#ifdef USE_MEMCPY
		copy_to_user(dest,dummy_ring[dummy_ring_tail],4096);
#else
		damnick_icopy_k2u(dest,dummy_ring[dummy_ring_tail],4096);
#endif
		dummy_ring_tail = (dummy_ring_tail+1) % DUMMY_RING_SIZE;
	}
#ifdef USE_MEMCPY
	copy_to_user(dest,dummy_ring[dummy_ring_tail],len%4096);
	copy_to_user(notif,&val,sizeof(unsigned long));
#else
	damnick_icopy_k2u_nu(dest,dummy_ring[dummy_ring_tail],len%4096,
		notif,val);
#endif
	dummy_ring_tail = (dummy_ring_tail+1) % DUMMY_RING_SIZE;
#ifdef MEASURE_ICOPY
	damnick_time(b);
	icopy_counter++;
	icopy_amount += damnick_time_diff_ticks(b,a);
#endif
#ifdef MEASURE_COPYCOMPL
	for (guard=10000000;guard;guard--) {
		copy_from_user(&tmp,notif,sizeof(unsigned long));
		if (tmp == val)
			break;
	}
	damnick_time(c);
	ccopy_counter++;
	ccopy_amount += damnick_time_diff_ticks(c,a);
	ccopy_bytes += len;
#endif
	return 0;
}


/************************************************************************
 *	Debug routines							*
 ************************************************************************/


#ifdef DRIVER_DEBUG

typedef union {
	unsigned long ll;
	unsigned short ss[2];
} l2s;

void rx_damnick
(struct net_device *dev)
{

l2s *h;
rx_common_decl_and_init;


DEBUG(printk("rx_damnick(): entering...\n"));

rx_read_pkt_info();

if (rx_err_pktnotavail())
 goto bogus_inv;

for(;;) {

DEBUG(printk(
"rx_damnick(): loop: rx_ring_head %d, rx_ring_tail %d, rx_status %4.4x\n",
rx_ring_head,rx_ring_tail,rx_status()
));

 rx_header(h);

DEBUG(printk(
"rx_damnick(): h=%8.8lx, h0 %8.8lx, h1 %8.8lx, h2 %8.8lx, h3 %8.8lx, h4 %8.8lx\n",
(unsigned long)h,
h[0].ll, h[1].ll, h[2].ll, h[3].ll, h[4].ll
));

 discard_pkt();

 rx_read_pkt_info();

 if (rx_err_pktnotavail())
  break;

DEBUG(printk("rx_damnick(): continuing loop...\n"));

};

DEBUG(printk(
"rx_damnick(): exiting: rx_ring_head %d, rx_ring_tail %d\n",
 rx_ring_head,rx_ring_tail
));

return;

/*  Here all the unusual branches.  */

bogus_inv:

DEBUG(printk(
"rx_damnick(): bogus invocation (rx_status %8.8x)\n",rx_status()
));

return;

}


#define REPEAT_POLL 10000  /*  About 457 usec polling on a Pentium 133 MHz  */

int sys_damnick_poll
(void)
{

	poll_common_decl_and_init;
	int i;


	for (i=REPEAT_POLL;i;i--) {
		rx_read_pkt_info();
		if (!(rx_err_pktnotavail())) {
/*  There are packets  */
			synchronize_irq();
			rx_damnick(damnick_device);
			clear_rx_irqs();
			return 0;
		}
	}
	return 1;

}

#else

int sys_damnick_poll (void) {return 0;}

#endif /* Debug routines */


/************************************************************************
 *	Other support routines						*
 ************************************************************************/


#include "sys_damnick_mem.c"
#include "sys_damnick_timers.c"
