/*  These are for time measures  */

/*
 *  Adjust this to the clock freq. of the machine where time intervals
 *  are computed.
 */
#ifdef DAMNICK_USERLEVEL
extern unsigned long damnick_clock;
#define CLOCK damnick_clock
#else
extern unsigned long cpu_khz;
#define CLOCK (cpu_khz/1000)
#endif

typedef struct {
        unsigned long hi;
        unsigned long lo;
} time_586;

#define damnick_time(x) \
__asm__ __volatile__("rdtsc" : "=d" (x.hi), "=a" (x.lo))

#ifdef DAMNICK_USERLEVEL
static inline double damnick_time_diff(time_586 b, time_586 a) {
        double db,da,res;
        db = (double)b.hi*(double)(1<<16)*(double)(1<<16)+(double)b.lo;
        da = (double)a.hi*(double)(1<<16)*(double)(1<<16)+(double)a.lo;
        if (db < da)
res = ((double)(1<<16)*(double)(1<<16)*(double)(1<<16)*(double)(1<<16)+db-da)
/(double)CLOCK;
	else
res = (db-da)/(double)CLOCK;
	return res;
}
#else
/*
 *  We cannot use floating point inside kernel, because FPU regs are
 *  not saved on IRQ.  So we use integer arithmetics, and get the
 *  timings in clock ticks rather than usecs.
 */
static inline unsigned long long damnick_time_diff_ticks
(time_586 b, time_586 a)
{
	unsigned long long bb,aa;
	bb = (((unsigned long long)b.hi)<<32) + (unsigned long long)b.lo;
	aa = (((unsigned long long)a.hi)<<32) + (unsigned long long)a.lo;
	return bb - aa;
}
#endif
