#include "libdamnick.h"
#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <asm/page.h>	/*  "#define PAGE_SIZE ..."  */
#include <signal.h>

/*
 *  Clock rate of the underlying CPU.
 *  Initialized by damnick_init().
 */
unsigned long damnick_clock;


int damnick_icopy
(void *dest, void *source, unsigned long len)
{
	long res;
	__asm__ __volatile__ (
		"int $0x82"
		:
		"=a"(res)
		:
		"0"(DAMNICK_ICOPY_U2U),
		"b"((long)dest),
		"c"((long)source),
		"d"((long)len)
	);
	return (int)res;
}


int damnick_icopy_notif
(
	void *dest,
	void *source,
	unsigned long len,
	unsigned long *notif_sem,
	unsigned long val
)
{
	long res;
	__asm__ __volatile__ (
		"int $0x82"
		:
		"=a"(res)
		:
		"0"(DAMNICK_ICOPY_U2U_NU),
		"b"((long)dest),
		"c"((long)source),
		"d"((long)len),
		"S"((long)notif_sem),
		"D"((long)val)
	);
	return (int)res;
}


int damnick_icopy_k2u
(void *dest, unsigned long len)
{
	long res;
	__asm__ __volatile__ (
		"int $0x82"
		:
		"=a"(res)
		:
		"0"(DAMNICK_ICOPY_K2U),
		"b"((long)dest),
		"c"((long)len)
	);
	return (int)res;
}


int damnick_icopy_k2u_notif
(void *dest, unsigned long len, unsigned long *notif_sem, unsigned long val)
{
	long res;
	__asm__ __volatile__ (
		"int $0x82"
		:
		"=a"(res)
		:
		"0"(DAMNICK_ICOPY_K2U_NU),
		"b"((long)dest),
		"c"((long)len),
		"d"((long)notif_sem),
		"S"((long)val)
	);
	return (int)res;
}


#include <asm/unistd.h>
int damnick_sched_yield
(void)
{
	long res;
	__asm__ __volatile__ ("int $0x80"
		:
		"=a"(res)
		:
		"0"(__NR_sched_yield)
	);
	return (int)res;
}


static int damnick_poll
(void)
{
	long res;
	__asm__ __volatile__ (
		"int $0x82"
		:
		"=a"(res)
		:
		"0"(DAMNICK_POLL)
	);
	return (int)res;
}


int copy_wait
(void)
{
	int i=10000;
	for (;i;i--)
		damnick_poll();
	return 0;
}


int damnick_init
(void)
{
	long res;
	__asm__ __volatile__ ("int $0x82"
		:
		"=a"(res)
		:
		"0"(DAMNICK_INIT)
	);
	damnick_clock = (unsigned long)res;
	return (int)res;
}


static volatile unsigned char defeat_cc_opt;

static inline int
damnick_mlock_lib
(void *buffer, unsigned long len,int flag)
{
	/*
	 *  Lock/Unlock a contiguous region of virtual memory into RAM.
	 *  This is good for both data structures (flag=1) and
	 *  receiver handlers (flag=0).
	 *  If flag=1, pre-fetch the memory buffer.
	 */
	unsigned char *p,*end;
	long res;
	if (buffer == NULL) return 0;
	p = (unsigned char *)buffer;
	end = (unsigned char *)
		(((unsigned long)(p+len-1+PAGE_SIZE))&~(PAGE_SIZE-1));

	/*  Lock buffer into RAM  */
        __asm__ __volatile__ ("int $0x82"
		:
		"=a"(res)
		:
		"0"(DAMNICK_MLOCK),
		"b"((long)buffer),
		"c"((long)len)
	);
	if (res) {
		printf("damnick_mlock(): ");
		perror("could not lock buffer:  ");
		return -DAMNICK_ERR_NOLOCK;
	}
	if (flag)
		/*
		 *  Buffer is for data:
		 *  Pre-fetch its logical pages by touching (on write)
		 *  one byte per page.
		 *  Must defeat compiler optimizations here...
		 */
		for (;p<end;p+=PAGE_SIZE) {
			defeat_cc_opt = *p;
			*p = defeat_cc_opt;
		}
	else
		/*
		 *  Buffer is for data:
		 *  Pre-fetch its logical pages by touching (on read)
		 *  one byte per page.
		 *  Must defeat compiler optimizations here...
		 */
		for (;p<end;p+=PAGE_SIZE)
			defeat_cc_opt += *p;
	return 0;
}


int damnick_mlock(void *buffer, unsigned long len)
{
	/*
	 *  Lock a data structure spanning a contiguous region.
	 *  This must be used in advance if the data structure is to
	 *  receive data through DAMNICK.
	 */
	return damnick_mlock_lib(buffer,len,1);
}


int damnick_munlock(void *buffer, unsigned long len)
{
	long res;

	/*
	 *  Unlock anything previously locked by damnick_mlock() or
	 *  damnick_mlock_funct().
	 */
/*
        __asm__ __volatile__ ("int $0x80"
		:
		"=a"(res)
		:
		"0"(LINUX_MUNLOCK),
		"b"((long)buffer),
		"c"((long)len)
	);
*/
        __asm__ __volatile__ ("int $0x82"
		:
		"=a"(res)
		:
		"0"(DAMNICK_MUNLOCK),
		"b"((long)buffer),
		"c"((long)len)
	);

	if (res) {
		printf("damnick_munlock(): ");
		perror("could not unlock:  ");
		return -DAMNICK_ERR_NOLOCK;
	}

	return 0;
}


int damnick_munlockall(void)
{
	/*  Unlock anything previously locked.  */
	long res;

        __asm__ __volatile__ ("int $0x82"
		:
		"=a"(res)
		:
		"0"(DAMNICK_MUNLOCKALL)
	);

	if (res) {
		printf("damnick_munlockall(): ");
		perror("could not unlock:  ");
		return -DAMNICK_ERR_NOLOCK;
	}
	
	return 0;
}


int damnick_timers(time_586 from_t, time_586 to_t)
{
	/*
	 *  Force printout of time measurements taken at kernel level.
	 */
        long res;
        __asm__ __volatile__ ("int $0x82"
		:
		"=a"(res)
		:
                "0"(DAMNICK_TIMERS),
                "b"((long)from_t.lo),
                "c"((long)from_t.hi),
                "d"((long)to_t.lo),
                "S"((long)to_t.hi)
        );
        return (int)res;
}


void damnick_busywait_usec
(unsigned long n)
{
	time_586 a,b;
	damnick_time(a);
	for (;;) {
		damnick_time(b);
		if ((unsigned long)damnick_time_diff(b,a) > n)
			break;
	}
}
