/*
 *  Format of GAMMA frame headers: 24 bytes.
 *
 *  i.		eth receiver address (leading 6 bytes)
 *  ii.		eth sender address (6 bytes)
 *  iii.	pkt type:
 *		ETH_GAMMA,ETH_GAMMA_RETX,ETH_GAMMA_KERNEL,ETH_GAMMA_KERNEL_RETX
 *		(1 byte)
 *  iv.		par_pid of receiver process (1 byte)
 *  v.		sequence frame number, relative to sender node (see ix.)
 *		(1 byte)
 *  vi.		GAMMA normal frame:  "procnum" of receiver process
 *		GAMMA kernel frame:  kind of kernel msg
 *		(1 byte)
 *  vii.	2 bytes:
 *		bits 0..12:	payload length in bytes (max 8191)
 *		bit 13:		sender requests credit renewal (flow control)
 *		bit 14:		is this the first chunk of a new msg?
 *		bit 15:		is this the last chunk of a msg?
 *  viii.	number of input port of receiver process
 *		(2 byte, so max 65536 ports per process instance; actually
 *		much fewer, for memory occupancy reasons; see MAX_PORT_NUMBER)
 *  ix.		physical no. of sender node (2 byte, so max 65536 nodes;
 *		actually much fewer, for memory occupancy reasons; see
 *		MAX_NUM_NODES)
 *  x.		unused (2 bytes)
 */
#define HEADER_SIZE 24

#define ETH_GAMMA 0x2			/*  GAMMA regular pkt.  */
#define ETH_GAMMA_KERNEL 0x3		/*  GAMMA kernel pkt.  */
#define ETH_GAMMA_RETX 0x4		/*  retransmitted GAMMA pkt.  */
#define ETH_GAMMA_KERNEL_RETX 0x5	/*  retransmitted GAMMA kernel pkt.  */

#define CNK_LEN_MASK 0x1fff
#define CREQ_MASK 0x2000
#define MSG_1ST_FRAME_MASK 0x4000
#define MSG_COMPLETE_MASK 0x8000
#define make_send_header\
( \
	dest_phys_node,par_pid,procnum_or_kernel_kind,port,kind,last_frame, \
	len, frame_no,h \
) \
{ \
	unsigned short *usp, cnk_len; \
	unsigned char *ucp; \
	cnk_len = ((len) + 3) & ~3; \
	if ((dest_phys_node) == BROADCAST) { \
		(h)[0].ll = 0xffffffff; \
		(h)[1].ss[0] = 0xffff; \
	} \
	else { \
		(h)[0].ll = eth_table[dest_phys_node].addrl; \
		(h)[1].ss[0] = eth_table[dest_phys_node].addrw; \
	} \
	usp = (unsigned short *)&(eth_table[phys_whoami].addrl); \
	(h)[1].ss[1] = *(usp++); \
	(h)[2].ss[0] = *usp; \
        (h)[2].ss[1] = eth_table[phys_whoami].addrw; \
	ucp = (unsigned char *)&((h)[3]); \
	*(ucp++) = kind; \
	*(ucp++) = par_pid; \
	*(ucp++) = (unsigned char)(frame_no); \
	*ucp = procnum_or_kernel_kind; \
	(h)[4].ss[0] = (last_frame) ? \
		cnk_len | MSG_1ST_FRAME_MASK | MSG_COMPLETE_MASK : \
		cnk_len | MSG_1ST_FRAME_MASK; \
	usp = (unsigned short *)&((h)[4].ss[1]); \
	*usp++ = port; \
	*usp = phys_whoami; \
}
#define set_send_header_frmtype(h,n) (*(unsigned char *)&((h)[3].ll)) = (n)
#define fill_send_header_cnklen(h,word) (h)[4].ss[0] = word
#define set_send_header_creq(h) (h)[4].ss[0] |= CREQ_MASK
#define set_send_header_frameno(h,n) (*(unsigned char *)&((h)[3].ss[1])) = (n)
#define read_send_header_frameno(h) (*(unsigned char *)&((h)[3].ss[1]))
#define rec_header_frmtype(h) ((unsigned char)(h)[3].ll)
#define rec_header_1st_frame(h) ((h)[4].ss[0] & MSG_1ST_FRAME_MASK)
#define rec_header_msg_complete(h) ((h)[4].ss[0] & MSG_COMPLETE_MASK)
#define rec_header_parpid_port(h,par_pid,port,procnum) \
{ \
	unsigned char *ucp; \
	ucp = (unsigned char *)&((h)[3]); \
	par_pid = *++ucp; \
	ucp += 2; \
	procnum = *ucp; \
	port = *(unsigned short *)&((h)[4].ss[1]); \
}
#define rec_header_cnk_len(h) ((h)[4].ss[0] & CNK_LEN_MASK)
#define rec_header_frame_no(h) ((unsigned char)(h)[3].ss[1])
#define rec_header_phys_sender(h) ((unsigned short)(h)[5].ss[0])
#define rec_header_creq(h) ((h)[4].ss[0] & CREQ_MASK)
#define rec_header_clean_creq(h) (h)[4].ss[0] &= ~(unsigned short)CREQ_MASK
#define rec_is_bcast(h) (((h)[0].ll == 0xffffffff) && ((h)[1].ss[0] == 0xffff))
#define rec_header_set_procnum(h,n) { \
	unsigned char *ucp; \
	ucp = (unsigned char *)&((h)[3].ss[1]); \
	*++ucp = n; \
}
