/***************************************************************************
	NIC-independent macros for the GAMMA protocol
	Read COPYRIGHT file in this directory
****************************************************************************/


#define gmin(a,b) ((a)<=(b)?(a):(b))
#define gmax(a,b) ((a)>=(b)?(a):(b))


#ifdef MORE_INSTANCES_ON_SAME_NODE
#define DELAY_FOR_SMP busywait(20)
#else
#define DELAY_FOR_SMP
#endif

#ifdef DRIVER_DEBUG_EVERYTHING
#define DEBUG_E(a) {a;DELAY_FOR_SMP;}
#define DEBUG(a) {a;DELAY_FOR_SMP;}
#define DEBUG_F(a) {a;DELAY_FOR_SMP;}
#define DEBUG_L(a) {a;DELAY_FOR_SMP;}
#else
#ifdef DRIVER_DEBUG
#define DEBUG_E(a)
#define DEBUG(a) {a;DELAY_FOR_SMP;}
#define DEBUG_F(a) {a;DELAY_FOR_SMP;}
#define DEBUG_L(a) {a;DELAY_FOR_SMP;}
#else
#ifdef DRIVER_DEBUG_FLWCTL_RETX
#define DEBUG_E(a)
#define DEBUG(a)
#define DEBUG_F(a) {a;DELAY_FOR_SMP;}
#define DEBUG_L(a) {a;DELAY_FOR_SMP;}
#else
#ifdef DRIVER_DEBUG_LIGHT
#define DEBUG_E(a)
#define DEBUG(a)
#define DEBUG_F(a)
#define DEBUG_L(a) {a;DELAY_FOR_SMP;}
#else
#define DEBUG_E(a)
#define DEBUG(a)
#define DEBUG_F(a)
#define DEBUG_L(a)
#endif
#endif
#endif
#endif


#define memcpy_tofs(a,b,c)   copy_to_user(a,b,c)
#define memcpy_fromfs(a,b,c) copy_from_user(a,b,c)


#define user_space_to_phys(P,addr) \
(unsigned long)virt_to_phys( \
	(void *)( \
		(unsigned long)page_address(pte_page( \
			*pte_offset( \
				pmd_offset( \
					pgd_offset( \
						(P)->mm, \
						(unsigned long)(addr) \
					), \
					(unsigned long)(addr) \
				), \
				(unsigned long)(addr) \
			) \
		)) \
		| ((unsigned long)(addr) & 0x00000fff) \
	) \
)


#define recv_payload_1st_chunk(dest,skip,len) \
memcpy_tofs((void *)(dest),(void *)(((u32)h)+skip),len)


#define recv_payload_2nd_chunk(dest,skip1,skip2,len) \
memcpy_tofs((void *)(dest),(void *)(((u32)h)+skip1+skip2),len)


#define invoke_rec_handler(handler_entry,par_pid,port,procnum,msglen) \
{ \
unsigned long a,b,c,d,S,D; \
handler_entry.offset = \
(unsigned long)(par_proc_table[(par_pid)]->inst_table[procnum] \
.handler_rec_dispatcher); \
__asm__ __volatile__ ( "lcall (%4)" \
	: \
	"=a" (a), \
	"=b" (b), \
	"=c" (c), \
	"=d" (d), \
	"=S" (S), \
	"=D" (D) \
	: \
	"4" (&(handler_entry)), \
	"1" ((unsigned long)(port)), \
	"2" ((unsigned long)(msglen)) \
); \
}


#define invoke_err_handler(handler_entry,par_pid,port,procnum,errno) \
{ \
unsigned long a,b,c,d,S,D; \
handler_entry.offset = \
(unsigned long)(par_proc_table[(par_pid)]->inst_table[procnum] \
.handler_err_dispatcher); \
__asm__ __volatile__ ("lcall (%4)" \
	: \
	"=a" (a), \
	"=b" (b), \
	"=c" (c), \
	"=d" (d), \
	"=S" (S), \
	"=D" (D) \
	: \
	"4" (&(handler_entry)), \
	"1" ((unsigned long)(port)), \
	"2" ((unsigned long)errno) \
); \
}


/*
 *  NOTE: most of what follows is NOT thread safe on SMP machines.
 *  Caller should raise spinlock.
 */


/*
 *  Shift user-space rec buffer's current ptr forward.
 *  Check for room availability.
 *  If not enough room in rec buffer, truncate the msg chunk
 *  by setting cnk_len equal to the available room.
 *  Decrement the available room by cnk_len.
 *  This macro is to prepare the download of a fresh packet into final dest.
 *  The fresh packet will be stored starting from address "rx_buffer".
 *  Check whether user-space or kernel-space buffer.
 */
#define update_ptr_to_user_recv_buffer(par_pid,port,procnum,len,rx_buffer) \
{ \
rx_buffer = (unsigned long) \
par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_progress; \
if \
(par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_room < len) \
len = par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_room; \
par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_room \
-= len; \
par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_progress \
+= len; \
}


/*
 *  Get the pointer to current user-level recv buffer.
 */
#define get_ptr_to_user_recv_buffer(par_pid,port,procnum,rx_buffer) \
{ \
rx_buffer = (unsigned long) \
par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_progress; \
}


/*  Notify a message arrival to the port semaphore.  */
#define notify_arrival(par_pid,port,procnum) \
{ \
unsigned long temp; \
memcpy_fromfs( \
	&temp, \
	((unsigned long *) \
(par_proc_table[par_pid]->inst_table[procnum].sem1_table)+ \
par_proc_table[par_pid]->inst_table[procnum].port_table[port].semaphore), \
	sizeof(temp) \
); \
temp++; \
memcpy_tofs( \
	((unsigned long *) \
(par_proc_table[par_pid]->inst_table[procnum].sem1_table)+ \
par_proc_table[par_pid]->inst_table[procnum].port_table[port].semaphore), \
	&temp, \
	sizeof(temp) \
); \
}


#ifdef USE_DAMNICK
/*  Pre-notify a message arrival to the port semaphore.  */
#define pre_notify_arrival(par_pid,port,procnum,pre_notif_val) \
{ \
memcpy_fromfs( \
	&pre_notif_val, \
	((unsigned long *) \
(par_proc_table[par_pid]->inst_table[procnum].pre_sem_table)+ \
par_proc_table[par_pid]->inst_table[procnum].port_table[port].semaphore), \
	sizeof(pre_notif_val) \
); \
pre_notif_val++; \
memcpy_tofs( \
	((unsigned long *) \
(par_proc_table[par_pid]->inst_table[procnum].pre_sem_table)+ \
par_proc_table[par_pid]->inst_table[procnum].port_table[port].semaphore), \
	&pre_notif_val, \
	sizeof(pre_notif_val) \
); \
}
#endif


/*  Notify a message error to the port error semaphore.  */
#define notify_error(par_pid,port,procnum) \
{ \
unsigned long temp; \
memcpy_fromfs( \
	&temp, \
	((unsigned long *) \
(par_proc_table[par_pid]->inst_table[procnum].semerr_table)+ \
par_proc_table[par_pid]->inst_table[procnum].port_table[port].semaphore), \
	sizeof(temp) \
); \
temp++; \
memcpy_tofs( \
	((unsigned long *) \
(par_proc_table[par_pid]->inst_table[procnum].semerr_table)+ \
par_proc_table[par_pid]->inst_table[procnum].port_table[port].semaphore), \
	&temp, \
	sizeof(temp) \
); \
}


#define reset_current_ptr(par_pid,port,procnum) \
{ \
par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_progress \
= par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_buffer; \
par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_room \
= par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_size; \
}


#define \
check_available_room_and_possibly_reset_current_ptr(par_pid,port,procnum) \
{ \
if (!(par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_room)) \
{ \
par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_progress \
= par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_buffer; \
par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_room \
= par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_size; \
} \
}


#define read_credit(node,credit) \
for (;;) { \
	unsigned long iii; \
	if (node != BROADCAST) { \
		credit = local_credit[node]; \
		break; \
	} \
	credit = 999999; \
	for (iii=0; iii<total_num_nodes; iii++) \
		if (iii != phys_whoami) \
			if (local_credit[iii] < credit) \
				credit = local_credit[iii]; \
	break; \
}


#define dec_credit(node,amount) \
for (;;) { \
	unsigned long iii; \
	if (node != BROADCAST) { \
		local_credit[node] -= amount; \
		break; \
	} \
	for (iii=0; iii<total_num_nodes; iii++) \
		if (iii != phys_whoami) \
			local_credit[iii] -= amount; \
	break; \
}


#ifdef MORE_INSTANCES_ON_SAME_NODE
#define TX_SPIN_LOCK spin_lock(&gamma_tx_lock)
#define TX_SPIN_UNLOCK spin_unlock(&gamma_tx_lock)
#define CREDIT_SPIN_TRYLOCK spin_trylock(&gamma_cr_lock)
#define CREDIT_SPIN_UNLOCK spin_unlock(&gamma_cr_lock)
#else
#define TX_SPIN_LOCK
#define TX_SPIN_UNLOCK
#define CREDIT_SPIN_TRYLOCK 1
#define CREDIT_SPIN_UNLOCK
#endif

#define wait_credit_refill(i,dest,par_pid,procnum) \
{ \
time_586 a,b; \
i = 1; \
gamma_time(a); \
if (CREDIT_SPIN_TRYLOCK) \
for (;;) { \
	if (dest != BROADCAST) { \
		if (local_credit1[dest]) { \
			local_credit[dest] = local_credit1[dest]; \
			local_credit1[dest] = 0; \
			credit = local_credit[dest]; \
			curr_credit = credit; \
DEBUG_F( \
if (credit) { \
unsigned long i; \
printk("credit refilled: "); \
for (i=0;i<total_num_nodes;i++) \
	printk("%ld ", local_credit[i]); \
printk("\n"); \
}) \
			CREDIT_SPIN_UNLOCK; \
			break; \
		} \
	} \
	else { \
		unsigned long j; \
		credit = 999999; \
		for (j=0;j<total_num_nodes;j++) { \
			if (j != phys_whoami) { \
				if (local_credit1[j]) { \
					local_credit[j] = local_credit1[j]; \
					local_credit1[j] = 0; \
				} \
				if (local_credit[j] < credit) \
					credit = local_credit[j]; \
			} \
		} \
		if (credit) { \
			curr_credit = credit; \
DEBUG_F({ \
unsigned long i; \
printk("credit refilled: "); \
for (i=0;i<total_num_nodes;i++) \
	printk("%ld ", local_credit[i]); \
printk("\n"); \
}) \
			CREDIT_SPIN_UNLOCK; \
			break; \
		} \
	} \
	/* avoid deadlock: poll() might invoke send() in turn,  */ \
	/* so we must deassert "gamma_tx_lock" (SMP only).	*/ \
	/* this however opens a window to reentrance in send(). */ \
	/* never mind, provided this macro does not reenter.    */ \
	/* that's the reason for the "gamma_cr_lock" (SMP only) */ \
        TX_SPIN_UNLOCK; \
	sys_gamma_poll(par_pid,procnum); \
        TX_SPIN_LOCK; \
	gamma_time(b); \
	if ( \
		gamma_time_diff_ticks(b,a) \
		 > (unsigned long long)MAX_CREDIT_WAIT*CLOCK \
	) { \
		i = 0; \
DEBUG_E({ \
unsigned long i; \
printk("credit NOT refilled yet: "); \
for (i=0;i<total_num_nodes;i++) \
	printk("%ld ", local_credit[i]); \
printk("\n"); \
}) \
		CREDIT_SPIN_UNLOCK; \
		break; \
	} \
} \
else i = 0; \
}


#define wait_credit_refill_samenode(i,par_pid,procnum) \
{ \
time_586 a,b; \
i = 1; \
gamma_time(a); \
if (CREDIT_SPIN_TRYLOCK) \
for (;;) { \
	ring_room = SHARED_RING_SIZE-(shared_ring.tail-shared_ring.head); \
	if (ring_room > 4) { \
		CREDIT_SPIN_UNLOCK; \
		break; \
	} \
	/* avoid deadlock: poll() might invoke send() in turn,  */ \
	/* so we must deassert "gamma_tx_lock" (SMP only).	*/ \
	/* this however opens a window to reentrance in send(). */ \
	/* never mind, provided this macro does not reenter.    */ \
	/* that's the reason for the "gamma_cr_lock" (SMP only) */ \
        TX_SPIN_UNLOCK; \
	sys_gamma_poll(par_pid,procnum); \
        TX_SPIN_LOCK; \
	gamma_time(b); \
	if ( \
		gamma_time_diff_ticks(b,a) \
		> (unsigned long long)MAX_CREDIT_WAIT*CLOCK \
	) { \
		i = 0; \
		CREDIT_SPIN_UNLOCK; \
		break; \
	} \
} \
else i = 0; \
}


#define busywait(n) \
{ \
	time_586 a,b; \
	gamma_time(a); \
	for (;;) { \
		gamma_time(b); \
		if ( \
gamma_time_diff_ticks(b,a) > (unsigned long long)(n)*CLOCK \
		) \
			break; \
	} \
}


#define tx_ring_is_full() (tx_ring_occupancy >= GAMMA_TX_RING_SIZE-5)


#define tx_ring_has_room() (tx_ring_occupancy < GAMMA_TX_RING_SIZE-20)


#define IRQ_enable_for_5usec(t_elaps,t_start) \
if ( \
gamma_time_diff_ticks((t_elaps),(t_start)) > (unsigned long long)9000*CLOCK \
) { \
/* \
 *  Open a 5 usec window for IRQs every 9 msec, necessary for \
 *  the correct management of the OS clock (gettimeoday()). \
 */ \
 save_flags(flags); \
 sti(); \
 busywait(5); \
 restore_flags(flags); \
 gamma_time(t_start); \
}
