#ifdef GAMMA_INSTALLING
#include "gamma_time.h"
#else
#include <linux/gamma/gamma_time.h>
#endif

#define DONT_MIND 0

#define MAX_PORT_NUMBER 1024	/*
				 *  Max port number allowed by each GAMMA
				 *  process; ports are numbered from 0 to
				 *  MAX_PORT_NUMBER, so MAX_PORT_NUMBER+1 ports
				 *  per process.
				 *  Must not exceed 2^15-1 (lower 15 bits useful)
				 *  Port #MAX_PORT_NUMBER reserved to barrier sync.
				 *  Port #MAX_PORT_NUMBER-1 reserved to bcast.
				 */
#define MAX_SYNC_SEM_NUMBER	MAX_PORT_NUMBER
				/*
				 *  Max sync. semaphore number allowed by each
				 *  GAMMA process; they are numbered from 0 to
				 *  MAX_SYNC_SEM_NUMBER, so MAX_SYNC_SEM_NUMBER+1
				 *  semaphores per process.
				 *  Must not exceed 2^15-1 (lower 15 bits useful)
				 *  Sem #MAX_PORT_NUMBER reserved to barrier sync.
				 *  Sem #MAX_PORT_NUMBER-1 reserved to bcast.
				 */
#define MAX_PAR_PID 255		/*
				 *  Max par_pid number allowed.
				 *  Must not exceed the unsigned char range.
				 *  par_pid #0 is reserved.
				 */
#define MAX_NUM_NODES 32	/*
				 *  Max no. of machines in the cluster.
				 *  Must not exceed the unsigned short range.
				 *  Must be < BROADCAST (below).
				 */
#define BROADCAST 255		/*
				 *  A conventional node number
				 *  which denotes "all nodes".
				 *  Must not exceed the unsigned short range.
				 */


				/*  error codes  */
#define GAMMA_ERR_GENERIC		1
#define GAMMA_ERR_FRAMELOST		2
#define GAMMA_ERR_MSGTOOLONG		3
#define GAMMA_ERR_SENDFAILURE		4
#define GAMMA_ERR_RECVQUEUEFULL		5
#define GAMMA_ERR_TRYAGAIN		6
#define GAMMA_ERR_MUSTYIELD		7
#define GAMMA_ERR_NOCREDIT		8
#define GAMMA_ISEND_IS_DONE		9
#define GAMMA_NO_RX_YET			10
#define GAMMA_RX_INCOMPLETE		11
#define GAMMA_ERR_BUSY			12
#define GAMMA_ERR_MAX			255  /*  No greater err codes!  */

/*
 *  These define the behaviour of GAMMA driver over user-provided buffers.
 *  GO_BACK means that each message directed to a buffer will be stored
 *  at its beginning.  If not otherwise programmed by a receiver handler,
 *  this implies that a subsequent message will overwrite the current one
 *  in the buffer.
 *  GO_AHEAD means that a subsequent message will be stored next and
 *  adjacent the current one in the buffer, unless otherwise specified by
 *  a receiver handler.
 */
/*  NONE OF THEM MAY BE 0  */
#define GO_BACK 1
#define GO_AHEAD 2

typedef struct {
	unsigned long addrl;
	unsigned short addrw;
	unsigned char addr[6];
} eth_addr;
