#undef USE_2P
#define USE_ISEND
#undef USE_FLOWCTL
#define USE_BCAST

#define SEND_TIME
#undef PERFORMANCE
#undef PERF_SHORT
#undef DELAY
#undef FREQUENCIES

#define CORRECTNESS

#ifdef USE_FLOWCTL
#define gamma_SEND(a,b,c)  gamma_send_flowctl(a,b,c)
#define gammaSEND2P(a,b,c,d,e)  gamma_send_2p_flowctl(a,b,c,d,e)
#define gamma_ISEND(a,b,c)  gamma_isend_flowctl(a,b,c)
#define gammaISEND2P(a,b,c,d,e)  gamma_isend_2p_flowctl(a,b,c,d,e)
#else
#define gamma_SEND(a,b,c)  gamma_send(a,b,c)
#define gammaSEND2P(a,b,c,d,e)  gamma_send_2p(a,b,c,d,e)
#define gamma_ISEND(a,b,c)  gamma_isend(a,b,c)
#define gammaISEND2P(a,b,c,d,e)  gamma_isend_2p(a,b,c,d,e)
#endif

#include <linux/gamma/libgamma.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include "measure.h"

#undef TRACE
#define MASTER_NODE 0
#define PORT 1


unsigned char *data,*data1;
int len;

#ifdef USE_2P
void handler(void) {
	gamma_switch_to_buffer(PORT,data+12,len);
	return;
}
#endif


int main(int argc, char **argv) {

#ifdef PERFORMANCE
#ifdef SEND_TIME
	double record_mc[NPROVE];
#endif
	double record_mm[NPROVE];
#endif
	int i,j,num_procs,z;
	time_586
		t_send_s,
		t_rec;
#ifdef PERFORMANCE
#ifdef SEND_TIME
	time_586
		t_send_e;
#endif
	double time_overhead;
#endif

   	num_procs = atoi(argv[1]);
	gamma_init(num_procs,argc,argv);

gamma_suspend_irq();

   	len = atoi(argv[2]);
	/*  Allocate data space (+4096 to allow page alignment)  */
#ifdef USE_2P
	data1 = (unsigned char *)malloc(4096+len*sizeof(unsigned char)+12);
#else
	data1 = (unsigned char *)malloc(4096+len*sizeof(unsigned char));
#endif
	/*  Align to page boundary  */
	data = (unsigned char *)(((unsigned long)data1+4096) & (~4095));

	/*  Init the message to be exchanged.  */
#ifdef USE_2P
	for (j=0;j<len+12;j++) {
		if (gamma_my_node() == MASTER_NODE)
			data[j] = (unsigned char)j;
		else
			data[j] = (unsigned char)0;
	}
#else
	for (j=0;j<len;j++) {
		if (gamma_my_node() == MASTER_NODE)
			data[j] = (unsigned char)j;
		else
			data[j] = (unsigned char)0;
	}
#endif


	if (gamma_my_node() == MASTER_NODE) {

		/*  this is "master" (inst. no. == 0)  */

#ifdef USE_2P
#ifndef USE_BCAST
		gamma_set_active_port_2p(
			PORT,
			(gamma_my_node()+1)%num_procs,
			gamma_my_par_pid(),PORT,
			handler,handler,
			PORT,GO_BACK,data,len+12,
			12
		);
#else
		gamma_set_active_port_2p(
			PORT,
			BROADCAST,
			gamma_my_par_pid(),PORT,
			handler,handler,
			PORT,GO_BACK,data,len+12,
			12
		);
#endif
#else
#ifndef USE_BCAST
		gamma_set_passive_port(
			PORT,
			(gamma_my_node()+1)%num_procs,
			gamma_my_par_pid(),PORT,
			PORT,GO_BACK,data,len
		);
#else
		gamma_set_passive_port(
			PORT,
			BROADCAST,gamma_my_par_pid(),PORT,
			PORT,GO_BACK,data,len
		);
#endif
#endif

#ifdef PERFORMANCE
		/*  Computing the gamma_time() overhead...  */
		for (i=0;i<NPROVE;i++) {
			gamma_time(t_send_s);
 			gamma_time(t_rec);
			record_mm[i] = gamma_time_diff(t_rec, t_send_s);
		}
		{
		double Mmm,medmm,varmm;
		mean_variance(record_mm,&time_overhead,&Mmm,&medmm,&varmm);
		}
#endif

		gamma_sync();

		for (i=0;i<NPROVE;i++) {
#ifdef CORRECTNESS
#ifdef USE_2P
			for (j=0;j<len+12;j++) data[j] = (unsigned char)j;
#else
			for (j=0;j<len;j++) data[j] = (unsigned char)j;
#endif
			gamma_sync();
#endif

#ifdef PERFORMANCE
			gamma_time(t_send_s);
#endif

#ifdef USE_ISEND
#ifdef USE_2P
			j = gammaISEND2P(PORT,data,12,data+12,len);
#else
			j = gamma_ISEND(PORT,data,len);
#endif
#else
#ifdef USE_2P
			gammaSEND2P(PORT,data,12,data+12,len);
#else
			gamma_SEND(PORT,data,len);
#endif
#endif

#ifdef PERFORMANCE
#ifdef SEND_TIME
			gamma_time(t_send_e);
#endif
#endif
#ifdef TRACE
printf("[%d]: inst 0 sent\n",i);fflush(stdout);
#endif
#ifndef USE_BCAST
			if (gamma_wait(PORT,1))
#else
			for (z=0;z<num_procs-1;z++) {
				gamma_sync();
				if (gamma_wait(PORT,1))
#endif
			{
				printf("instance 0:  ERROR\n");
				fflush(stdout);
				free(data1);
				gamma_exit();
				exit(-1);
			}
#ifdef USE_BCAST
			}
#endif
#ifdef TRACE
printf("[%d]: inst 0 recvd\n",i);fflush(stdout);
#endif
#ifdef PERFORMANCE
 			gamma_time(t_rec);
#endif
#ifdef USE_ISEND
			gamma_wsend(j);
#endif

			/*
			 *  Force printout of kernel-level time measurements,
			 *  if enabled.  Otherwise nothing happens.
			 */
			gamma_timers(t_send_s,t_rec);

#ifdef CORRECTNESS
			/*  Check data correctness.  */
#ifdef USE_2P
                        for (j=0;j<len+12;j++)
                                if (data[j] != (unsigned char)j) {
                                        int k;
                                        printf("data error\n");
                                        for (k=j-5;k<j+10;k++)
						printf("%d:%d ",k,data[k]);
                                        printf("\n");
                                }
#else
                        for (j=0;j<len;j++)
                                if (data[j] != (unsigned char)j) {
                                        int k;
                                        printf("data error\n");
                                        for (k=j-5;k<j+10;k++)
						printf("%d:%d ",k,data[k]);
                                        printf("\n");
				sleep(1);
                                }
#endif
#endif

#ifdef PERFORMANCE
#ifdef SEND_TIME
			record_mc[i] =
			gamma_time_diff(t_send_e, t_send_s)-time_overhead;
#endif
			record_mm[i] =
			gamma_time_diff(t_rec, t_send_s)/num_procs
			-time_overhead;
#endif

		}

gamma_exit();
#ifdef PERFORMANCE
#ifdef PERF_SHORT
		{
#ifdef SEND_TIME
			double mmc,Mmc,medmc,varmc;
			mean_variance(record_mc,&mmc,&Mmc,&medmc,&varmc);
#ifdef DELAY
		        printf("%d %g\n",len,medmc);
#else
			if (!len)
			        printf("%d %g\n",len,medmc);
			else
			        printf("%d %g\n",len,len/medmc);
#endif
#else
			double mmm,Mmm,medmm,varmm;
			mean_variance(record_mm,&mmm,&Mmm,&medmm,&varmm);
#ifdef DELAY
		        printf("%d %g\n",len,medmm);
#else
			if (!len)
			        printf("%d %g\n",len,medmm);
			else
			        printf("%d %g\n",len,len/medmm);
#endif
#endif
		}
#else
		{
		printf("\n\n");
		printf("PERFORMANCE (microsec, MB/sec) (best < average < worst):\n\n");
		{
#ifdef SEND_TIME
			double mmc,Mmc,medmc,varmc;
#endif
			double mmm,Mmm,medmm,varmm;
#ifdef SEND_TIME
			mean_variance(record_mc,&mmc,&Mmc,&medmc,&varmc);
		        printf("total send time        %g < %g < %g\n",
				mmc,medmc,Mmc);
			if (len)
		        printf("send throughput        %g < %g < %g\n",
				len/Mmc,len/medmc,
				len/mmc);
#endif
			mean_variance(record_mm,&mmm,&Mmm,&medmm,&varmm);
		        printf("total delay            %g < %g < %g\n",
				mmm,medmm,Mmm);
			if (len)
		        printf("end-to-end throughput  %g < %g < %g\n",
				len/Mmm,len/medmm,
				len/mmm);
			printf("\n\n");

		        printf("VARIANCE:  \n");
#ifdef SEND_TIME
		        printf("send time        %g\n", varmc);
#endif
		        printf("end-to-end delay %g\n", varmm);
			printf("\n\n");

#ifdef FREQUENCIES
		        printf("FREQUENCIES:\n");
#ifdef SEND_TIME
			printf("send time:\n");
			frequencies(record_mc,mmc,Mmc);
#endif
			printf("end-to-end delay:\n");
			frequencies(record_mm,mmm,Mmm);
#endif
			fflush(stdout);

		}
		}
#endif
#endif
	}


	else {
		/*  this is a "slave" (inst. no.  != 0)  */
#ifdef USE_2P
#ifndef USE_BCAST
		gamma_set_active_port_2p(
			PORT,
			(gamma_my_node()+1)%num_procs,
			gamma_my_par_pid(),PORT,
			handler,handler,
			PORT,GO_BACK,data,len+12,
			12
		);
#else
		gamma_set_active_port_2p(
			PORT,
			BROADCAST,
			gamma_my_par_pid(),PORT,
			handler,handler,
			PORT,GO_BACK,data,len+12,
			12
		);
#endif
#else
#ifndef USE_BCAST
		gamma_set_passive_port(
			PORT,
			(gamma_my_node()+1)%num_procs,
			gamma_my_par_pid(),PORT,
			PORT,GO_BACK,data,len
		);
#else
		gamma_set_passive_port(
			PORT,
			BROADCAST,gamma_my_par_pid(),PORT,
			PORT,GO_BACK,data,len
		);
#endif
#endif
		gamma_sync();

		for (i=0;i<NPROVE;i++) {

#ifdef CORRECTNESS
#ifdef USE_2P
			for (j=0;j<len+12;j++) data[j] = 0;
#else
			for (j=0;j<len;j++) data[j] = 0;
#endif
			gamma_sync();
#endif

#ifndef USE_BCAST
			if (gamma_wait(PORT,1))
#else
			for (z=0;z<gamma_my_node();z++) {
				gamma_sync();
				if (gamma_wait(PORT,1))
#endif
			{
				printf("instance %d:  ERROR\n",gamma_my_node());
				fflush(stdout);
				free(data1);
				gamma_exit();
				exit(-1);
			}
#ifdef USE_BCAST
			}
#endif
#ifdef TRACE
printf("[%d]: inst %d recvd\n",i,gamma_my_node());fflush(stdout);
#endif
#ifdef USE_ISEND
#ifdef USE_2P
			j = gammaISEND2P(PORT,data,12,data+12,len);
			gamma_wsend(j);
#else
			j = gamma_ISEND(PORT,data,len);
			gamma_wsend(j);
#endif
#else
#ifdef USE_2P
			gammaSEND2P(PORT,data,12,data+12,len);
#else
			gamma_SEND(PORT,data,len);
#endif
#endif
#ifdef TRACE
printf("[%d]: inst %d sent\n",i,gamma_my_node());fflush(stdout);
#endif
#ifdef USE_BCAST
			for (z=0;z<num_procs-gamma_my_node()-1;z++) {
				gamma_sync();
				gamma_wait(PORT,1);
			}
#endif
			continue;
		}
gamma_exit();
	}
free(data1);
return 0;
}
